##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Ape configuration assembler.

$Id: apeconf.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""
from __future__ import nested_scopes

from apelib.core.mapper import Mapper
from apelib.core.serializers import CompositeSerializer, AnyObjectSerializer
from apelib.core.gateways import CompositeGateway
from apelib.core.interfaces import IDatabaseInitializer

from minitables import Table, TableSchema
from common import Directive, DirectiveReader, ComponentSystem

class AssemblyError(Exception):
    """Error while assembling components"""


class MapperDeclaration(Directive):
    schema = TableSchema()
    schema.addColumn('mapper_name', primary=1)


class MapperAttribute(Directive):
    schema = TableSchema()
    schema.addColumn('mapper_name', primary=1)
    # Attribute names: 'class', 'parent', 'extends'
    schema.addColumn('name', primary=1, indexed=1)
    schema.addColumn('value')


class ComponentDefinition(Directive):
    schema = TableSchema()
    # comptypes: 'serializer', 'gateway', 'classifier', 'keygen'
    schema.addColumn('comptype', primary=1)
    schema.addColumn('name', primary=1)
    schema.addColumn('producer')


class MapperComponent(Directive):
    schema = TableSchema()
    schema.addColumn('mapper_name', primary=1)
    schema.addColumn('comptype', primary=1)
    schema.addColumn('producer')


class MapperCompositeComponent(Directive):
    schema = TableSchema()
    schema.addColumn('mapper_name', primary=1)
    schema.addColumn('comptype', primary=1)
    schema.addColumn('name', primary=1)
    schema.addColumn('producer')
    schema.addColumn('order')


class MapperRegistration(Directive):
    # Contains a use-for directive
    schema = TableSchema()
    schema.addColumn('mapper_name', primary=1)
    schema.addColumn('attr', primary=1)
    schema.addColumn('value', primary=1)  # Multiple values allowed


class ClassifierOption(Directive):
    schema = TableSchema()
    schema.addColumn('mapper_name', primary=1)
    schema.addColumn('option', primary=1)
    schema.addColumn('value')


class DisabledProducer:
    def __init__(self, source):
        self.source = source

    def __call__(self, compsys):
        return None


class UseProducer:
    def __init__(self, source, comptype, name):
        self.source = source
        self.comptype = comptype
        self.name = name

    def __call__(self, compsys):
        return compsys.get(self.comptype, self.name)


class FactoryProducer:

    def __init__(self, source, factory_name, param=None):
        self.source = source
        pos = factory_name.rfind('.')
        if pos <= 0:
            raise ValueError, "Module and class name required"
        self.module_name = factory_name[:pos]
        self.class_name = factory_name[pos + 1:]
        self.param = param
        self.sub_producer = None

    def __call__(self, compsys):
        params = []
        if self.param is not None:
            params.append(self.param)
        if self.sub_producer is not None:
            params.append(self.sub_producer(compsys))
        m = __import__(self.module_name, {}, {}, ('__doc__',))
        try:
            c = getattr(m, self.class_name)
        except AttributeError:
            raise ImportError, "No class %s in module %s" % (
                self.class_name, self.module_name)
        return c(*params)


def makeProducer(source, comptype, attrs, raise_exc=1):
    if attrs.get('enabled', '').lower() == 'false':
        return DisabledProducer(source)
    elif attrs.has_key('use'):
        if attrs.has_key('factory'):
            raise ValueError, "Both 'use' and 'factory' not allowed"
        return UseProducer(source, comptype, attrs['use'])
    elif attrs.has_key('factory'):
        param = attrs.get('param')
        return FactoryProducer(source, attrs['factory'], param)
    elif raise_exc:
        raise ValueError, "Need a 'use', 'factory', or 'enabled' attribute"
    else:
        return None


def getElementHandlers():
    """Returns a dictionary of XML element handlers.
    """

    def handle_configuration(source, vars, attrs):
        assert vars.has_key('directives')

    def handle_variation(source, vars, attrs):
        variations = vars['variations']
        name = attrs['name']
        d = variations.get(name)
        if d is None:
            # Make a new variation.
            d = []
            variations[name] = d
        vars['directives'] = d

    def handle_mapper(source, vars, attrs):
        d = vars['directives']
        mapper_name = str(attrs['name'])
        vars['mapper_name'] = mapper_name
        d.append(MapperDeclaration(source, mapper_name))
        for key in ('class', 'parent', 'extends'):
            if attrs.has_key(key):
                d.append(MapperAttribute(source, mapper_name, key, attrs[key]))

    def handle_component(source, vars, attrs, comptype, multiple):
        d = vars['directives']
        producer = makeProducer(source, comptype, attrs)
        mapper_name = vars.get('mapper_name')
        if mapper_name is None:
            # Reusable component
            name = attrs['name']
            directive = ComponentDefinition(source, comptype, name, producer)
        elif multiple:
            # Composite component of a mapper
            name = attrs['name']
            directive = MapperCompositeComponent(
                source, mapper_name, comptype,
                name, producer, attrs.get('order', 'middle'))
        else:
            # Singular component of a mapper
            directive = MapperComponent(
                source, mapper_name, comptype, producer)
        d.append(directive)
        return producer

    def handle_serializer(source, vars, attrs):
        handle_component(source, vars, attrs, 'serializer', multiple=1)

    def handle_gateway(source, vars, attrs):
        p = vars.get('classifier_producer')
        if p is not None:
            # Set a gateway for a classifier.
            if not hasattr(p, 'sub_producer'):
                raise ValueError(
                    "Classifier at %s needs a factory in order to "
                    "use a gateway" % source)
            if p.sub_producer is not None:
                raise ValueError(
                    "Multiple gateways in classifiers not allowed at %s" %
                    source)
            p.sub_producer = makeProducer(source, 'gateway', attrs)
        else:
            handle_component(source, vars, attrs, 'gateway', multiple=1)

    def handle_classifier(source, vars, attrs):
        p = handle_component(source, vars, attrs, 'classifier', multiple=0)
        vars['classifier_producer'] = p

    def handle_keygen(source, vars, attrs):
        handle_component(source, vars, attrs, 'keygen', multiple=0)

    def handle_use_for(source, vars, attrs):
        d = vars['directives']
        mapper_name = vars['mapper_name']
        for attr in ('class', 'extensions', 'fallback', 'key'):
            if attrs.has_key(attr):
                v = attrs[attr]
                if attr == 'extensions':
                    first = 1
                    for ext in v.split():
                        if not ext.startswith('.'):
                            ext = '.' + ext
                        ext = ext.lower()
                        d.append(MapperRegistration(
                            source, mapper_name, 'extension', ext))
                        if first:
                            # Use a classifier option to set the default
                            # extension.
                            first = 0
                            d.append(ClassifierOption(
                                source, mapper_name, 'default_extension', ext))
                elif attr == 'key':
                    try:
                        v = int(v)
                    except ValueError:
                        v = str(v)
                    d.append(MapperRegistration(
                        source, mapper_name, attr, v))
                else:
                    d.append(MapperRegistration(
                        source, mapper_name, attr, v))

    def handle_option(source, vars, attrs):
        d = vars['directives']
        mapper_name = vars['mapper_name']
        name = attrs['name']
        value = attrs['value']
        d.append(ClassifierOption(source, mapper_name, name, value))

    handlers = {
        'configuration': handle_configuration,
        'variation':     handle_variation,
        'mapper':        handle_mapper,
        'serializer':    handle_serializer,
        'gateway':       handle_gateway,
        'classifier':    handle_classifier,
        'keygen':        handle_keygen,
        'use-for':       handle_use_for,
        'option':        handle_option,
        }

    return handlers



class BasicComponentAssembler:
    """Assembler for simple components.

    Configures at the time of creation.
    """

    def __init__(self, compsys, comptype, name):
        self.compsys = compsys
        records = compsys.dtables.query(
            ComponentDefinition, comptype=comptype, name=name)
        if not records:
            raise AssemblyError("No %s component named %s exists"
                                % (comptype, repr(name)))
        assert len(records) == 1
        producer = records[0]['producer']
        self.producer = producer

    def create(self):
        return self.producer(self.compsys)

    def configure(self):
        pass


class MapperAssembler:
    """Assembler for mapper components.
    """

    def __init__(self, compsys, comptype, name):
        self.compsys = compsys
        self.dtables = dtables = compsys.dtables
        self.mapper_name = name
        if not dtables.query(MapperDeclaration, mapper_name=name):
            raise AssemblyError("No mapper named %s exists" % repr(name))
        self.subobjs = []  # all subobjects
        self.attrs = {}
        for record in dtables.query(MapperAttribute, mapper_name=name):
            self.attrs[record['name']] = record['value']
        self.prepareSubComponents()

    def prepareSubComponents(self):
        self.single_comps = {} # comptype -> record
        self.multi_comps = {}  # comptype -> name -> record
        name = self.mapper_name
        all_names = []  # mapper_name and all of its base mapper_names
        while name:
            all_names.append(name)
            records = self.dtables.query(MapperComponent, mapper_name=name)
            for r in records:
                self.single_comps.setdefault(r['comptype'], r)
            records = self.dtables.query(
                MapperCompositeComponent, mapper_name=name)
            for r in records:
                d = self.multi_comps.setdefault(r['comptype'], {})
                d.setdefault(r['name'], r)
            name = self.dtables.queryField(
                MapperAttribute, 'value', mapper_name=name, name='extends')
            if name and name in all_names:
                raise AssemblyError(
                    "Circular extension in mappers %s" % repr(all_names))
        self.sub_mapper_names = []
        records = self.dtables.query(
            MapperAttribute, name='parent', value=self.mapper_name)
        for r in records:
            self.sub_mapper_names.append(r['mapper_name'])

    def create(self):
        self.obj = Mapper()
        return self.obj

    def configure(self):
        self.setParent()
        self.addSerializers()
        self.addGateways()
        self.setClassifier()
        self.setKeygen()
        self.addInitializers()
        self.addSubMappers()
        self.registerClassifications()

    def setParent(self):
        if self.attrs.get('parent'):
            p = self.compsys.get('mapper', self.attrs['parent'])
            self.obj.setParent(p)
        else:
            self.obj.setParent(None)

    def addSerializers(self):
        cname = self.attrs.get('class')
        if cname == 'any':
            # This mapper is usable for many classes
            s = AnyObjectSerializer()
        elif cname == 'none':
            # This mapper is abstract (usable for no classes)
            s = CompositeSerializer(None, None)
        else:
            # This mapper is concrete (usable for one class only)
            if cname is None:
                cname = self.mapper_name
            pos = cname.rfind('.')
            if pos < 0:
                raise AssemblyError("Class name must include a module name")
            s = CompositeSerializer(cname[:pos], cname[pos + 1:])
        
        d = self.multi_comps.get('serializer')
        if d:
            ordered = [(r.get('order', '').lower(), name, r)
                       for name, r in d.items()]
            ordered.sort()
            for order, name, r in ordered:
                o = r['producer'](self.compsys)
                if o is not None:
                    s.addSerializer(str(name), o)
                    self.subobjs.append(o)
        self.obj.setSerializer(s)

    def addGateways(self):
        g = CompositeGateway()
        d = self.multi_comps.get('gateway')
        if d:
            for name, r in d.items():
                o = r['producer'](self.compsys)
                if o is not None:
                    g.addGateway(str(name), o)
                    self.subobjs.append(o)
        self.obj.setGateway(g)

    def setClassifier(self):
        r = self.single_comps.get('classifier')
        if r:
            o = r['producer'](self.compsys)
            self.classifier = o
            if o is not None:
                self.obj.setClassifier(o)
                self.subobjs.append(o)
                gw = o.getGateway()
                if gw is not None:
                    self.subobjs.append(gw)

    def setKeygen(self):
        r = self.single_comps.get('keygen')
        if r:
            o = r['producer'](self.compsys)
            if o is not None:
                self.obj.setKeychainGenerator(o)
                self.subobjs.append(o)

    def addInitializers(self):
        for o in self.subobjs:
            if IDatabaseInitializer.isImplementedBy(o):
                self.obj.addInitializer(o)

    def addSubMappers(self):
        for name in self.sub_mapper_names:
            o = self.compsys.get('mapper', name)
            self.obj.addSubMapper(name, o)

    def registerClassifications(self):
        """Registers classifications on behalf of sub-mappers."""
        all_regs = {}     # { (attr, value) -> mapper_name }
        all_options = {}  # { (mapper_name, option) -> value }
        for name in self.sub_mapper_names:
            # use-for directives
            records = self.dtables.query(
                MapperRegistration, mapper_name=name)
            for r in records:
                key = ((r['attr'], r['value']))
                if all_regs.has_key(key) and all_regs[key] != name:
                    raise AssemblyError(
                        "Mappers %s and %s are contending over %s == %s" % (
                        name, all_regs[key],
                        r['attr'], repr(r['value'])))
                all_regs[key] = name

            # class="" attributes
            class_name = self.dtables.queryField(
                MapperAttribute, 'value', mapper_name=name, name='class')
            if class_name is None:
                class_name = name
            elif class_name in ('none', 'any'):
                class_name = None
            if class_name is not None:
                # Add an implicit use-for directive
                key = ('class', class_name)
                if all_regs.has_key(key) and all_regs[key] != name:
                    raise AssemblyError(
                        "Mappers %s and %s are contending over %s == %s" % (
                        name, all_regs[key],
                        'class', repr(class_name)))
                all_regs[key] = name

            # options
            records = self.dtables.query(
                ClassifierOption, mapper_name=name)
            for r in records:
                all_options[(name, r['option'])] = r['value']

        if all_regs or all_options:
            if self.classifier is None:
                raise AssemblyError(
                    "Mapper %s needs a classifier because it has "
                    "sub-mappers with registrations" % self.mapper_name)
            for (attr, value), name in all_regs.items():
                self.classifier.register(attr, value, name)
            for (name, option), value in all_options.items():
                self.classifier.setOption(name, option, value)


def wireMapper(filenames, vname='', mapper_name='root'):
    """Returns an Ape mapper built according to the configuration files.
    """
    handlers = getElementHandlers()
    reader = DirectiveReader(handlers)
    for fn in filenames:
        reader.read(fn)
    directives = reader.getDirectives(vname)
    cs = ComponentSystem(directives)
    cs.addComponentType('mapper', MapperAssembler)
    for comptype in ('serializer', 'gateway', 'classifier', 'keygen'):
        cs.addComponentType(comptype, BasicComponentAssembler)
    mapper = cs.get('mapper', mapper_name)
    mapper.checkConfiguration(path=mapper_name)
    return mapper

