##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""In-memory tables with support for basic relational operations.

$Id: minitables.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""


from BTrees.IIBTree import IISet, intersection


class DuplicateError(Exception):
    """Duplicated data record"""


class Column:

    def __init__(self, name, primary, indexed):
        self.name = name        # string
        self.primary = primary  # boolean
        self.indexed = indexed  # boolean


class TableSchema:

    reserved_names = ('rid',)

    def __init__(self):
        self.columns = []
        self.column_names = {}
        self.primary_names = []

    def addColumn(self, name, primary=0, indexed=0):
        if name in self.reserved_names:
            raise ValueError, "Column name %s is reserved" % repr(name)
        if self.column_names.has_key(name):
            raise ValueError, "Column %s already exists" % repr(name)
        self.column_names[name] = 1
        self.columns.append(Column(name, primary, indexed))
        if primary:
            self.primary_names.append(name)

    def getColumns(self):
        return tuple(self.columns)

    def getPrimaryNames(self):
        return self.primary_names



class Table:
    """Simple, generic relational table.
    """
    schema = None

    def __init__(self, schema=None):
        if schema is not None:
            self.schema = schema
        self.columns = self.schema.getColumns()
        self.next_rid = 1
        self.data = {}     # {rid -> record as {column name -> value}}
        self.indexes = {}  # {index_name -> {value -> IISet}}
        self.primary_index = {}  # {primary key -> rid}
        for c in self.columns:
            if c.indexed:
                self.indexes[c.name] = {}


    def makeRecord(self, params):
        """Returns a record containing {name -> value}.
        """
        res = {}
        columns = self.columns
        for column in columns:
            name = column.name
            if params.has_key(name):
                res[name] = params[name]
        if len(params) > len(res):
            raise ValueError, "Too many parameters"
        return res


    def insert(self, params):
        record = self.makeRecord(params)

        # Determine the primary key.
        primary_key = []
        for column in self.columns:
            if column.primary:
                if not record.has_key(column.name):
                    raise ValueError, (
                        "No value provided for primary key column %s"
                        % repr(column.name))
                primary_key.append(record[column.name])
        if primary_key:
            primary_key = tuple(primary_key)
            if self.primary_index.has_key(primary_key):
                raise DuplicateError(
                    "Primary key %s in use" % repr(primary_key))

        # Add a record.
        rid = self.next_rid
        self.next_rid += 1
        record['rid'] = rid
        self.data[rid] = record
        if primary_key:
            self.primary_index[primary_key] = rid

        # Add to indexes.
        for name, value in record.items():
            if self.indexes.has_key(name):
                set = self.indexes[name].get(value)
                if set is None:
                    set = IISet()
                    self.indexes[name][value] = set
                set.insert(rid)

        # Return the number of rows inserted.
        return 1


    def update(self, filter, changes):
        rids = self._select_rids(self.makeRecord(filter))
        if rids is None:
            rids = self.data.keys()
        elif not rids:
            # Nothing needs to be updated.
            return 0

        # Identify changes.
        old_data = {}    # rid -> old record
        new_data = {}    # rid -> new record
        old_to_new = {}  # old primary key -> new primary key
        new_to_rid = {}  # new primary key -> rid

        changes = self.makeRecord(changes)
        for rid in rids:
            old_r = self.data[rid]
            old_data[rid] = old_r
            new_r = old_r.copy()
            new_r.update(changes)
            new_data[rid] = new_r
            opk = []
            npk = []
            for column in self.columns:
                if column.primary:
                    opk.append(old_r[column.name])
                    npk.append(new_r[column.name])
            if opk != npk:
                opk = tuple(opk)
                npk = tuple(npk)
                old_to_new[opk] = npk
                new_to_rid[npk] = rid

        # Look for primary key conflicts.  A primary key conflict can
        # occur when changing a record to a different primary key and
        # the new primary key is already in use.
        for pk in old_to_new.values():
            if (self.primary_index.has_key(pk)
                and not old_to_new.has_key(pk)):
                raise DuplicateError("Primary key %s in use" % repr(pk))

        # Update the data.
        self.data.update(new_data)

        # Remove old primary key indexes and insert new primary key indexes.
        for pk in old_to_new.keys():
            del self.primary_index[pk]
        self.primary_index.update(new_to_rid)

        # Update indexes.
        for rid, old_r in old_data.items():
            for column in self.columns:
                name = column.name
                index = self.indexes.get(name)
                if index is not None and changes.has_key(name):
                    if changes[name] != old_r[name]:
                        # Remove an index entry.
                        old_value = old_r[name]
                        set = index[old_value]
                        set.remove(rid)
                        if not set:
                            del index[old_value]
                        # Add an index entry.
                        new_value = changes[name]
                        set = index.get(new_value)
                        if set is None:
                            set = IISet()
                            index[new_value] = set
                        set.insert(rid)

        # Return the number of rows affected.
        return len(rids)


    def select(self, filter):
        rids = self._select_rids(self.makeRecord(filter))
        if rids is None:
            # All
            return self.data.values()
        elif rids:
            # Some
            return [self.data[rid] for rid in rids]
        else:
            # None
            return []


    def _select_rids(self, query):
        """Searches the table for matches, returning record ids.

        Returns a sequence of record ids, or None for all records.
        """
        # Shortcut: if no query, return all.
        if not query:
            return None

        # First strategy: try to satisfy the request by consulting
        # the primary key index.
        primary_key = []
        filter_columns = []
        for column in self.columns:
            name = column.name
            if query.has_key(name):
                if column.primary:
                    primary_key.append(query[name])
                else:
                    # Specified a value that's not in the primary key
                    filter_columns.append((name, query[name]))
            elif column.primary:
                # Didn't fully specify a primary key
                break
        else:
            if primary_key:
                # To satisfy the request, we only need to look at
                # primary index.
                primary_key = tuple(primary_key)
                rid = self.primary_index.get(primary_key)

                # Possibly filter out the single item
                if rid is not None and filter_columns:
                    cand = self.data[rid]
                    for name, value in filter_columns:
                        if cand[name] != value:
                            # Not a match.
                            rid = None
                            break

                if rid is None:
                    return ()
                else:
                    return (rid,)

        # Second strategy: try to satisfy the request by intersecting
        # indexes.
        filter_columns = []
        rids = None
        for name, value in query.items():
            if self.indexes.has_key(name):
                set = self.indexes[name].get(value)
                if set is None:
                    # No rows satisfy this criterion.
                    return ()
                if rids is None:
                    rids = set
                else:
                    rids = intersection(rids, set)
                if not rids:
                    # No rows satisfy all criteria.
                    return ()
            else:
                filter_columns.append((name, value))
        if not filter_columns:
            # No need to search each record.
            return rids

        # Fallback strategy: Eliminate items one by one.
        if rids is None:
            # Use the whole data set.
            candidates = self.data.items()
        else:
            # Use the specified records.
            candidates = [(rid, self.data[rid]) for rid in rids]

        rids = []
        for rid, cand in candidates:
            for name, value in filter_columns:
                if not cand.has_key(name) or cand[name] != value:
                    # Not a match.
                    break
            else:
                # A match.
                rids.append(rid)
        return rids

