##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Standard event implementations

$Id: events.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

import interfaces


SIMPLE_IMMUTABLE_OBJECTS = (None, (), 0, 1, '', u'')

try:
    True
except NameError:
    pass
else:
    SIMPLE_IMMUTABLE_OBJECTS += (False, True)


class DatabaseInitEvent:
    """Database initialization event."""

    __implements__ = interfaces.IDatabaseInitEvent

    def __init__(self, connections, clearing):
        self._connections = connections
        self._clearing = clearing

    def getConnection(self, name):
        """Returns the named connection."""
        return self._connections[name]

    def clearing(self):
        """Returns true if the database is to be cleared."""
        return self._clearing


class MapperEvent:

    __implements__ = interfaces.IMapperEvent

    def __init__(self, mapper, keychain):
        self._mapper = mapper
        self._keychain = keychain

    def getMapper(self):
        """Returns the mapper for the object being (de)serialized."""
        return self._mapper

    def getKeychain(self):
        """Returns the keychain of the object being (de)serialized."""
        return self._keychain

    def getKey(self):
        """Returns the last element of the keychain."""
        return self._keychain[-1]

    def makeKeychain(self, name, stored):
        kcg = self.getMapper().getKeychainGenerator()
        return kcg.makeKeychain(self, name, stored)


class GatewayEvent (MapperEvent):

    __implements__ = interfaces.IGatewayEvent

    def __init__(self, mapper, keychain, connections, classification):
        MapperEvent.__init__(self, mapper, keychain)
        self._connections = connections
        self._classification = classification

    def getConnection(self, name):
        """Returns the named connection."""
        return self._connections[name]

    def getClassification(self):
        """Returns the innermost classification or None."""
        return self._classification


class LoadEvent (GatewayEvent):
    """Object loading event."""

    __implements__ = interfaces.ILoadEvent


class StoreEvent (GatewayEvent):
    """Object storing event."""

    __implements__ = interfaces.IStoreEvent


class SDEvent (MapperEvent):

    __implements__ = interfaces.ISDEvent

    _serializer_name = ''

    def __init__(self, keyed_ob_sys, object_mapper, keychain, object):
        MapperEvent.__init__(self, object_mapper, keychain)
        self._keyed_ob_sys = keyed_ob_sys
        self._object = object
        self._unmanaged = []
        # _refs is the list of externally referenced objects.
        # It has the form [(keychain, value)]
        self._refs = []

    def getKeyedObjectSystem(self):
        """Returns the IKeyedObjectSystem that generated this event.

        It is needed for loading and identifying other objects.
        """
        return self._keyed_ob_sys

    def setSerializerName(self, name):
        """Sets the name of the next serializer."""
        assert ':' not in name
        self._serializer_name = name

    def getSerializerName(self):
        """Returns the name of the serializer in use."""
        return self._serializer_name

    def getObject(self):
        """Returns the object being serialized."""
        return self._object

    def addUnmanagedPersistentObjects(self, obs):
        """Notifies that there are unmanaged persistent objects in the object.
        """
        self._unmanaged.extend(obs)

    def getUnmanagedPersistentObjects(self):
        """Returns the list of unmanaged persistent objects."""
        return self._unmanaged

    def getExternalRefs(self):
        """Returns the list of external references"""
        return self._refs


class DeserializationEvent (SDEvent):

    __implements__ = interfaces.IFullDeserializationEvent

    def __init__(self, keyed_ob_sys, object_mapper, keychain, object):
        SDEvent.__init__(
            self, keyed_ob_sys, object_mapper, keychain, object)
        self._loaded_refs = {}  # { (serializer_name, name) -> object }

    # IDeserializationEvent interface methods:

    def notifyDeserialized(self, name, value):
        """See the IDeserializationEvent interface."""
        assert self._serializer_name is not None
        self._loaded_refs['%s:%s' % (self._serializer_name, name)] = value

    def dereference(self, name, keychain, hints=None):
        """Retrieves a referenced subobject (usually ghosted initially).
        """
        kos = self.getKeyedObjectSystem()
        ob = kos.getObject(keychain, hints)
        self._refs.append((keychain, ob))
        self.notifyDeserialized(name, ob)
        return ob

    # IFullDeserializationEvent interface methods:

    def loadInternalRef(self, ref):
        """Returns an object already deserialized by another serializer.

        'ref' is a tuple containing (serializer_name, name).
        """
        return self._loaded_refs[ref]


class SerializationEvent (SDEvent):

    __implements__ = interfaces.IFullSerializationEvent

    def __init__(self, keyed_ob_sys, object_mapper, keychain, object):
        SDEvent.__init__(
            self, keyed_ob_sys, object_mapper, keychain, object)
        self._attrs = {}
        # _internal_refs:
        # id(ob) -> (serializer_name, name)
        self._internal_refs = {}
        # _internal_ref_list contains all objects that may be referenced
        # internally.  This only ensures that id(ob) stays consistent.
        self._internal_ref_list = []

    # ISerializationEvent interface methods:

    def notifySerialized(self, name, value, is_attribute):
        """See the ISerializationEvent interface."""
        assert self._serializer_name is not None
        for ob in SIMPLE_IMMUTABLE_OBJECTS:
            # If value is a simple immutable object, don't make a
            # reference to it.  Compare by identity rather than
            # equality, otherwise rich comparison leads to surprises.
            if value is ob:
                break
        else:
            # Make internal references only for mutable or complex objects.
            idx = id(value)
            if not self._internal_refs.has_key(idx):
                self._internal_ref_list.append(value)
                if name is not None:
                    self._internal_refs[idx] = (
                        '%s:%s' % (self._serializer_name, name))
                else:
                    self._internal_refs[idx] = None
        if is_attribute and name is not None:
            self._attrs[name] = 1

    def identifyObject(self, value):
        kos = self.getKeyedObjectSystem()
        return kos.identifyObject(value)

    def notifySerializedRef(self, name, value, is_attribute, keychain):
        assert keychain is not None
        self._refs.append((keychain, value))
        self.notifySerialized(name, value, is_attribute)

    def ignoreAttribute(self, name):
        """See the ISerializationEvent interface."""
        self._attrs[name] = 1

    def ignoreAttributes(self, names):
        """See the ISerializationEvent interface."""
        for name in names:
            self._attrs[name] = 1


    # IFullSerializationEvent interface methods:

    def getSerializedAttributeNames(self):
        """Returns the name of all attributes serialized."""
        return self._attrs.keys()

    def getInternalRef(self, ob):
        """Returns (serializer_name, name) or None."""
        return self._internal_refs.get(id(ob))


