##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Standard gateway classes.

$Id: gateways.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

import time

from interfaces import IGateway
import exceptions


class CompositeGateway:
    """Gateway that delegates to multiple smaller gateways."""

    __implements__ = IGateway

    def __init__(self, base=None):
        self._gws = {}
        if base is not None:
            self._gws.update(base._gws)

    def addGateway(self, name, gw, force=0):
        if not force and self._gws.has_key(name):
            raise KeyError, "Gateway name %s in use" % name
        self._gws[name] = gw

    def removeGateway(self, name):
        del self._gws[name]  # raise KeyError if not in use

    def hasGateway(self, name):
        return self._gws.has_key(name)

    def getSchema(self):
        """Returns the ISchema of data stored by this gateway.

        See interfaces.ISchema.
        """
        res = {}
        for name, gw in self._gws.items():
            s = gw.getSchema()
            if s is not None:
                res[name] = s
        return res

    def load(self, event):
        """Loads data.

        Returns a pair containing the data and an object
        that acts as a serial number or a hash of the data.
        The serial number is either a time stamp or some other object
        that can be consistently compared to detect conflicts.
        """
        full_state = {}
        serials = {}
        for name, gw in self._gws.items():
            state, serial = gw.load(event)
            if state is not None:
                full_state[name] = state
                if serial is not None:
                    serials[name] = serial
        serials = serials.items()
        serials.sort()
        return full_state, tuple(serials)

    def store(self, event, full_state):
        """Stores data.

        Returns a new serial.
        """
        serials = {}
        for name, gw in self._gws.items():
            state = full_state.get(name)
            # print 'gateway storing', keychain, name, state
            serial = gw.store(event, state)
            if serial is not None:
                serials[name] = serial
        serials = serials.items()
        serials.sort()
        return tuple(serials)

    def getSources(self, event):
        """Returns data source information.  See IGateway.
        """
        res = {}
        for gw in self._gws.values():
            sources = gw.getSources(event)
            if sources is not None:
                res.update(sources)
        return res


class MappingGateway:
    """Gateway to a simple dictionary (primarily for testing).
    """

    __implements__ = IGateway

    def __init__(self, schema):
        self.schema = schema
        self.data = {}

    def getSchema(self):
        return self.schema

    def load(self, event):
        # Returns (data, serial)
        keychain = event.getKeychain()
        try:
            return self.data[keychain]
        except KeyError:
            raise exceptions.NoStateFoundError(keychain)

    def store(self, event, data):
        h = time.time()
        self.data[event.getKeychain()] = (data, h)
        return h

    def getSources(self, event):
        return None

