##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Public interfaces

$Id: interfaces.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

from Interface import Interface, Attribute


class ISchema(Interface):
    """Data schema"""

    def __eq__(other):
        """Checks equality with another schema.

        This is the only method common to all schema objects.  The
        idea is that serializers and data gateways need only agree on
        schemas, and this method allows you to check compatibility.
        """


class IRelationalSchema(ISchema):
    """Schema oriented for relational storage.

    Designed to assist integration with relational databases."""

    def getColumnDefs():
        """Returns a sequence of (column_name, column_type, unique_flag)"""



class IClassFactory(Interface):
    """Class finder."""

    def getClass(module_name, class_name):
        """Returns the named class.

        A default implementation may use the Python's standard import
        mechanism.
        """


class IKeyedObjectSystem (IClassFactory):
    """A collection of objects identifiable by keychain.

    In apelib.zodb3, the ZODB Connection object (the _p_jar) is
    an IKeyedObjectSystem.
    """

    def getObject(keychain, hints=None):
        """Returns a class instance, possibly ghosted.

        Used during deserialization (loading/import).
        The hints argument, a mapping, may be provided as an
        optimization.  Without it, implementations of this method may
        have to load a full object rather than a ghosted object.

        Some hints may be:

        classification
        mapper_name (the mapper name for the last item in the keychain)
        mapper_names (a list of all mapper names)
        """

    def identifyObject(obj):
        """Returns the keychain of an object.

        Used during serialization (storing/export).
        Returns None if the object is not in the keyed object system.
        """

    def newKey():
        """Returns a new, unique key (which might be used in a keychain).

        Used during serialization (storing/export).
        """

    # Deprecated
    def loadStub(keychain, hints=None):
        """Deprecated alias for getObject()."""




class IDatabaseInitializer (Interface):
    """Provides a way to initialize a database."""

    def init(event):
        """Initializes the database, creating tables etc.

        event is an IDatabaseInitEvent.
        """


class IDatabaseInitEvent (Interface):
    """Interface for events involved in initializing databases."""

    def getConnection(name):
        """Returns the named connection."""

    def clearing():
        """Returns true if the database is to be cleared.

        This method is designed for testing purposes.
        """


class IMapperEvent (Interface):
    """The base interface for events occurring in context of a mapper."""

    def getMapper():
        """Returns the mapper for the object being (de)serialized."""

    def getKeychain():
        """Returns the keychain of the object being (de)serialized."""

    def getKey():
        """Returns the last element of the keychain.

        For most purposes, the key can be used as an OID.
        """

    def makeKeychain(name, stored):
        """Generates a keychain for a subobject.
        """


class IGatewayEvent (IMapperEvent):
    """Interface for events used by gateways."""

    def getConnection(name):
        """Returns the named connection."""

    def getClassification():
        """Returns the innermost classification (for reading only) or None.

        The innermost classification is the classification of the last
        key in the keychain.  During traversal of the first element of
        a keychain, no classification has been determined yet, so this
        method returns None.
        """


class ILoadEvent (IGatewayEvent):
    """Interface for events involved in loading objects."""


class IStoreEvent (IGatewayEvent):
    """Interface for events involved in storing objects."""


class ISDEvent (IMapperEvent):
    """Base for serialization and deserialization events."""

    def getKeyedObjectSystem():
        """Returns the IKeyedObjectSystem involved in the event."""

    def getObject():
        """Returns the object being (de)serialized."""

    def setSerializerName(name):
        """Sets the name of the next serializer to be used."""

    def getSerializerName():
        """Returns the name of the serializer in use."""

    def addUnmanagedPersistentObjects(obs):
        """Notifies that there are unmanaged persistent objects in the object.

        If no attention is paid to unmanaged persistent objects
        (UPOs), they will not notify ZODB when they are changed, and
        hence can be a challenge for the application programmer.  Use
        this method to tell ZODB about the UPOs so that ZODB will see
        changes made to them and save the corresponding managed
        persistent object.
        """

    def getUnmanagedPersistentObjects():
        """Returns the list of unmanaged persistent objects."""

    def getExternalRefs():
        """Returns the list of external references.

        The list is built up during (de)serialization.
        The returned list is of the form [(keychain, subobject)].
        """


class IDeserializationEvent(ISDEvent):
    """A helper in the object deserialization process.

    Implementations of ISerializer.deserialize() call
    methods of this interface to restore internal and external
    references.
    """

    def notifyDeserialized(name, value):
        """Indicates that a named internal subobject was deserialized.
        """

    def dereference(name, keychain, mapper_names=None):
        """Retrieves a referenced subobject (usually ghosted initially).
        """


class IFullDeserializationEvent(IDeserializationEvent):
    """Deserialization event with features for deserializing remainder data.
    """

    def loadInternalRef(ref):
        """Returns an object already deserialized by another serializer.

        'ref' is a tuple containing (serializer_name, name).
        """


class ISerializationEvent(ISDEvent):
    """A helper in the object serialization process.

    Implementations of ISerializer.serialize() call
    methods of this interface to create internal and external
    references.
    """

    def notifySerialized(name, value, is_attribute):
        """Indicates that a named internal subobject was serialized.

        This allows a 'remainder pickler' to refer to subobjects by name.
        Be careful to unwrap acquisition and/or context wrappers around
        subob before calling this method.
        """

    def identifyObject(value):
        """Returns the keychain of an existing object.

        Returns None if the object is not yet known.
        """

    def notifySerializedRef(name, value, is_attribute, keychain):
        """Indicates that a reference to an external subobject was serialized.

        TODO: write more here. ;-)
        """

    def ignoreAttribute(name):
        """Indicates that an attribute should be ignored when storing."""

    def ignoreAttributes(names):
        """Indicates that several attributes should be ignored."""


class IFullSerializationEvent(ISerializationEvent):
    """Serialization event with features for ensuring complete serialization.

    Used for generating a 'remainder pickle'.
    """

    def getSerializedAttributeNames():
        """Returns the name of all attributes serialized."""

    def getInternalRef(ob):
        """Returns (serializer_name, name) or None."""


class ISerializer(Interface):
    """Object serializer / deserializer"""

    def getSchema():
        """Returns the schema of records (de)serialized by this component.
        """

    def canSerialize(obj):
        """Returns true if this serializer can serialize the given object.
        """

    def serialize(obj, event):
        """Returns the state of this part of the object.

        Use the ISerializationEvent to set up internal and external
        references.
        """

    def deserialize(obj, event, state):
        """Fills in the state of this part of the object.

        Use the IDeserializationEvent to resolve external references.
        Returns nothing.
        """


class IFullObjectSerializer(ISerializer):
    """Serializes/deserializes the complete state of objects.

    The serialized state does not need to include the class of the object,
    which is maintained separately.

    IFullObjectSerializers usually delegate to multiple ISerializers
    to do the actual work of (de)serialization.  The schema of
    IFullObjectSerializers is usually a dictionary containing the name
    and schema of its constituent ISerializers.
    """

    def createEmptyInstance(class_factory, classification=None):
        """Returns a new instance.

        If this serializer works with instances of only one class,
        createEmptyInstance() should not require the use of the
        classification argument.  Implementations that need the
        classification argument can return None when classification is
        None, but doing so may incur a performance penalty.

        class_factory is an IClassFactory.  The serializer may use it
        instead of the standard Python import mechanism to find a
        class.
        """


class IGateway (Interface):
    """Loads and stores data by keychain.

    Implementations can store in entire tables, pieces of tables, translate
    for storage in joined tables, or store in some entirely different way.

    Based on _Patterns of Enterprise Application Architecture_
    by Martin Fowler.
    """

    def getSchema():
        """Returns the ISchema of data stored by this gateway.

        See serial.interfaces.ISchema.
        """

    def load(event):
        """Loads data.

        event is an ILoadEvent.

        Returns a pair containing the data and a hash of the data.
        The hash value is either an integer or an object that is
        hashable using the Python hash() function.  The hashable
        object is used to detect storage conflicts.
        """

    def store(event, data):
        """Stores data.

        event is an IStoreEvent.

        Returns a new hash value.
        """

    def getSources(event):
        """Returns source information for a keychain.

        The source information allows the system to poll for changes
        to keep caches in sync with the data.  Where polling is not
        necessary, gateways are free to return None.

        The source information is a dictionary in the format:
        {(repository, source): state}.  The repository must be an
        ISourceRepository.  The source and state must be in a form
        recognized by the repository.  Both the repository and source
        must be hashable.
        """


class IClassifier(Interface):
    """Object classifier

    Implementations of this interface are a little like biologists.
    During serialization, the classifyObject() method returns a
    mapping containing the classification of subob (like a biologist
    identifying a creature's genus and species).  During
    deserialization, the classifyState() method decides what kind of
    objects to create for a stored state (like a biologist showing you
    a creature of a certain genus and species).

    The keys in classifications are implementation-dependent.
    """

    def classifyObject(value, keychain):
        """Returns a classification and mapper_name.
        """

    def classifyState(event):
        """Returns a classification and mapper_name.

        event is an ILoadEvent.

        May load the classification from storage.
        """

    def store(event, classification):
        """Stores the classification of an object.

        event is an IStoreEvent.
        """


class IConfigurableClassifier (IClassifier):
    """Classifier that accepts registrations.
    """

    def register(condition, value, mapper_name):
        """Registers a sub-mapper to be used when a condition is met.

        The kinds of conditions available depend on the classifier,
        but the following conditions are expected to be common:

        'class'     - matches a fully-qualified class name
        'extension' - matches a filename extension
        'key'       - matched the end of the keychain
        'fallback'  - matches when no other condition is met.  The
                      fallback values depend on the classifier.
        """


class IKeychainGenerator (Interface):
    """A utility for generating keychains.

    Many serializers encode references to other objects, but it is
    desirable to not hard code the format of keychains into
    ISerializer implementations.  This utility, normally accessed
    through an IMapperEvent, allows keychain knowledge to be kept in
    one place.
    """

    def makeKeychain(event, name, stored):
        """Returns a new keychain.

        event is an IMapperEvent.  Additionally, it is always either
        an ISDEvent or an IGatewayEvent.

        name is the name of the subobject.

        stored is a flag indicating whether the returned keychain will
        be stored.  If it will not be stored, the generated keychain
        can not be a random label.
        """


class IMapper (Interface):
    """A hub for mapping a certain kind of object.
    """

    def getSerializer():
        """Returns the IObjectSerializer for this mapper."""

    def getGateway():
        """Returns the IGateway for this mapper."""

    def getClassifier():
        """Returns the IClassifier for objects referenced by this mapper.

        If this mapper references no other objects, it's safe to
        return None.
        """

    def getSubMapper(name):
        """Returns the named sub-IMapper.

        The name of a sub-mapper is chosen by either a classifier or
        a serializer.
        """

    def listSubMapperNames():
        """Returns the name of all sub-IMappers."""

    def getKeychainGenerator():
        """Returns the IKeychainGenerator for subobjects.

        Some serializers and gateways make use of this.  If no serializer
        or gateway needs this, it's safe to return None.
        """

    def getInitializers():
        """Returns the list of configured IDatabaseInitializers."""


class IConfigurableMapper (IMapper):
    """Adds operations to IMapper for configuration.
    """

    def setSerializer(s):
        """Sets the IFullObjectSerializer for this mapper."""

    def setGateway(g):
        """Returns the IGateway for this mapper."""

    def setClassifier(c):
        """Sets the IClassifier for this mapper."""

    def setKeychainGenerator(k):
        """Sets the IKeychainGenerator for subobjects."""

    def addSubMapper(name, m=None, replace=0):
        """Adds a named sub-IMapper, returning the mapper added.

        If m is None, a default IConfigurableMapper
        implementation is created.  If replace is not set,
        the implemenation prevents overriding an existing mapper.
        """

    def addInitializer(obj):
        """Adds a database initializer.

        Database initializers get notified of new connections,
        allowing them to prepare the database.
        """

    def checkConfiguration(path='root', recursive=1):
        """Verifies the mapper configuration is sane.

        Raises an exception if there are errors.

        'path' gives the path to the mapper, for debugging purposes.
        'recursive' can be turned off to not descend into sub-mappers.
        """


class ITPCConnection(Interface):

    def connect():
        """Opens any resources needed for transactions.  Called only once."""

    def sortKey():
        """Returns a sort key for consistent ordering."""

    def getName():
        """Returns a human-readable name."""

    def begin():
        """Called before the first phase of two-phase commit."""

    def vote():
        """Called upon transition to the second phase of two-phase commit."""

    def abort():
        """Aborts the transaction."""

    def finish():
        """Commits the transaction."""

    def close():
        """Closes resources.  Called only once."""


class ISourceRepository(Interface):
    """Represents a collection of object sources.

    Designed to helps keep a cache in sync with its sources.
    """

    def freshen(sources):
        """Returns changed source information.

        The source information is a mapping that maps
        (source_repository, path) to a state object.  The
        source_repository will always be this object (the redundancy
        keeps things simpler for scanners).  This method returns a
        mapping containing only the items of the input dictionary
        whose state has changed.
        """

