##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Ape I/O facades.

These facades implement commonly useful high-level mapper operations.

$Id: io.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

from weakref import proxy

from events \
     import DatabaseInitEvent, GatewayEvent, LoadEvent, StoreEvent, \
     SerializationEvent, DeserializationEvent
from interfaces import ITPCConnection, IKeyedObjectSystem


class ClassifiedState:
    """Object state with classification information."""

    def __init__(self, state, classification, mapper_names):
        self.state = state
        self.classification = classification
        self.mapper_names = mapper_names


class GatewayIO:
    """Gateway operations facade."""

    def __init__(self, root_mapper, connections):
        self._root_mapper = root_mapper
        self._conn_map = connections
        # Sort the connections by sort key.  Use an extra index to avoid
        # using connections as sort keys.
        items = []  # [(sort_key, index, conn)]
        index = 0
        for c in connections.values():
            assert ITPCConnection.isImplementedBy(c)
            sort_key = c.sortKey()
            items.append((sort_key, index, c))
            index += 1
        items.sort()
        conn_list = []
        for sort_key, index, c in items:
            conn_list.append(c)
        self._conn_list = conn_list

    def openConnections(self):
        try:
            opened = []
            for c in self._conn_list:
                c.connect()
                opened.append(c)
        except:
            for c in opened:
                c.close()
            raise

    def closeConnections(self):
        for conn in self._conn_list:
            conn.close()

    def getConnectionList(self):
        return self._conn_list

    def getConnectionMap(self):
        return self._conn_map


    def initDatabases(self, clear_all=0):
        """Creates tables, etc.
        """
        # Find all initializers, eliminating duplicates.
        initializers = {}  # obj -> 1
        todo = [self._root_mapper]
        while todo:
            mapper = todo.pop()
            for obj in mapper.getInitializers():
                initializers[obj] = 1
            sub = mapper.listSubMapperNames()
            if sub:
                for name in sub:
                    m = mapper.getSubMapper(name)
                    todo.append(m)
        # Now call them.
        event = DatabaseInitEvent(self._conn_map, clear_all)
        for initializer in initializers.keys():
            initializer.init(event)


    def classifyState(self, keychain):
        mapper = self._root_mapper
        mapper_names = []
        # Follow the keychain to find the right mapper.
        classification = None
        for i in range(len(keychain)):
            k = keychain[:i + 1]
            cfr = mapper.getClassifier()
            assert cfr is not None, keychain
            event = LoadEvent(mapper, k, self._conn_map, classification)
            classification, sub_mapper_name = cfr.classifyState(event)
            mapper_names.append(sub_mapper_name)
            mapper = mapper.getSubMapper(sub_mapper_name)
        return classification, mapper_names, mapper


    def load(self, keychain):
        classification, mapper_names, mapper = self.classifyState(keychain)
        event = LoadEvent(mapper, keychain, self._conn_map, classification)
        state, hash_value = mapper.getGateway().load(event)
        cs = ClassifiedState(state, classification, mapper_names)
        return event, cs, hash_value


    def store(self, keychain, classified_state):
        classification = classified_state.classification
        mapper_names = classified_state.mapper_names
        assert len(keychain) == len(mapper_names)
        mapper = self._root_mapper
        prev_mapper = mapper
        for mapper_name in mapper_names:
            prev_mapper = mapper
            mapper = mapper.getSubMapper(mapper_name)
        cfr = prev_mapper.getClassifier()
        event = StoreEvent(mapper, keychain, self._conn_map, classification)
        new_hash = mapper.getGateway().store(event, classified_state.state)
        if cfr is not None:
            cfr.store(event, classification)
        return event, new_hash


    def getSources(self, keychain):
        classification, mapper_names, mapper = self.classifyState(keychain)
        event = LoadEvent(mapper, keychain, self._conn_map, classification)
        return mapper.getGateway().getSources(event)


    def newKeychain(self):
        # Try to use the root keychain generator to make a keychain.
        kgen = self._root_mapper.getKeychainGenerator()
        event = GatewayEvent(self._root_mapper, (), self._conn_map, None)
        return kgen.makeKeychain(event, None, 1)


class ObjectSystemIO:
    """Object system (de)serialization facade."""

    def __init__(self, root_mapper, kos):
        self._root_mapper = root_mapper
        self._kos = kos


    def classifyObject(self, obj, keychain):
        mapper = self._root_mapper
        mapper_names = []
        classification = None
        if keychain:
            # Classify the parents first to discover what mapper to
            # use for storage.
            for i in range(1, len(keychain)):
                k = keychain[:i]
                o = self._kos.getObject(k)
                cfr = mapper.getClassifier()
                classification, sub_mapper_name = cfr.classifyObject(o, k)
                mapper_names.append(sub_mapper_name)
                mapper = mapper.getSubMapper(sub_mapper_name)
            # Now classify the object being stored.
            cfr = mapper.getClassifier()
            classification, sub_mapper_name = cfr.classifyObject(obj, keychain)
            mapper_names.append(sub_mapper_name)
            mapper = mapper.getSubMapper(sub_mapper_name)
        return classification, mapper_names, mapper


    def serialize(self, keychain, obj):
        classification, mapper_names, mapper = self.classifyObject(
            obj, keychain)
        # Now serialize.
        ser = mapper.getSerializer()
        event = SerializationEvent(self._kos, mapper, keychain, obj)
        state = ser.serialize(obj, event)
        cs = ClassifiedState(state, classification, mapper_names)
        return event, cs


    def deserialize(self, keychain, obj, classified_state):
        mapper = self._root_mapper
        assert len(keychain) == len(classified_state.mapper_names)
        for mapper_name in classified_state.mapper_names:
            mapper = mapper.getSubMapper(mapper_name)
        ser = mapper.getSerializer()
        event = DeserializationEvent(self._kos, mapper, keychain, obj)
        ser.deserialize(obj, event, classified_state.state)
        return event


    def newObject(self, classified_state):
        mapper = self._root_mapper
        for mapper_name in classified_state.mapper_names:
            mapper = mapper.getSubMapper(mapper_name)
        ser = mapper.getSerializer()
        return ser.createEmptyInstance(
            self._kos, classification=classified_state.classification)



class ExportImport:
    """Simple import/export facade.
    """

    __implements__ = IKeyedObjectSystem

    def __init__(self, root_mapper, connections, class_factory=None):
        self._objects = {}     # { keychain -> obj }
        self._keychains = {}   # { id(obj) -> keychain }
        # _incomplete contains the keychains of objects not yet
        # imported fully.
        self._incomplete = {}   # { keychain -> 1 }
        self._class_factory = class_factory
        # Avoid a circular reference by making a weakref proxy
        self.obj_io = ObjectSystemIO(root_mapper, proxy(self))
        self.gw_io = GatewayIO(root_mapper, connections)


    def _register(self, keychain, obj):
        """Registers obj in the temporary object index.

        Returns true if the object was added to the index for the first
        time.  If the registration conflicts, raises an exception.
        """
        is_new = 0
        if self._objects.has_key(keychain):
            if self._objects[keychain] is not obj:
                raise ValueError, (
                    "Multiple objects for keychain %s" % repr(keychain))
        else:
            self._objects[keychain] = obj
            is_new = 1
        obj_id = id(obj)
        if self._keychains.has_key(obj_id):
            if self._keychains[obj_id] != keychain:
                raise ValueError, (
                    "Multiple keychains for object %s" % repr(obj))
        else:
            self._keychains[obj_id] = keychain
            is_new = 1
        return is_new


    def exportObject(self, src_obj, dest_keychain=None, deactivate_func=None):
        count = 0
        if dest_keychain is None:
            dest_keychain = (self.newKey(),)
        self._register(dest_keychain, src_obj)
        todo = [(dest_keychain, src_obj)]
        while todo:
            keychain, obj = todo.pop()
            event, classified_state = self.obj_io.serialize(keychain, obj)
            count += 1
            if deactivate_func is not None:
                deactivate_func(obj, count)
            self.gw_io.store(keychain, classified_state)
            ext_refs = event.getExternalRefs()
            if ext_refs:
                for ext_keychain, ext_obj in ext_refs:
                    if self._register(ext_keychain, ext_obj):
                        todo.append((ext_keychain, ext_obj))


    def importObject(self, src_keychain, dest_obj=None, commit_func=None):
        count = 0
        if dest_obj is None:
            dest_obj = self.getObject(src_keychain)
        root_obj = dest_obj
        self._register(src_keychain, dest_obj)
        todo = [(src_keychain, dest_obj)]
        while todo:
            keychain, obj = todo.pop()
            e, classified_state, hash_value = self.gw_io.load(keychain)
            event = self.obj_io.deserialize(keychain, obj, classified_state)
            if self._incomplete.has_key(keychain):
                del self._incomplete[keychain]
            count += 1
            if commit_func is not None:
                commit_func(obj, count)
            ext_refs = event.getExternalRefs()
            if ext_refs:
                for ext_keychain, ext_obj in ext_refs:
                    if (self._register(ext_keychain, ext_obj)
                        or self._incomplete.has_key(ext_keychain)):
                        todo.append((ext_keychain, ext_obj))
        return root_obj


    # IKeyedObjectSystem implementation

    def getClass(self, module, name):
        # Normally called only while importing
        if self._class_factory is not None:
            return self._class_factory.getClass(module, name)
        else:
            m = __import__(module, {}, {}, ('__doc__',))
            return getattr(m, name)

    def getObject(self, keychain, hints=None):
        # Should be called only while importing
        try:
            return self._objects[keychain]
        except KeyError:
            # This object has not been loaded yet.  Make a stub.
            e, classified_state, hash_value = self.gw_io.load(keychain)
            obj = self.obj_io.newObject(classified_state)
            # Don't fill in the state yet, to avoid infinite
            # recursion.  Just register it.
            self._incomplete[keychain] = 1
            self._register(keychain, obj)
            return obj

    loadStub = getObject

    def identifyObject(self, obj):
        # Normally called only while exporting
        return self._keychains.get(id(obj))

    def newKey(self):
        # Should be called only while exporting
        return self.gw_io.newKeychain()[-1]

