##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Standard mapper class.

$Id: mapper.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

from types import DictType

import interfaces
from exceptions import ConfigurationError


class Mapper:
    """Standard mapper class."""

    __implements__ = interfaces.IConfigurableMapper

    def __init__(self,
                 parent=None,
                 serializer=None,
                 gateway=None,
                 classifier=None,
                 kgen=None):
        self._sub_mappers = {}
        self._parent = parent
        self._serializer = serializer
        self._gateway = gateway
        self._classifier = classifier
        self._kgen = kgen
        self._initializers = []

    # IConfigurableMapper implementation

    def setParent(self, p):
        self._parent = p

    def setSerializer(self, s):
        self._serializer = s

    def setGateway(self, g):
        self._gateway = g

    def setClassifier(self, c):
        self._classifier = c

    def setKeychainGenerator(self, k):
        self._kgen = k

    def addSubMapper(self, name, m=None, replace=0):
        if not replace and self._sub_mappers.has_key(name):
            raise KeyError('mapper name %s already in use' % name)
        if m is None:
            m = Mapper(self)
        self._sub_mappers[name] = m
        return m

    def addInitializer(self, obj):
        self._initializers.append(obj)

    def checkConfiguration(self, path='root', recursive=1):
        s = self._serializer
        if s is None:
            raise ConfigurationError(
                'No serializer configured for mapper %s' % repr(path))
        if not interfaces.IFullObjectSerializer.isImplementedBy(s):
            raise ConfigurationError(
                'Not an IFullObjectSerializer: %s' % repr(s))
        g = self._gateway
        if g is None:
            raise ConfigurationError(
                'No gateway configured for mapper %s' % repr(path))
        if not interfaces.IGateway.isImplementedBy(g):
            raise ConfigurationError(
                'Not an IGateway: %s' % repr(g))
        if s.getSchema() != g.getSchema():
            # Try to show a descriptive error
            ss = s.getSchema()
            gs = g.getSchema()
            msg = None
            if isinstance(ss, DictType) and isinstance(gs, DictType):
                for key in ss.keys():
                    if not gs.has_key(key):
                        msg = 'No gateway provided for serializer "%s"' % key
                        break
                    elif ss[key] != gs[key]:
                        msg = 'Mismatch on name "%s". %s != %s' % (
                            key, ss[key], gs[key])
                        break
                if msg is None:
                    for key in gs.keys():
                        if not ss.has_key(key):
                            msg = ('No serializer provided for gateway "%s"'
                                   % key)
                            break
            if msg is None:
                msg = '%s != %s' % (ss, gs)
            raise ConfigurationError(
                'Mismatched schemas in mapper "%s": %s' % (path, msg))
        if self._parent is None:
            if self._classifier is None:
                raise ConfigurationError('No root classifier configured')
            if self._kgen is None:
                raise ConfigurationError(
                    'No root keychain generator configured')
        else:
            if not interfaces.IMapper.isImplementedBy(self._parent):
                raise ConfigurationError(
                    'Not an IMapper: %s' % repr(self._parent))
        if (self._classifier is not None
            and not interfaces.IClassifier.isImplementedBy(self._classifier)):
            raise ConfigurationError(
                'Not an IClassifier: %s' % repr(self._classifier))
        if (self._kgen is not None
            and not interfaces.IKeychainGenerator.isImplementedBy(self._kgen)):
            raise ConfigurationError(
                'Not an IKeychainGenerator: %s' % repr(self._kgen))

        if recursive:
            for n, m in self._sub_mappers.items():
                if not interfaces.IMapper.isImplementedBy(m):
                    raise ConfigurationError(
                        'Not an IMapper: %s' % repr(m))
                m.checkConfiguration(('%s/%s' % (path, n)), recursive)

    # IMapper implementation

    def getSerializer(self):
        return self._serializer

    def getGateway(self):
        return self._gateway

    def getSubMapper(self, name):
        return self._sub_mappers[name]

    def listSubMapperNames(self):
        return self._sub_mappers.keys()

    def getClassifier(self):
        if self._classifier is not None:
            return self._classifier
        if self._parent is not None:
            return self._parent.getClassifier()
        return None

    def getKeychainGenerator(self):
        if self._kgen is not None:
            return self._kgen
        if self._parent is not None:
            return self._parent.getKeychainGenerator()
        return None

    def getInitializers(self):
        return self._initializers

