##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Standard serializers.

$Id: serializers.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

from types import StringType

from interfaces import ISerializer, IFullObjectSerializer
from exceptions import DeserializationError, SerializationError
from schemas import FieldSchema


class CompositeSerializer:
    """Full serializer based on partial serializers.
    """

    __implements__ = IFullObjectSerializer

    def __init__(self, module, name, base=None):
        self._module = module
        self._name = name
        self.init(base)

    def init(self, base=None):
        self._part_names = {}  # { name -> 1 }
        self._parts = []       # [(name, serializer)] -- Order matters.
        self._final_parts = [] # [(name, serializer)]
        if base is not None:
            self._part_names.update(base._part_names)
            self._parts[:] = base._parts
            self._final_parts[:] = base._final_parts

    def addSerializer(self, name, serializer, force=0, final=0):
        if self._part_names.has_key(name):
            if not force:
                raise KeyError, "Serializer name %s in use" % repr(name)
            self.removeSerializer(name)
        if final:
            self._final_parts.append((name, serializer))
        else:
            self._parts.append((name, serializer))
        self._part_names[name] = 1

    def removeSerializer(self, name):
        if not self._part_names.has_key(name):
            raise KeyError, "Serializer name %s not in use" % repr(name)
        for lst in (self._parts, self._final_parts):
            for i in range(len(lst)):
                if lst[i][0] == name:
                    del lst[i]
                    break
        del self._part_names[name]

    def hasSerializer(self, name):
        return self._part_names.has_key(name)

    def getSerializers(self):
        return self._parts + self._final_parts

    def getSchema(self):
        res = {}
        for name, serializer in self.getSerializers():
            s = serializer.getSchema()
            if s is not None:
                res[name] = s
        return res

    def canSerialize(self, object):
        if not hasattr(object, '__class__'):
            return 0
        c = object.__class__
        return (c.__module__ == self._module and c.__name__ == self._name)

    def serialize(self, object, event):
        full_state = {}
        for name, serializer in self.getSerializers():
            event.setSerializerName(name)
            state = serializer.serialize(object, event)
            if state is not None:
                full_state[name] = state
        return full_state

    def deserialize(self, object, event, full_state):
        for name, serializer in self.getSerializers():
            state = full_state.get(name)
            event.setSerializerName(name)
            serializer.deserialize(object, event, state)

    def createEmptyInstance(self, class_factory, classification=None):
        c = class_factory.getClass(self._module, self._name)
        return c.__basicnew__()


class AnyObjectSerializer (CompositeSerializer):
    """Full serializer that's not tied to a specific class
    """

    __implements__ = IFullObjectSerializer

    def __init__(self, base=None):
        self.init(base)

    def canSerialize(self, object):
        return 1

    def createEmptyInstance(self, class_factory, classification=None):
        if classification is None:
            # This serializer can't do anything without the classification.
            return None
        cn = classification['class_name']
        pos = cn.rfind('.')
        if pos < 0:
            raise ValueError, "class_name must include the module"
        module = cn[:pos]
        name = cn[pos + 1:]
        c = class_factory.getClass(module, name)
        return c.__basicnew__()


class FullState:
    """Serializer that reads/writes the entire state of an object."""

    __implements__ = ISerializer

    schema = FieldSchema('data', 'object')

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return 1

    def serialize(self, object, event):
        return object.__getstate__()

    def deserialize(self, object, event, state):
        object.__setstate__(state)



class IgnoredAttribute:
    """Serializer that explicitly ignores an attribute
    """
    __implements__ = ISerializer

    def __init__(self, attrname):
        self.attrname = attrname

    def getSchema(self):
        return None  # No storage

    def canSerialize(self, object):
        return 1

    def serialize(self, object, event):
        event.ignoreAttribute(self.attrname)
        return None

    def deserialize(self, object, event, state):
        assert state is None


class OptionalSerializer:
    """Serializer wrapper that serializes only if the object is compatible.
    """

    __implements__ = ISerializer

    def __init__(self, real, default_state=None):
        self._real = real
        self._default_state = default_state

    def getSchema(self):
        return self._real.getSchema()

    def canSerialize(self, object):
        return 1

    def serialize(self, object, event):
        if self._real.canSerialize(object):
            return self._real.serialize(object, event)
        else:
            return self._default_state

    def deserialize(self, object, event, state):
        if self._real.canSerialize(object):
            self._real.deserialize(object, event, state)
        else:
            if state is not None and state != self._default_state:
                raise DeserializationError(
                    "Optional serializer unable to install state %s into %s" %
                    (repr(state), repr(object)))


class StringDataAttribute:
    """Serializer of a simple string data attribute."""

    __implements__ = ISerializer

    schema = FieldSchema('data', 'string')

    def __init__(self, attrname):
        self.attrname = attrname

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return 1

    def serialize(self, object, event):
        attrname = self.attrname
        assert attrname
        v = getattr(object, attrname)
        assert isinstance(v, StringType)
        event.notifySerialized(attrname, v, 1)
        return v

    def deserialize(self, object, event, state):
        attrname = self.attrname
        assert attrname
        assert isinstance(state, StringType)
        setattr(object, attrname, state)
        event.notifyDeserialized(attrname, state)

