##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Full-featured Filesystem connection class.

$Id: connection.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

import re
from types import StringType

from apelib.core.interfaces import ITPCConnection, ISourceRepository
from apelib.core.exceptions import NoStateFoundError
from interfaces import IFSConnection
from exceptions import FSWriteError
from cache import ShortLivedCache
from fileops import StandardFileOperations


# Try to decipher this regular expression ;-)
# It basically matches "\n[sectionname]...\n", where len(sectionname) > 0.
section_re = re.compile(r'^\[([^\[\]\n]+)\][^\r\n]*(?:\r\n|\r|\n)',
                      re.MULTILINE)

# For a NODE_TYPE_SECTION, the value is 'f' (file) or 'd' (directory)
NODE_TYPE_SECTION = '@node_type'

# For a DATA_SECTION, the value is a two-item tuple containing a
# string (file) or list of names (directory) and the as_text flag.
DATA_SECTION = '@data'

SUGGESTED_EXTENSION_SECTION = '@s_ext'  # The suggested filename extension.
OBJECT_NAMES_SECTION = 'object_names'   # For directories.  The value is text.
REMAINDER_SECTION = 'remainder'         # The value is a binary string.

PROPERTIES_EXTENSION = 'properties'
REMAINDER_EXTENSION = 'remainder'

# Match 'foo.properties', 'foo.remainder', 'properties', or 'remainder'.
# This is for filtering out metadata filenames.
metadata_re = re.compile('(|.+[.])(%s|%s)$' % (
    PROPERTIES_EXTENSION, REMAINDER_EXTENSION))


class FSConnection:
    """Reads / writes files with 'sections'.

    The required 'type' section specifies whether the object is a file or
    a directory.  The optional 'data' section specifies either the main
    file contents or the names of the files in the directory.  All other
    sections get stored in a '.properties' file.  The properties file uses
    square-bracket section headers and encodes sections by doubling
    left-square brackets.
    """
    __implements__ = IFSConnection, ITPCConnection, ISourceRepository

    basepath = ''

    def __init__(self, basepath, metadata_prefix='.', hidden_filenames='_',
                 ops=None):
        self.basepath = basepath
        self.metadata_prefix = metadata_prefix
        self.hidden_re = re.compile(hidden_filenames)
        self._final = 0
        # _pending holds the data to be written.
        # _pending: { subpath string -> { section_name -> data } }
        self._pending = {}
        self._props_cache = ShortLivedCache()
        self._dir_cache = ShortLivedCache()
        if ops is None:
            ops = StandardFileOperations()
        self.ops = ops


    def _isLegalFilename(self, fn):
        mp = self.metadata_prefix
        if (not fn or
            (fn.startswith(mp) and metadata_re.match(fn, len(mp)))
            or self.hidden_re.match(fn) is not None):
            return 0
        return 1
        

    def _computeDirectoryContents(self, path, ignore_error=0):
        """Computes and returns intermediate directory contents info.

        Returns (filenames, object_names, translations).  The results
        are cached for a short time.
        """
        res = self._dir_cache.get(path)
        if res is not None:
            return res

        obj_names = []
        trans = {}     # { base name -> filename with extension or None }
        try:
            fns = self.ops.listdir(path)
        except OSError:
            if ignore_error:
                return ([], obj_names, trans)

        filenames = filter(self._isLegalFilename, fns)
        props = self._getPropertiesFromFile(path)
        text = props.get(OBJECT_NAMES_SECTION)
        if text:
            # Prepare a dictionary of translations.
            for fn in filenames:
                if '.' in fn:
                    base, ext = fn.split('.', 1)
                    if trans.has_key(base):
                        # Name collision: two or more files have the same base
                        # name.  Don't use an extension for this name.
                        trans[base] = None
                    else:
                        trans[base] = fn
                else:
                    trans[fn] = None
            obj_names = [line.strip() for line in text.split('\n')]
            for obj_name in obj_names:
                if '.' in obj_name:
                    base, ext = obj_name.split('.', 1)
                    trans[base] = None

        res = (filenames, obj_names, trans)
        self._dir_cache.set(path, res)
        return res


    def _listDirectoryAsMapping(self, path, ignore_error=0):
        """Returns the translated filenames at path.

        The ignore_error flag makes this method return an empty
        dictionary if the directory is not found.

        Returns {filename -> obj_name}.
        """
        filenames, obj_names, trans = self._computeDirectoryContents(
            path, ignore_error)
        res = {}
        for fn in filenames:
            res[fn] = fn
        # Translate names.
        for obj_name in obj_names:
            fn = trans.get(obj_name)
            if fn:
                res[fn] = obj_name
        return res


    def _expandPath(self, subpath):
        if self.basepath:
            while subpath.startswith('/') or subpath.startswith('\\'):
                subpath = subpath[1:]
            path = self.ops.join(self.basepath, subpath)
        else:
            # unchanged.
            path = subpath
        if not self.ops.exists(path):
            dir_path, obj_name = self.ops.split(path)
            if '.' not in obj_name:
                # This object might have an automatic filename extension.
                filenames, obj_names, trans = self._computeDirectoryContents(
                    dir_path, 1)
                fn = trans.get(obj_name)
                if fn is not None:
                    # Use the filename with an extension.
                    path = self.ops.join(dir_path, fn)
        return path


    def _checkSectionName(self, section_name):
        if (not isinstance(section_name, StringType)
            or not section_name
            or '[' in section_name
            or ']' in section_name
            or '\n' in section_name
            or section_name.startswith('@')
            or section_name == OBJECT_NAMES_SECTION):
            raise ValueError, section_name


    def writeSection(self, subpath, section_name, data):
        self._checkSectionName(section_name)
        self._queue(subpath, section_name, data)


    def writeNodeType(self, subpath, data):
        self._queue(subpath, NODE_TYPE_SECTION, data)


    def writeData(self, subpath, data, as_text=0):
        self._queue(subpath, DATA_SECTION, (data, as_text))


    def suggestExtension(self, subpath, ext):
        self._queue(subpath, SUGGESTED_EXTENSION_SECTION, ext)


    def readSection(self, subpath, section_name, default=None):
        self._checkSectionName(section_name)
        path = self._expandPath(subpath)
        sections = self._getPropertiesFromFile(path)
        return sections.get(section_name, default)


    def readNodeType(self, subpath):
        path = self._expandPath(subpath)
        if not self.ops.exists(path):
            raise NoStateFoundError(subpath)
        return self.ops.isdir(path) and 'd' or 'f'


    def readData(self, subpath, allow_missing=0, as_text=0):
        path = self._expandPath(subpath)
        isdir = self.ops.isdir(path)
        # Read either the directory listing or the file contents.
        if isdir:
            # Return a sequence of object names.
            return self._listDirectoryAsMapping(path).values()
        # Return a string.
        try:
            return self.ops.readfile(path, as_text)
        except IOError:
            if allow_missing:
                return None
            raise


    def getExtension(self, subpath):
        path = self._expandPath(subpath)
        stuff, ext = self.ops.splitext(path)
        return ext


    def getModTime(self, subpath, default=0):
        """Returns the time an object was last modified.

        Since objects are split into up to three files, this
        implementation returns the modification time of the most
        recently modified of the three.
        """
        path = self._expandPath(subpath)
        props, remainder = self._getPropertyPaths(path)
        maxtime = -1
        for p in (path, props, remainder):
            try:
                t = self.ops.getmtime(p)
            except OSError:
                pass
            else:
                if t > maxtime:
                    maxtime = t
        if maxtime == -1:
            maxtime = default
        return maxtime


    def _getPropertyPaths(self, path):
        """Returns the property and remainder paths for a path."""
        if self.ops.isdir(path):
            base_fn = self.ops.join(path, self.metadata_prefix)
        else:
            dirname, filename = self.ops.split(path)
            base_fn = self.ops.join(dirname, '%s%s.' % (
                self.metadata_prefix, filename))
        return (base_fn + PROPERTIES_EXTENSION, base_fn + REMAINDER_EXTENSION)


    def _getPropertiesFromFile(self, path):
        """Reads the properties and remainder for a path."""
        res = self._props_cache.get(path)
        if res is not None:
            return res

        props_fn, rem_fn = self._getPropertyPaths(path)

        res = {}
        try:
            data = self.ops.readfile(rem_fn, 0)
        except IOError:
            # The remainder file apparently does not exist
            pass
        else:
            res[REMAINDER_SECTION] = data
            # Note that the remainder can be overridden by the properties
            # file.  Perhaps that should be prevented in the future.

        try:
            data = self.ops.readfile(props_fn, 1)
        except IOError:
            # The properties file apparently does not exist
            self._props_cache.set(path, res)
            return res

        pos = 0
        prev_section_name = None
        while 1:
            match = section_re.search(data, pos)
            if match is None:
                endpos = len(data)
            else:
                endpos = match.start()
            if prev_section_name is not None:
                # get the data and decode.
                section = data[pos:endpos].replace('[[', '[')
                res[prev_section_name] = section
            if match is None:
                break
            else:
                prev_section_name = match.group(1)
                pos = match.end()

        self._props_cache.set(path, res)
        return res


    def _writeFinal(self, subpath, sections):
        """Performs an actual write of a file or directory to disk."""
        # sections is a mapping.
        path = self._expandPath(subpath)
        t = sections[NODE_TYPE_SECTION]
        if not self.ops.exists(path):
            if t == 'd':
                self.ops.mkdir(path)
            else:
                fn = self.ops.split(path)[1]
                if '.' not in fn:
                    # This object has no extension and doesn't yet exist.
                    ext = sections.get(SUGGESTED_EXTENSION_SECTION)
                    if ext:
                        # Try to use the suggested extension.
                        if not ext.startswith('.'):
                            ext = '.' + ext
                        p = path + ext
                        if not self.ops.exists(p):
                            # No file is in the way.
                            # Use the suggested extension.
                            path = p
        props_fn, rem_fn = self._getPropertyPaths(path)
        props_data = ''
        rem_data = ''
        items = sections.items()
        items.sort()
        try:
            for name, value in items:
                if name == NODE_TYPE_SECTION:
                    continue
                elif name == DATA_SECTION:
                    data, as_text = value
                    if t == 'd':
                        # Change the list of subobjects.
                        self._removeUnlinkedItems(path, data)
                        props_data += self._formatSection(
                            OBJECT_NAMES_SECTION, '\n'.join(data))
                        self._disableConflictingExtensions(subpath, data)
                        self._dir_cache.invalidate(path)
                    else:
                        # Change the file contents.
                        self.ops.writefile(path, as_text, data)
                elif name == SUGGESTED_EXTENSION_SECTION:
                    # This doesn't need to be written.
                    pass
                elif name == REMAINDER_SECTION:
                    # Write to the remainder file.
                    rem_data = value
                else:
                    # Write a metadata section.
                    props_data += self._formatSection(name, value)
        finally:
            self._writeOrRemove(props_fn, 1, props_data)
            self._writeOrRemove(rem_fn, 0, rem_data)
            self._props_cache.invalidate(path)
            # The file might be new, so invalidate the directory.
            self._dir_cache.invalidate(self.ops.dirname(path))


    def _formatSection(self, name, text):
        s = '[%s]\n%s\n' % (name, text.replace('[', '[['))
        if not text.endswith('\n'):
            s += '\n'
        return s


    def _writeOrRemove(self, fn, as_text, data):
        """If data is provided, write it.  Otherwise remove the file.
        """
        if data:
            self.ops.writefile(fn, as_text, data)
        else:
            if self.ops.exists(fn):
                self.ops.remove(fn)


    def _removeUnlinkedItems(self, path, names):
        """Removes unused files/subtrees from a directory."""
        linked = {}
        for name in names:
            linked[name] = 1
        for fn, obj_name in self._listDirectoryAsMapping(path).items():
            if not linked.get(obj_name):
                item_fn = self.ops.join(path, fn)
                if self.ops.isdir(item_fn):
                    self.ops.rmtree(item_fn)
                else:
                    self.ops.remove(item_fn)
                    props_fn, rem_fn = self._getPropertyPaths(item_fn)
                    if self.ops.exists(props_fn):
                        self.ops.remove(props_fn)
                    if self.ops.exists(rem_fn):
                        self.ops.remove(rem_fn)


    def _disableConflictingExtensions(self, subpath, obj_names):
        """Fixes collisions before writing files in a directory.

        Enforces the rule: if 'foo.*' is in the
        database, 'foo' may not have an automatic extension.
        Enforces by un-queuing suggested extensions.
        """
        reserved = {}  # { object name without extension -> 1 }
        for obj_name in obj_names:
            if '.' in obj_name:
                base, ext = obj_name.split('.', 1)
                reserved[base] = 1
        if not reserved:
            # No objects have extensions.
            return

        while subpath.endswith('/'):
            subpath = subpath[:-1]
        for obj_name in obj_names:
            if reserved.has_key(obj_name):
                # Prevent obj_name from using an automatic extension.
                child_subpath = '%s/%s' % (subpath, obj_name)
                self._queue(child_subpath, SUGGESTED_EXTENSION_SECTION,
                            '', force=1)


    def _beforeWrite(self, items):
        """Does some early checking while it's easy to bail out.

        This helps avoid exceptions during the second phase of
        transaction commit.
        """
        non_containers = {}
        for subpath, sections in items:
            path = self._expandPath(subpath)
            exists = self.ops.exists(path)
            if exists and not self.ops.canwrite(path):
                raise FSWriteError(
                    "Can't get write access to %s" % subpath)
            # type must be provided and must always be either 'd' or 'f'.
            if (not sections.has_key(NODE_TYPE_SECTION)
                or not sections.has_key(DATA_SECTION)):
                raise FSWriteError(
                    'Data or node type not specified for %s' % subpath)
            t = sections[NODE_TYPE_SECTION]
            dir = self.ops.dirname(subpath)
            if non_containers.get(dir):
                raise FSWriteError(
                    "Not a directory: %s" % dir)
            data, as_text = sections[DATA_SECTION]
            if t == 'f':
                if exists and self.ops.isdir(path):
                    raise FSWriteError(
                        "Can't write file data to directory at %s"
                        % subpath)
                non_containers[subpath] = 1
                if not isinstance(data, StringType):
                    raise FSWriteError(
                        'Data for a file must be a string at %s'
                        % subpath)
            elif t == 'd':
                if exists and not self.ops.isdir(path):
                    raise FSWriteError(
                        "Can't write directory contents to file at %s"
                        % subpath)
                if isinstance(data, StringType):
                    raise FSWriteError(
                        'Data for a directory must be a list or tuple at %s'
                        % subpath)
                for item in data:
                    if not self._isLegalFilename(item):
                        raise FSWriteError(
                            'Not a legal object name: %s' % repr(item))
            else:
                raise FSWriteError(
                    'Node type must be "d" or "f" at %s' % subpath)


    def _queue(self, subpath, section_name, data, force=0):
        """Queues data to be written at commit time"""
        m = self._pending
        sections = m.get(subpath)
        if sections is None:
            sections = {}
            m[subpath] = sections
        if sections.has_key(section_name) and not force:
            if sections[section_name] != data:
                raise FSWriteError(
                    'Conflicting data storage at %s (%s)' %
                    (subpath, section_name))
        else:
            sections[section_name] = data


    #
    # ITPCConnection implementation
    #

    def sortKey(self):
        return self.basepath

    def getName(self):
        return self.basepath

    def connect(self):
        if not self.ops.exists(self.basepath):
            self.ops.makedirs(self.basepath)

    def begin(self):
        self._props_cache.clear()
        self._dir_cache.clear()

    def vote(self):
        """Do some early verification

        This is done while the transaction can still be vetoed safely.
        """
        items = self._pending.items()
        items.sort()  # Ensure that base directories come first.
        self._beforeWrite(items)
        self._final = 1

    def reset(self):
        self._final = 0
        self._pending.clear()
        self._props_cache.clear()
        self._dir_cache.clear()

    def abort(self):
        self.reset()

    def finish(self):
        if self._final:
            try:
                items = self._pending.items()
                items.sort()  # Ensure that base directories come first.
                for subpath, sections in items:
                    self._writeFinal(subpath, sections)
            finally:
                self.reset()

    def close(self):
        self.reset()


    def _get_paths_mtime(self, paths):
        t = []
        for path in paths:
            try:
                t.append(self.ops.getmtime(path))
            except OSError:
                t.append(None)
        return t


    def getSources(self, subpath):
        p = self._expandPath(subpath)
        props, remainder = self._getPropertyPaths(p)
        paths = (p, props, remainder)
        t = self._get_paths_mtime(paths)
        return {(self, paths): t}


    def freshen(self, sources):
        """ISourceRepository implementation.

        Returns the changed items.
        """
        res = {}
        for source, t in sources.items():
            myself, paths = source
            assert myself is self
            new_t = self._get_paths_mtime(paths)
            if t != new_t:
                res[source] = new_t
        return res


