##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Basic filesystem gateways.

$Id: structure.py,v 1.2 2004/01/09 14:05:32 johanc Exp $
"""

from types import StringType

from apelib.core.interfaces import IGateway
from apelib.core.schemas import FieldSchema, RowSequenceSchema

from base import FSGatewayBase


class FSFileData (FSGatewayBase):
    """File data gateway, where data is a string.
    """

    __implements__ = IGateway

    schema = FieldSchema('data', 'string')

    def __init__(self, text=0, conn_name='fs'):
        if text == 'text':
            text = 1
        elif text == 'binary':
            text = 0
        self.text = text
        FSGatewayBase.__init__(self, conn_name)

    def load(self, event):
        c = self.getConnection(event)
        p = event.getKey()
        assert c.readNodeType(p) == 'f'
        state = c.readData(p, as_text=self.text)
        return state, state

    def store(self, event, state):
        if not isinstance(state, StringType):
            raise ValueError('Not a string: %s' % repr(state))
        c = self.getConnection(event)
        p = event.getKey()
        c.writeNodeType(p, 'f')
        c.writeData(p, state, as_text=self.text)
        return state


class FSAutoId (FSGatewayBase):
    """Automatic ID gateway based on the key of the item."""

    __implements__ = IGateway

    schema = FieldSchema('id', 'string')

    def getIdFrom(self, event):
        path = event.getKey()
        pos = path.rfind('/')
        if pos >= 0:
            return path[pos + 1:]
        else:
            return path

    def load(self, event):
        id = self.getIdFrom(event)
        return id, id

    def store(self, event, state):
        id = self.getIdFrom(event)
        if state != id:
            raise ValueError('Mismatched file ID')
        return id

    def getSources(self, event):
        fs_conn = self.getConnection(event)
        return fs_conn.getSources(event.getKey())
        


class FSDirectoryItems (FSGatewayBase):
    """Read/write objects in a filesystem directory."""

    __implements__ = IGateway

    schema = RowSequenceSchema()
    schema.addField('id', 'string', 1)
    schema.addField('keychain', 'keychain')

    def load(self, event):
        p = event.getKey()
        c = self.getConnection(event)
        assert c.readNodeType(p) == 'd'
        names = c.readData(p)
        names.sort()
        res = []
        for name in names:
            keychain = event.makeKeychain(name, 0)
            res.append((name, keychain))
        res = tuple(res)
        return res, res

    def store(self, event, state):
        p = event.getKey()
        c = self.getConnection(event)
        c.writeNodeType(p, 'd')
        state = list(state)
        state.sort()
        if __debug__:
            for name, keychain in state:
                expect = event.makeKeychain(name, 0)
                assert expect == keychain, (
                    "Child of %s named %s must use keychain %s, but used %s" %
                    (event.getKeychain(), name, expect, keychain))
        names = [row[0] for row in state]
        c.writeData(p, names)
        return tuple(state)


class FSModTime (FSGatewayBase):
    """Reads the modification time of a file."""

    __implements__ = IGateway

    schema = FieldSchema('mtime', 'int')

    def load(self, event):
        p = event.getKey()
        fs_conn = self.getConnection(event)
        state = long(fs_conn.getModTime(p))
        return state, None  # Use None as the hash (see store())

    def store(self, event, state):
        # Under normal circumstances, there is no need to change the mod
        # time of a file.  Ignore by returning None as the hash.
        return None

