##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""DB-API to Ape gateway connectivity

$Id: dbapi.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import os
from time import time
from types import StringType

from apelib.core.interfaces import ITPCConnection
from apelib.sql.interfaces import ISQLConnection
from querygen import makeQueryGenerator

DEBUG = os.environ.get('APE_DEBUG_SQL')
PROFILE = os.environ.get('APE_PROFILE_SQL')


class DBAPIConnector:

    __implements__ = ITPCConnection, ISQLConnection

    _final = 0
    db = None
    cursor = None

    def __init__(self, module_name, params=(), kwparams='', prefix='zodb'):
        self.module_name = module_name
        self.module = __import__(module_name, {}, {}, ('__doc__',))
        self.error = self.module.DatabaseError
        self.prefix = prefix
        self.connect_callbacks = []
        if isinstance(kwparams, StringType):
            # Convert the keyword string to a dictionary.
            kw = {}
            for p in kwparams.split():
                k, v = p.split(':')
                kw[k.strip()] = v.strip()
            kwparams = kw
        self.kwparams = kwparams
        if isinstance(params, StringType):
            params = (params,)
        self.params = params
        self.query_gens = {}      # { (table, columns) -> QueryGenerator }
        self.query_cache = {}     # { (table, columns, operation) -> query }

    def isConnected(self):
        return self.db is not None

    # ISQLConnection implementation

    def getQuery(self, table_name, column_defs, operation):
        query = self.query_cache.get((table_name, column_defs, operation))
        if query is not None:
            return query
        gen = self.query_gens.get((table_name, column_defs))
        if gen is None:
            if self.prefix:
                real_table_name = '%s_%s' % (self.prefix, table_name)
            else:
                real_table_name = table_name
            gen = makeQueryGenerator(
                self.module_name, real_table_name, column_defs)
            self.query_gens[(table_name, column_defs)] = gen
        query = gen.generate(operation)
        self.query_cache[(table_name, column_defs, operation)] = query
        return query

    def execute(self, query, fetch=0, cursor=None, _data=None, **kw):
        if cursor is None:
            cursor = self.cursor
            if cursor is None:
                raise RuntimeError('Not connected')
        if _data is not None:
            # Callers can pass _data instead of keyword arguments
            # to avoid argument name conflicts.
            kw.update(_data)
        if DEBUG or PROFILE:
            if kw:
                print 'Query: %s, %s' % (repr(query), kw)
            else:
                print 'Query: %s' % repr(query)
        if PROFILE:
            start = time()
            cursor.execute(query, kw)
            end = time()
            print 'Query time: %0.6fs' % (end - start)
        else:
            cursor.execute(query, kw)
        if fetch:
            res = list(cursor.fetchall())
            if DEBUG:
                print 'Query result: %s' % repr(res)
            return res
        return None

    def asBinary(self, data):
        return self.module.Binary(data)

    # ITPCConnection implementation

    def connect(self):
        if self.kwparams:
            self.db = self.module.connect(*self.params, **self.kwparams)
        else:
            self.db = self.module.connect(*self.params)
        self.cursor = self.db.cursor()
        for f in self.connect_callbacks:
            f()
        self.connect_callbacks = []

    def sortKey(self):
        return repr(self)

    def getName(self):
        return repr(self)

    def begin(self):
        pass

    def vote(self):
        self._final = 1

    def reset(self):
        self._final = 0

    def abort(self):
        try:
            self.db.rollback()
        finally:
            self.reset()

    def finish(self):
        if self._final:
            try:
                self.db.commit()
            finally:
                self.reset()

    def close(self):
        if self.isConnected():
            self.cursor.close()
            self.cursor = None
            self.db.close()
            self.db = None

