##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL generator.

$Id: querygen.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

query_generator_factories = {}

def addFactory(name, f):
    if query_generator_factories.has_key(name):
        raise KeyError("Factory name %s conflicts" % repr(name))
    query_generator_factories[name] = f

def makeQueryGenerator(module_name, table_name, column_defs):
    f = query_generator_factories.get(module_name)
    if f is None:
        raise KeyError(
            "No query generator registered for %s" % repr(module_name))
    return f(table_name, column_defs)


class AbstractQueryGenerator:

    column_type_translations = None  # { local type name -> db type name }
    column_name_translations = None  # { local col name -> db col name }

    key_column = ('key', 'int', 1)

    def __init__(self, table_name, column_defs):
        # column_defs is a sequence of (column_name, column_type, unique)
        self.table_name = table_name
        self.column_defs = (self.key_column,) + tuple(column_defs)

    def translateName(self, column_name):
        """Returns a column name for a variable name.

        Defaults to no translation.
        """
        return self.column_name_translations.get(column_name, column_name)

    def translateType(self, column_type):
        """Returns a database type for a variable type.

        If the type is unknown, raises KeyError.
        """
        return self.column_type_translations.get(column_type, column_type)


    def generate(self, operation):
        m = getattr(self, 'gen_' + operation)
        return m()


    def gen_check(self):
        kc = self.translateName(self.key_column[0])
        # This is meant to raise an error if the table is missing.
        # Returns nothing if the table exists.
        return 'SELECT %s FROM %s WHERE 0 = 1' % (kc, self.table_name)


    def gen_create(self):
        cols = []
        for name, typ, u in self.column_defs:
            trans_name = self.translateName(name)
            db_type = self.translateType(typ)
            cols.append("%s %s" % (trans_name, db_type))
        return "CREATE TABLE %s (%s)" % (self.table_name, ', '.join(cols))


    def gen_read(self):
        cols = []
        for name, typ, u in self.column_defs[1:]:
            trans_name = self.translateName(name)
            cols.append(trans_name)
        key_column_name = self.key_column[0]
        trans_kc = self.translateName(key_column_name)
        return 'SELECT %s FROM %s WHERE %s = %%(%s)s' % (
            ', '.join(cols), self.table_name, trans_kc, key_column_name)


    def gen_simple_search(self):
        cols = []
        where_clauses = []
        for name, typ, unique in self.column_defs[1:]:
            trans_name = self.translateName(name)
            cols.append(trans_name)
            if unique:
                # use this column to choose a record
                clause = "%s = %%(%s)s" % (trans_name, name)
                where_clauses.append(clause)
        return 'SELECT %s FROM %s WHERE %s' % (
            ', '.join(cols), self.table_name, ' and '.join(where_clauses))


    def gen_update(self):
        update_cols = []
        where_clauses = []
        for name, typ, unique in self.column_defs:
            trans_name = self.translateName(name)
            clause = "%s = %%(%s)s" % (trans_name, name)
            if unique:
                # use this column to choose a record
                where_clauses.append(clause)
            else:
                # update this column
                update_cols.append(clause)
        return 'UPDATE %s SET %s WHERE %s' % (
            self.table_name, ', '.join(update_cols),
            ' and '.join(where_clauses))


    def gen_insert(self):
        colnames = []
        colfmts = []
        for name, typ, unique in self.column_defs:
            trans_name = self.translateName(name)
            colnames.append(trans_name)
            colfmts.append('%%(%s)s' % name)
        return 'INSERT INTO %s (%s) VALUES (%s)' % (
            self.table_name, ', '.join(colnames), ', '.join(colfmts))
        

    def gen_delete(self):
        where_clauses = []
        for name, typ, unique in self.column_defs:
            if unique:
                trans_name = self.translateName(name)
                clause = "%s = %%(%s)s" % (trans_name, name)
                where_clauses.append(clause)
        return 'DELETE FROM %s WHERE %s' % (
            self.table_name, ' and '.join(where_clauses))


    def gen_clear(self):
        return 'DELETE FROM %s' % self.table_name



class PostgreSQLQueryGenerator (AbstractQueryGenerator):

    column_type_translations = {
        'long':   'bigint',
        'string': 'character varying(255)',
        'blob':   'bytea',
        'date_international': 'date',
        }

    column_name_translations = {}

    def gen_table_names(self):
        # Returns the names of the tables.
        return "SELECT tablename FROM pg_tables"

    def gen_sequence_check(self):
        return "SELECT last_value FROM %s" % self.table_name

    def gen_sequence_create(self):
        return "CREATE SEQUENCE %s" % self.table_name

    def gen_sequence_insert(self):
        return ""  # Do nothing

    def gen_sequence_read(self):
        return "SELECT nextval('%s')" % self.table_name

    def gen_sequence_update(self):
        return ""  # Do nothing

    def gen_sequence_clear(self):
        return "SELECT setval('%s', 1)" % self.table_name

addFactory('psycopg', PostgreSQLQueryGenerator)



class MySQLQueryGenerator (AbstractQueryGenerator):

    column_type_translations = {
        'long':   'bigint',
        'string': 'character varying(255)',
        'blob':   'longblob',
        'date_international': 'date',
        }

    column_name_translations = {
        'key': 'objkey',
        }

    def gen_table_names(self):
        # Returns the names of the tables.
        return "SHOW TABLES"

    def gen_sequence_check(self):
        return "SELECT last_value FROM %s" % self.table_name

    def gen_sequence_create(self):
        return "CREATE TABLE %s (last_value int)" % self.table_name

    def gen_sequence_insert(self):
        return "INSERT INTO %s VALUES (0)" % self.table_name

    def gen_sequence_read(self):
        return "SELECT LAST_INSERT_ID()"

    def gen_sequence_update(self):
        return ("UPDATE %s SET last_value=LAST_INSERT_ID(last_value+1)"
                % self.table_name)

    def gen_sequence_clear(self):
        return "UPDATE %s SET last_value=0" % self.table_name

addFactory('MySQLdb', MySQLQueryGenerator)

