##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL gateways for security information.

$Id: security.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

from apelib.core.schemas import RowSequenceSchema
from sqlbase import SQLGatewayBase


class SQLSecurityAttributes (SQLGatewayBase):
    """SQL security attribute storage"""

    __implements__ = SQLGatewayBase.__implements__

    schema = RowSequenceSchema()
    schema.addField('declaration_type', 'string')
    schema.addField('role', 'string')
    schema.addField('permission', 'string')
    schema.addField('username', 'string')

    table_base_name = 'security'

    def load(self, event):
        key = long(event.getKey())
        items = self.execute(event, 'read', 1, key=key)
        items.sort()
        return items, tuple(items)

    def store(self, event, state):
        key = long(event.getKey())
        items = self.execute(event, 'read', 1, key=key)
        items.sort()
        state = list(state)
        state.sort()
        if state != items:
            self.execute(event, 'delete', key=key)
            for row in state:
                self.execute(
                    event, 'insert', key=key, declaration_type=row[0],
                    role=row[1], permission=row[2], username=row[3])
        return tuple(state)



class SQLUserList (SQLGatewayBase):
    """Stores and retrieves all users for a folder at once."""

    __implements__ = SQLGatewayBase.__implements__

    schema = RowSequenceSchema()
    schema.addField('id', 'string', 1)
    schema.addField('password', 'string')
    schema.addField('roles', 'string:list')
    schema.addField('domains', 'string:list')

    table_defs = {
        'users':        (('id', 'string', 1),
                         ('password', 'string', 0),),
        'user_roles':   (('id', 'string', 1),
                         ('role', 'string', 0),),
        'user_domains': (('id', 'string', 1),
                         ('domain', 'string', 0),),
        }


    def init(self, event):
        conn = event.getConnection(self.conn_name)
        try:
            if event.clearing():
                self.execute(event, 'users', 'clear')
                self.execute(event, 'user_roles', 'clear')
                self.execute(event, 'user_domains', 'clear')
            else:
                self.execute(event, 'users', 'check')
        except conn.error:
            conn.db.rollback()
            self.execute(event, 'users', 'create')
            self.execute(event, 'user_roles', 'create')
            self.execute(event, 'user_domains', 'create')
            conn.db.commit()


    def execute(self, event, table, operation, *args, **kw):
        conn = event.getConnection(self.conn_name)
        query = conn.getQuery(
            table, self.table_defs[table], operation)
        if query == '':
            # No query needed for this operation
            return
        return conn.execute(query, *args, **kw)


    def load(self, event):
        key = event.getKeychain()[-1]
        rows = self.execute(event, 'users', 'read', 1, key=key)
        data = {}
        for id, password in rows:
            data[id] = (password, [], [])
        rows = self.execute(event, 'user_roles', 'read', 1, key=key)
        for id, role in rows:
            row = data.get(id)
            if row is not None:
                row[1].append(role)
        rows = self.execute(event, 'user_domains', 'read', 1, key=key)
        for id, domain in rows:
            row = data.get(id)
            if row is not None:
                row[2].append(domain)
        records = []
        for id, (password, roles, domains) in data.items():
            records.append((id, password, tuple(roles), tuple(domains)))
        records.sort()
        return records, tuple(records)


    def store(self, event, state):
        keychain = event.getKeychain()
        keychain1 = keychain[:-1]
        key = long(keychain[-1])
        old_rows, old_serial = self.load(event)
        new_dict = {}
        for rec in state:
            new_dict[rec[0]] = rec
        old_dict = {}
        for rec in old_rows:
            old_dict[rec[0]] = rec

        # Compare / remove rows.
        for old_row in old_rows:
            id = old_row[0]
            new_row = new_dict.get(id)
            if new_row is None:
                # Remove this row.
                self.execute(event, 'users', 'delete', key=key, id=id)
                self.execute(event, 'user_roles', 'delete', key=key, id=id)
                self.execute(event, 'user_domains', 'delete', key=key, id=id)
            elif new_row == old_row:
                # Don't need to update this row.
                del new_dict[id]

        # Insert / update rows.
        for new_row in new_dict.values():
            id, password, roles, domains = new_row
            old_row = old_dict.get(id)
            if old_row is not None:
                old_id, old_password, old_roles, old_domains = old_row
                if old_password != password:
                    # Update the password.
                    self.execute(event, 'users', 'update',
                                 key=key, id=id, password=password)
            else:
                # Insert a new record.
                self.execute(event, 'users', 'insert',
                             key=key, id=id, password=password)
                old_roles = ()
                old_domains = ()

            # Update the role list.
            if tuple(roles) != tuple(old_roles):
                self.execute(event, 'user_roles', 'delete', key=key, id=id)
                for role in roles:
                    assert role
                    self.execute(event, 'user_roles', 'insert',
                                 key=key, id=id, role=role)

            # Update the domain list.
            if tuple(domains) != tuple(old_domains):
                self.execute(event, 'user_domains', 'delete', key=key, id=id)
                for domain in domains:
                    assert domain
                    self.execute(event, 'user_domains', 'insert',
                                 key=key, id=id, domain=domain)
                
        state = list(state)
        state.sort()
        return tuple(state)

