##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL gateways for a filesystem-like tree structure.

$Id: structure.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

from apelib.core.schemas import FieldSchema, RowSequenceSchema
from sqlbase import SQLGatewayBase


class SQLObjectData (SQLGatewayBase):
    """SQL object data gateway"""

    __implements__ = SQLGatewayBase.__implements__

    schema = FieldSchema('data', 'string')

    table_base_name = 'object_data'

    column_defs = (
        ('data', 'blob', 0),
        )

    def load(self, event):
        key = long(event.getKey())
        items = self.execute(event, 'read', 1, key=key)
        if items:
            state = items[0][0]
        else:
            state = ''
        return state, state

    def store(self, event, state):
        key = long(event.getKey())
        items = self.execute(event, 'read', 1, key=key)
        col_name = self.column_defs[0][0]
        conn = event.getConnection(self.conn_name)
        kw = {'key': key, col_name: conn.asBinary(state)}
        if items:
            # update.
            self.execute(event, 'update', **kw)
        else:
            # insert.
            self.execute(event, 'insert', **kw)
        return state


class SQLFolderItems (SQLGatewayBase):
    """SQL folder items gateway"""

    __implements__ = SQLGatewayBase.__implements__

    schema = RowSequenceSchema()
    schema.addField('id', 'string', 1)
    schema.addField('keychain', 'keychain')

    table_base_name = 'folder_items'

    column_defs = (
        ('name', 'string', 1),
        ('child_key', 'int', 0),
        )

    def load(self, event):
        keychain = event.getKeychain()
        key = long(keychain[-1])
        prefix = keychain[:-1]
        rows = self.execute(event, 'read', 1, key=key)
        rows.sort()
        res = [(row[0], prefix + (long(row[1]),)) for row in rows]
        return res, tuple(res)

    def store(self, event, state):
        keychain = event.getKeychain()
        keychain1 = keychain[:-1]
        key = long(keychain[-1])
        state_dict = {}
        for name, child_keychain in state:
            state_dict[name] = 1

        rows = self.execute(event, 'read', 1, key=key)
        db_dict = {}
        for name, child_key in rows:
            if state_dict.has_key(name):
                db_dict[name] = child_key
            else:
                # Remove this item from the database.
                self.execute(event, 'delete', key=key, name=name)

        state = list(state)
        state.sort()
        for name, child_keychain in state:
            if child_keychain[:-1] != keychain1:
                raise ValueError(
                    "SQLFolderItems cannot store cross-domain references")
            child_key = long(child_keychain[-1])
            kw = {'key': key, 'name': name, 'child_key': child_key}
            if db_dict.has_key(name):
                if db_dict[name] != child_key:
                    # Change this item to point to a different OID.
                    self.execute(event, 'update', **kw)
            else:
                # Add this item to the database.
                self.execute(event, 'insert', **kw)
        return tuple(state)


class SQLItemId (SQLGatewayBase):
    """SQL item ID gateway.

    Piggybacks SQLFolderItems for init and store.
    Makes the assumption that the item is stored in only one place.
    """

    __implements__ = SQLGatewayBase.__implements__

    schema = FieldSchema('id', 'string')

    table_base_name = 'folder_items'

    column_defs = (('child_key', 'int', 1),
                   ('name', 'string', 0),)

    def init(self, event):
        pass

    def load(self, event):
        key = long(event.getKey())
        rows = self.execute(event, 'simple_search', 1, child_key=key)
        assert len(rows) >= 1
        name = rows[0][1]  # Accept only the first result
        return name, None

    def store(self, event, state):
        return None


class SQLRemainder (SQLObjectData):
    """SQL remainder pickle gateway"""

    __implements__ = SQLGatewayBase.__implements__

    table_base_name = 'remainder'

    column_defs = (
        ('pickle', 'blob', 0),
        )


class SQLModTime (SQLGatewayBase):
    """SQL object mod time gateway"""

    __implements__ = SQLGatewayBase.__implements__

    schema = FieldSchema('mtime', 'int')  # second

    table_base_name = 'mtime'

    column_defs = (
        ('mtime', 'long', 0),
        )

    def load(self, event):
        key = long(event.getKey())
        items = self.execute(event, 'read', 1, key=key)
        if items:
            state = long(items[0][0])
        else:
            state = 0L
        return state, state

    def store(self, event, state):
        key = long(event.getKey())
        state = long(state)
        items = self.execute(event, 'read', 1, key=key)
        kw = {'key': key, 'mtime': state}
        if items:
            # update.
            self.execute(event, 'update', **kw)
        else:
            # insert.
            self.execute(event, 'insert', **kw)
        return state

