##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Serialization test setup/teardown

$Id: serialtestbase.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import ZODB
from Persistence import PersistentMapping
from cPickle import dumps, loads

from apelib.core \
     import classifiers, gateways, keygen, mapper, schemas, serializers
from apelib.core.interfaces import ISerializer

from apelib.zodb3.serializers \
     import FixedPersistentMapping, RollCall, RemainingState


class SimpleItemsSerializer:

    __implements__ = ISerializer

    schema = schemas.RowSequenceSchema()
    schema.addField('name')
    schema.addField('pickle')

    def getSchema(self):
        return self.schema

    def serialize(self, object, event):
        res = []
        for k, v in object.items():
            res.append((k, dumps(v)))
            event.notifySerialized(k, v, 0)
        res.sort()
        event.ignoreAttribute('data')
        event.ignoreAttribute('_container')
        return res

    def deserialize(self, object, event, state):
        d = {}
        for k, v in state:
            o = loads(v)
            d[k] = o
            event.notifyDeserialized(k, o)
        object.__init__(d)


class SerialTestBase:

    def setUp(self):
        self.conns = {}

        cfr = classifiers.FixedClassifier()
        cfr.register('test', 'test_mapper')
        cfr.register('test2', 'test_mapper_2')

        ser2 = serializers.CompositeSerializer(
            'Persistence', 'PersistentMapping')
        fixed_items_serializer = FixedPersistentMapping()
        fixed_items_serializer.add('TestRoot', ('test',), ('test_mapper',))
        fixed_items_serializer.add('TestRoot2', ('test2',), ('test_mapper_2',))
        ser2.addSerializer('fixed_items', fixed_items_serializer)
        ser2.addSerializer('roll_call', RollCall())

        root_mapper = mapper.Mapper(
            None, ser2, gateways.CompositeGateway(), cfr)
        self.root_mapper = root_mapper
        # Since all of the keychains used by this mapper are fixed, we
        # don't want any keychains to be generated.
        root_mapper.setKeychainGenerator(keygen.NullKeychainGenerator())

        # Create "test_mapper", which allows a "strdata" attribute.

        ser1 = serializers.CompositeSerializer(
            'Persistence', 'PersistentMapping')
        items_serializer = SimpleItemsSerializer()
        ser1.addSerializer('items', items_serializer)
        props_serializer = serializers.StringDataAttribute('strdata')
        ser1.addSerializer('properties', props_serializer)
        ser1.addSerializer('roll_call', RollCall())

        gw1 = gateways.CompositeGateway()
        items_gw = gateways.MappingGateway(items_serializer.getSchema())
        self.items_gw = items_gw
        gw1.addGateway('items', items_gw)
        props_gw = gateways.MappingGateway(props_serializer.getSchema())
        self.props_gw = props_gw
        gw1.addGateway('properties', props_gw)

        om1 = mapper.Mapper(root_mapper, ser1, gw1)
        self.om1 = om1
        root_mapper.addSubMapper('test_mapper', om1)

        # Create "test_mapper_2", which stores a remainder.

        ser = serializers.CompositeSerializer(
            'Persistence', 'PersistentMapping')
        items_serializer = SimpleItemsSerializer()
        ser.addSerializer('items', items_serializer)
        remainder_serializer = RemainingState()
        ser.addSerializer('remainder', remainder_serializer)

        gw = gateways.CompositeGateway()
        items_gw = gateways.MappingGateway(items_serializer.getSchema())
        gw.addGateway('items', items_gw)
        remainder_gw = gateways.MappingGateway(
            remainder_serializer.getSchema())
        gw.addGateway('remainder', remainder_gw)

        om = mapper.Mapper(root_mapper, ser, gw)
        root_mapper.addSubMapper('test_mapper_2', om)

        root_mapper.checkConfiguration()


    def tearDown(self):
        pass

