##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of apelib.core.io

$Id: testio.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import unittest

import ZODB
from Persistence import PersistentMapping

from apelib.core import io
from apelib.core.interfaces import IKeyedObjectSystem
from serialtestbase import SerialTestBase


class TestKeyedObjectSystem:
    __implements__ = IKeyedObjectSystem

    def getObject(self, keychain, hints=None):
        raise NotImplementedError

    loadStub = getObject

    def identifyObject(self, obj):
        raise NotImplementedError

    def newKey(self):
        raise NotImplementedError
    
    def getClass(self, module, name):
        m = __import__(module)
        return getattr(m, name)



class ApeIOTests(SerialTestBase, unittest.TestCase):

    def getKeyedObjectSystem(self):
        return TestKeyedObjectSystem()

    def testImpl(self):
        # Test of test :-)
        from Interface.Verify import verifyClass
        verifyClass(IKeyedObjectSystem, TestKeyedObjectSystem)

    def testSerializeAndDeserialize(self):
        ob = PersistentMapping()
        ob.strdata = '345'
        ob['a'] = 'b'
        ob['c'] = 'd'
        keychain = ('test',)
        kos = self.getKeyedObjectSystem()
        obsys = io.ObjectSystemIO(self.root_mapper, kos)
        event, classified_state = obsys.serialize(keychain, ob)

        ob2 = obsys.newObject(classified_state)
        obsys.deserialize(keychain, ob2, classified_state)
        self.assertEqual(ob.strdata, ob2.strdata)
        self.assertEqual(ob.data, ob2.data)


    def testStoreAndLoad(self):
        # Tests both serialization and storage
        ob = PersistentMapping()
        ob.strdata = '345'
        ob['a'] = 'b'
        ob['c'] = 'd'
        keychain = ('test',)
        kos = self.getKeyedObjectSystem()
        obsys = io.ObjectSystemIO(self.root_mapper, kos)
        gwsys = io.GatewayIO(self.root_mapper, self.conns)
        event, classified_state = obsys.serialize(keychain, ob)
        gwsys.store(keychain, classified_state)

        event, cs, hash_value = gwsys.load(keychain)
        ob2 = obsys.newObject(cs)
        obsys.deserialize(keychain, ob2, cs)
        self.assertEqual(ob.strdata, ob2.strdata)
        self.assertEqual(ob.data, ob2.data)


    def testExportImport(self):
        root = PersistentMapping()

        test1 = PersistentMapping()
        test1.strdata = '345'
        test1['a'] = 'b'
        test1['c'] = 'd'
        root['TestRoot'] = test1
        test2 = PersistentMapping()
        test2.leftover = 'oops'
        test2['true'] = 'undecided'
        root['TestRoot2'] = test2

        keychain = ()
        exporter = io.ExportImport(self.root_mapper, self.conns)
        exporter.exportObject(root, keychain)

        importer = io.ExportImport(self.root_mapper, self.conns)
        roota = importer.importObject(keychain)
        self.assert_(root is not roota)
        self.assert_(root['TestRoot'] is not roota['TestRoot'])
        self.assert_(root['TestRoot2'] is not roota['TestRoot2'])
        self.assertEqual(root['TestRoot'].data, roota['TestRoot'].data)
        self.assertEqual(root['TestRoot2'].data, roota['TestRoot2'].data)
        self.assertEqual(root['TestRoot'].strdata, roota['TestRoot'].strdata)
        self.assertEqual(root['TestRoot2'].leftover,
                         roota['TestRoot2'].leftover)



if __name__ == '__main__':
    unittest.main()
