##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Serialization tests

$Id: testserialization.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import unittest

import ZODB
from Persistence import PersistentMapping

from apelib.core.events \
     import LoadEvent, StoreEvent, SerializationEvent, DeserializationEvent
from apelib.core.exceptions import SerializationError
from serialtestbase import SerialTestBase


class SimpleInstance:

    def __init__(self, data):
        self.data = data


class SerializationTests(SerialTestBase, unittest.TestCase):

    def getKeyedObjectSystem(self):
        # XXX works for now
        return None

    def testSerializeAndDeserialize(self):
        ob = PersistentMapping()
        ob.strdata = '345'
        ob['a'] = 'b'
        ob['c'] = 'd'
        kos = self.getKeyedObjectSystem()
        mapper = self.root_mapper.getSubMapper('test_mapper')
        event = SerializationEvent(kos, mapper, ('',), ob)
        full_state = mapper.getSerializer().serialize(ob, event)
        ob2 = PersistentMapping()
        event = DeserializationEvent(kos, mapper, ('',), ob2)
        mapper.getSerializer().deserialize(ob2, event, full_state)
        self.assertEqual(ob.strdata, ob2.strdata)
        self.assertEqual(ob.data, ob2.data)

    def testStoreAndLoad(self):
        ob = PersistentMapping()
        ob.strdata = '345'
        ob['a'] = 'b'
        ob['c'] = 'd'
        kos = self.getKeyedObjectSystem()
        mapper = self.root_mapper.getSubMapper('test_mapper')
        event = SerializationEvent(kos, mapper, ('',), ob)
        full_state = mapper.getSerializer().serialize(ob, event)
        event = StoreEvent(mapper, ('',), self.conns, None)
        mapper.getGateway().store(event, full_state)

        event = LoadEvent(mapper, ('',), self.conns, None)
        full_state, serial = mapper.getGateway().load(event)
        ob2 = PersistentMapping()
        event = DeserializationEvent(kos, mapper, ('',), ob2)
        mapper.getSerializer().deserialize(ob2, event, full_state)
        self.assertEqual(ob.strdata, ob2.strdata)
        self.assertEqual(ob.data, ob2.data)

    def testCatchExtraAttribute(self):
        ob = PersistentMapping()
        ob.strdata = '345'
        ob.extra = '678'
        ob['a'] = 'b'
        ob['c'] = 'd'
        kos = self.getKeyedObjectSystem()
        mapper = self.root_mapper.getSubMapper('test_mapper')
        event = SerializationEvent(kos, mapper, ('',), ob)
        self.assertRaises(SerializationError,
                          mapper.getSerializer().serialize, ob, event)

    def testSharedAttribute(self):
        # Test of an attribute shared between a normal serializer and
        # a remainder serializer.
        ob = PersistentMapping()
        data = SimpleInstance('This is a shared piece of data')
        ob.extra = data
        ob['a'] = data
        kos = self.getKeyedObjectSystem()
        mapper = self.root_mapper.getSubMapper('test_mapper_2')
        event = SerializationEvent(kos, mapper, ('',), ob)
        full_state = mapper.getSerializer().serialize(ob, event)
        event = StoreEvent(mapper, ('',), self.conns, None)
        mapper.getGateway().store(event, full_state)

        # Now load the state into a different object
        event = LoadEvent(mapper, ('',), self.conns, None)
        full_state, serial = mapper.getGateway().load(event)
        ob2 = PersistentMapping()
        event = DeserializationEvent(kos, mapper, ('',), ob2)
        mapper.getSerializer().deserialize(ob2, event, full_state)
        self.assertEqual(ob.extra.data, ob2.extra.data)
        self.assertEqual(ob.keys(), ob2.keys())

        # Check that both see the *same* object
        self.assert_(ob2['a'] is ob2.extra, (ob2['a'], ob2.extra))
        self.assert_(ob2['a'] is not data)  # Verify it didn't cheat somehow


if __name__ == '__main__':
    unittest.main()
