##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Storage tests (with data stored in simple mappings)

$Id: teststorage.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import unittest
from thread import start_new_thread, allocate_lock

import ZODB
from Persistence import PersistentMapping

from apelib.zodb3.db import ApeDB
from apelib.zodb3.storage import ApeStorage
from apelib.zodb3.resource import StaticResource
from apelib.zodb3.utils import copyOf
from serialtestbase import SerialTestBase


def run_in_thread(f):
    """Calls a function in another thread and waits for it to finish."""
    lock = allocate_lock()
    def run(f=f, lock=lock):
        try:
            f()
        finally:
            lock.release()
    lock.acquire()
    start_new_thread(run, ())
    lock.acquire()
    lock.release()


class ApeStorageTests (SerialTestBase, unittest.TestCase):
    # Tests of ApeStorage and ApeConnection.

    def setUp(self):
        SerialTestBase.setUp(self)
        resource = StaticResource(self.root_mapper)
        storage = ApeStorage(resource, self.conns)
        self.storage = storage
        db = ApeDB(storage, resource)
        self.db = db

    def tearDown(self):
        get_transaction().abort()
        self.db.close()
        SerialTestBase.tearDown(self)

    def testStoreAndLoad(self):
        ob = PersistentMapping()
        ob.strdata = '345'
        ob['a'] = 'b'
        ob['c'] = 'd'

        dummy = PersistentMapping()

        conn1 = self.db.open()
        conn2 = None
        conn3 = None
        try:

            # Load the root and create a new object
            root = conn1.root()
            get_transaction().begin()
            root['TestRoot'] = ob
            root['TestRoot2'] = dummy
            get_transaction().commit()
            ob1 = conn1.getObject(('test',))
            self.assertEqual(ob1.strdata, ob.strdata)
            self.assertEqual(ob1.items(), ob.items())

            # Verify a new object was stored and make a change
            get_transaction().begin()
            conn2 = self.db.open()
            ob2 = conn2.getObject(('test',))
            self.assertEqual(ob2.strdata, ob.strdata)
            self.assertEqual(ob2.items(), ob.items())
            ob2.strdata = '678'
            get_transaction().commit()

            # Verify the change was stored and make another change
            conn3 = self.db.open()
            ob3 = conn3.getObject(('test',))
            self.assertEqual(ob3.strdata, '678')
            self.assertEqual(ob3.items(), ob.items())
            ob3.strdata = '901'
            get_transaction().commit()
            conn3.close()
            conn3 = None
            conn3 = self.db.open()
            ob3 = conn3.getObject(('test',))
            self.assertEqual(ob3.strdata, '901')

            # Verify we didn't accidentally change the original object
            self.assertEqual(ob.strdata, '345')

        finally:
            conn1.close()
            if conn2 is not None:
                conn2.close()
            if conn3 is not None:
                conn3.close()


    def testUnmanaged(self):
        ob = PersistentMapping()
        ob['a'] = 'b'
        ob.stowaway = PersistentMapping()
        ob.stowaway['c'] = 'd'

        dummy = PersistentMapping()
        dummy.strdata = 'foo'

        conn1 = self.db.open()
        conn2 = None
        conn3 = None
        try:

            # Load the root and create a new object
            root = conn1.root()
            get_transaction().begin()
            root['TestRoot'] = dummy
            root['TestRoot2'] = ob
            get_transaction().commit()
            ob1 = conn1.getObject(('test2',))
            self.assert_(ob1 is ob)
            self.assertEqual(ob1.items(), [('a', 'b')])
            self.assertEqual(ob1.stowaway.items(), [('c', 'd')])

            # Verify a new object was stored
            get_transaction().begin()
            conn2 = self.db.open()
            ob2 = conn2.getObject(('test2',))
            self.assertEqual(ob2.items(), [('a', 'b')])
            self.assertEqual(ob2.stowaway.items(), [('c', 'd')])

            # Make a change only to the unmanaged persistent object
            # (the "stowaway").
            ob.stowaway['c'] = 'e'
            get_transaction().commit()

            # Verify the change was stored and make a change to the
            # managed persistent object.
            conn3 = self.db.open()
            ob3 = conn3.getObject(('test2',))
            self.assertEqual(ob3.items(), [('a', 'b')])
            self.assertEqual(ob3.stowaway.items(), [('c', 'e')])
            ob3['a'] = 'z'
            get_transaction().commit()
            conn3.close()
            conn3 = None
            conn3 = self.db.open()
            ob3 = conn3.getObject(('test2',))
            self.assertEqual(ob3['a'], 'z')
            self.assertEqual(ob3.stowaway.items(), [('c', 'e')])

            # Verify we didn't accidentally change the original object.
            self.assertEqual(ob['a'], 'b')

            # sync and verify the current state.
            conn1.sync()
            self.assertEqual(ob1.items(), [('a', 'z')])
            self.assertEqual(ob1.stowaway.items(), [('c', 'e')])

        finally:
            conn1.close()
            if conn2 is not None:
                conn2.close()
            if conn3 is not None:
                conn3.close()


    def testStoreAndLoadBinary(self):
        ob = PersistentMapping()
        # strdata contains binary characters
        ob.strdata = ''.join([chr(n) for n in range(256)]) * 2

        dummy = PersistentMapping()

        conn1 = self.db.open()
        try:
            root = conn1.root()
            get_transaction().begin()
            root['TestRoot'] = ob
            root['TestRoot2'] = dummy
            get_transaction().commit()
            ob1 = conn1.getObject(('test',))
            self.assertEqual(ob1.strdata, ob.strdata)
            self.assertEqual(ob1.items(), ob.items())
        finally:
            conn1.close()


    def _writeBasicObjects(self, conn):
        ob = PersistentMapping()
        ob.strdata = 'abc'
        dummy = PersistentMapping()
        root = conn.root()
        get_transaction().begin()
        root['TestRoot'] = ob
        root['TestRoot2'] = dummy
        get_transaction().commit()
        return ob, dummy


    def _changeTestRoot(self):
        conn2 = self.db.open()
        try:
            ob2 = conn2.root()['TestRoot']
            ob2.strdata = 'ghi'
            get_transaction().commit()
        finally:
            conn2.close()


    def testConflictDetection(self):
        conn1 = self.db.open()
        try:
            ob1, dummy = self._writeBasicObjects(conn1)
            ob1.strdata = 'def'
            run_in_thread(self._changeTestRoot)
            # Verify that "def" doesn't get written, since it
            # conflicts with "ghi".
            self.assertRaises(ZODB.POSException.ConflictError,
                              get_transaction().commit)
            self.assertEqual(ob1.strdata, "ghi")
        finally:
            conn1.close()


    def testNewObjectConflictDetection(self):
        # Verify a new object won't overwrite existing objects by accident
        conn1 = self.db.open()
        try:
            ob1, dummy = self._writeBasicObjects(conn1)
            ob1.strdata = 'def'
            conn1.setSerial(ob1, '\0' * 8)  # Pretend that it's new
            self.assertRaises(ZODB.POSException.ConflictError,
                              get_transaction().commit)
        finally:
            conn1.close()


    def testRemainderCyclicReferenceRestoration(self):
        # test whether the remainder pickler properly stores cyclic references
        # back to the object itself.
        ob1 = PersistentMapping()
        ob1.myself = ob1

        dummy = PersistentMapping()
        dummy.strdata = ''

        conn1 = self.db.open()
        try:
            root = conn1.root()
            get_transaction().begin()
            root['TestRoot'] = dummy
            root['TestRoot2'] = ob1
            get_transaction().commit()

            conn2 = self.db.open()
            try:
                ob2 = conn2.root()['TestRoot2']
                self.assert_(ob2.myself is ob2)
                self.assert_(ob2 is not ob1)  # Verify it didn't cheat somehow
            finally:
                conn2.close()
        finally:
            conn1.close()


    def testCopyOf(self):
        # Verifies the functionality of copyOf().
        ob1 = PersistentMapping()
        ob1._p_oid = 'xxx'
        self.assertEqual(ob1._p_oid, 'xxx')  # Precondition
        ob1['fish'] = PersistentMapping()
        ob1['fish']['trout'] = 1
        ob1['fish']['herring'] = 2

        ob2 = copyOf(ob1)
        self.assert_(ob2 is not ob1)
        self.assert_(ob2['fish'] is not ob1['fish'])
        self.assert_(ob2._p_oid is None)
        self.assertEqual(list(ob2.keys()), ['fish'])
        self.assertEqual(len(ob2['fish'].keys()), 2)


    def testCopyOfZClassInstance(self):
        # Verifies that copyOf() can copy instances that look like ZClass
        # instances.
        class weird_class (ZODB.Persistent):
            pass
        weird_class.__module__ = '*IAmAZClassModule'
        self.assertEqual(weird_class.__module__, '*IAmAZClassModule')

        ob1 = PersistentMapping()
        ob1['fishy'] = weird_class()

        ob2 = copyOf(ob1)
        self.assert_(ob2 is not ob1)
        self.assert_(ob2['fishy'] is not ob1['fishy'])
        self.assert_(ob2['fishy'].__class__ is weird_class)


    def test_p_serial_untouched(self):
        # _p_serial isn't safe to use for hashes, since _p_mtime
        # interprets it as a date stamp.  Verify Ape doesn't
        # use _p_serial for hashes.
        conn1 = self.db.open()
        try:
            ob1, dummy = self._writeBasicObjects(conn1)
            self.assertEqual(ob1._p_serial, "\0" * 8)
            self.assertEqual(dummy._p_serial, "\0" * 8)
        finally:
            conn1.close()


    def testGetSerial(self):
        # Verifies the behavior of getSerial().
        conn1 = self.db.open()
        try:
            new_ob = PersistentMapping()
            self.assertEqual(conn1.getSerial(new_ob), '\0' * 8)
            ob1, dummy = self._writeBasicObjects(conn1)
            self.assertNotEqual(conn1.getSerial(ob1), '\0' * 8)
        finally:
            conn1.close()


    def testGetSerialDetectsNewObjects(self):
        # Verifies the behavior of getSerial() and setSerial().
        conn1 = self.db.open()
        try:
            ob1, dummy = self._writeBasicObjects(conn1)
            self.assertNotEqual(conn1.getSerial(ob1), '\0' * 8)
            # Replace the object and verify it gets a new serial.
            ob1 = PersistentMapping()
            ob1.strdata = 'cba'
            ob1._p_oid = conn1.root()['TestRoot']._p_oid
            conn1.root()['TestRoot'] = ob1
            self.assertEqual(conn1.getSerial(ob1), '\0' * 8)
        finally:
            conn1.close()


    def testSerialCleanup(self):
        # Verify that setSerial() cleans up.
        conn1 = self.db.open()
        try:
            conn1.SERIAL_CLEANUP_THRESHOLD = 10
            for n in range(conn1.SERIAL_CLEANUP_THRESHOLD + 1):
                new_ob = PersistentMapping()
                new_ob._p_oid = 'fake_oid_' + str(n)
                old_size = len(conn1._serials or ())
                conn1.setSerial(new_ob, '01234567')
                new_size = len(conn1._serials)
                if new_size < old_size:
                    # Cleaned up.  Success.
                    break
            else:
                self.fail("setSerial() did not clean up")
        finally:
            conn1.close()


    def testGetSources(self):
        sources = self.storage.getSources('\0' * 8)
        self.assertEqual(sources, {})
        # The test passed, but check for a false positive.
        oid = self.storage._oid_encoder.encode(('nonexistent-oid',))
        self.assertRaises(KeyError, self.storage.getSources, oid)


if __name__ == '__main__':
    unittest.main()

