##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test of storing folders on the filesystem via ZODB

$Id: testzope2fs.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import os
import sys
from shutil import rmtree
import unittest
from tempfile import mktemp
from cStringIO import StringIO

from ZODB.POSException import ConflictError
from OFS.Image import File, manage_addImage, manage_addFile
from Products.PythonScripts.PythonScript import PythonScript
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate

from apelib.zodb3.db import ApeDB
from apelib.zodb3.storage import ApeStorage
from apelib.zodb3.resource import StaticResource
from apelib.zope2.mapper import createMapper
from apelib.fs.exceptions import FSWriteError
from apelib.fs.connection import FSConnection
from zope2testbase import Zope2TestBase, Folder


try:
    __file__
except NameError:
    __file__ = os.path.abspath(sys.argv[0])

tmpdir = mktemp()

root_mapper = createMapper('filesystem')


class Zope2FSTests (unittest.TestCase, Zope2TestBase):

    def _createConnections(self, path):
        conn = FSConnection(path)
        return {'fs': conn}

    def setUp(self):
        if not os.path.exists(tmpdir):
            os.mkdir(tmpdir)
        self.path = tmpdir
        self.root_mapper = root_mapper
        conns = self._createConnections(tmpdir)
        assert len(conns) == 1
        self.conn = conns['fs']
        resource = StaticResource(self.root_mapper)
        storage = ApeStorage(resource, conns)
        self.storage = storage
        db = ApeDB(storage, resource)
        self.db = db
        get_transaction().begin()

    def tearDown(self):
        get_transaction().abort()
        self.db.close()
        rmtree(self.path)

    def testClassificationPreservation(self):
        # Ensure that classification doesn't get forgotten.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'Holidays'
            app._setObject(f.id, f, set_owner=0)
            get_transaction().commit()

            f2 = Folder()
            f2.id = 'Christmas'
            f._setObject(f2.id, f2, set_owner=0)
            get_transaction().commit()

            f3 = Folder()
            f3.id = 'Eve'
            f2._setObject(f3.id, f3, set_owner=0)
            get_transaction().commit()

            for folder in (f, f2, f3):
                keychain = conn.db()._oid_encoder.decode(folder._p_oid)
                key = keychain[-1]
                text = self.conn.readSection(key, 'classification')
                self.assert_(text.find('class_name=OFS.Folder.Folder') >= 0)
        finally:
            conn.close()


    def testMismatchedIdDetection(self):
        # FSAutoID should detect when the keychain and ID don't match.
        # Normally, the only time they don't match is when an object
        # has been moved.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'Holidays'
            app._setObject(f.id, f, set_owner=0)
            get_transaction().commit()

            ob = app.Holidays
            ob._setId('HolidayCalendar')
            self.assertRaises(ValueError, get_transaction().commit)
        finally:
            conn.close()


    def testReuseId(self):
        # Verifies that ApeConnection doesn't trip over reusing an OID that's
        # no longer in use.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'Holidays'
            app._setObject(f.id, f, set_owner=0)
            get_transaction().commit()

            f = None  # Forget the reference to folder
            app._delObject('Holidays')
            get_transaction().commit()

            f = Folder()
            f.id = 'Holidays'
            app._setObject(f.id, f, set_owner=0)
            get_transaction().commit()
        finally:
            conn.close()


    def testAutomaticPageTemplateExtension(self):
        text = '<span tal:content="string:Hello">example</span>'
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            template = ZopePageTemplate('template', text)
            app._setObject(template.id, template, set_owner=0)
            get_transaction().commit()

            dir = self.conn._expandPath('/')
            names = os.listdir(dir)
            self.assert_('template.html' in names, names)
            self.assert_('template' not in names, names)
        finally:
            conn.close()


    def testPreserveNamesWithoutExtensions(self):
        # Verifies that FSConnection retains original object names,
        # even though the files might be stored with extensions.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'folder'
            app._setObject(f.id, f, set_owner=0)
            for n in range(3):
                script = PythonScript('script%d' % n)
                script.write('##title=test script\nreturn "OK"')
                f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            conn2 = self.db.open()
            try:
                app = conn2.root()['Application']
                f = app.folder
                for n in range(3):
                    self.assert_(hasattr(f, 'script%d' % n))
                    self.assert_(not hasattr(f, 'script%d.py' % n))
                # white box test: verify the scripts were actually stored
                # with .py extensions.
                dir = self.conn._expandPath('/folder')
                names = os.listdir(dir)
                for n in range(3):
                    self.assert_(('script%d.py' % n) in names, names)
            finally:
                conn2.close()
        finally:
            conn.close()


    def testPreserveNamesWithExtensions(self):
        # Verifies that FSConnection retains original object names
        # even though the object names already have extensions.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'folder'
            app._setObject(f.id, f, set_owner=0)
            for n in range(3):
                script = PythonScript('script%d.py' % n)
                script.write('##title=test script\nreturn "OK"')
                f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            conn2 = self.db.open()
            try:
                app = conn2.root()['Application']
                f = app.folder
                for n in range(3):
                    self.assert_(hasattr(f, 'script%d.py' % n))
                    self.assert_(not hasattr(f, 'script%d' % n))
                # white box test: verify the scripts were actually stored
                # with .py extensions.
                dir = self.conn._expandPath('/folder')
                names = os.listdir(dir)
                for n in range(3):
                    self.assert_(('script%d.py' % n) in names, names)
            finally:
                conn2.close()
        finally:
            conn.close()


    def testNameExtensionConflictDetection(self):
        # Verifies that conflicting names resulting from automatic extensions
        # don't go unnoticed.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'folder'
            app._setObject(f.id, f, set_owner=0)

            # Can't write to 'script0' then 'script0.py'.
            script = PythonScript('script0')
            script.write('##title=test script\nreturn "OK"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            dir = self.conn._expandPath('/folder')
            names = os.listdir(dir)
            self.assert_(('script0.py') in names, names)
            self.assert_(('script0') not in names, names)

            # script0.py already exists, so the transaction will fail.
            script = PythonScript('script0.py')
            script.write('##title=test script\nreturn "Hello, world!"')
            f._setObject(script.id, script, set_owner=0)
            self.assertRaises(ConflictError, get_transaction().commit)
        finally:
            conn.close()


    def testNonConflictingNameExtensions1(self):
        # Verifies that FSConnection can write to 'script0.py' then 'script0'
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'folder'
            app._setObject(f.id, f, set_owner=0)

            # It's OK to write to 'script0.py' then 'script0'.
            script = PythonScript('script0.py')
            script.write('##title=test script\nreturn "OK"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            script = PythonScript('script0')
            script.write('##title=test script\nreturn "Hello, world!"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            dir = self.conn._expandPath('/folder')
            names = os.listdir(dir)
            self.assert_(('script0.py') in names, names)
            self.assert_(('script0') in names, names)

            conn2 = self.db.open()
            try:
                app = conn2.root()['Application']
                f = app.folder
                self.assertEqual(f['script0.py'](), 'OK')
                self.assertEqual(f['script0'](), 'Hello, world!')
            finally:
                conn2.close()
        finally:
            conn.close()


    def testNonConflictingNameExtensions2(self):
        # Verifies that FSConnection can write to 'script0.py' and 'script0'
        # at the same time
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'folder'
            app._setObject(f.id, f, set_owner=0)

            # It's OK to write to 'script0.py' then 'script0'.
            script = PythonScript('script0.py')
            script.write('##title=test script\nreturn "OK"')
            f._setObject(script.id, script, set_owner=0)
            script = PythonScript('script0')
            script.write('##title=test script\nreturn "Hello, world!"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            conn2 = self.db.open()
            try:
                app = conn2.root()['Application']
                f = app.folder
                self.assertEqual(f['script0.py'](), 'OK')
                self.assertEqual(f['script0'](), 'Hello, world!')
            finally:
                conn2.close()
        finally:
            conn.close()


    def testNonConflictingNameExtensions3(self):
        # Verifies that FSConnection can write to 'script0.py'
        # then 'script0.dtml', then 'script0'.
        # Then verifies that removal of items works correctly.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = 'folder'
            app._setObject(f.id, f, set_owner=0)

            script = PythonScript('script0.py')
            script.write('##title=test script\nreturn "OK"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            script = PythonScript('script0.dtml')
            script.write('##title=test script\nreturn "No DTML here"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            script = PythonScript('script0')
            script.write('##title=test script\nreturn "Hello, world!"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()

            dir = self.conn._expandPath('/folder')
            names = os.listdir(dir)
            self.assert_(('script0.py') in names, names)
            self.assert_(('script0.dtml') in names, names)
            self.assert_(('script0') in names, names)

            conn2 = self.db.open()
            try:
                app2 = conn2.root()['Application']
                f2 = app2.folder
                self.assertEqual(f2['script0.py'](), 'OK')
                self.assertEqual(f2['script0.dtml'](), 'No DTML here')
                self.assertEqual(f2['script0'](), 'Hello, world!')
            finally:
                conn2.close()

            f._delObject('script0.py')
            get_transaction().commit()
            names = os.listdir(dir)
            self.assert_(('script0.py') not in names, names)
            self.assert_(('script0.dtml') in names, names)
            self.assert_(('script0') in names, names)

            f._delObject('script0')
            get_transaction().commit()
            names = os.listdir(dir)
            self.assert_(('script0.py') not in names, names)
            self.assert_(('script0.dtml') in names, names)
            self.assert_(('script0') not in names, names)

            script = PythonScript('script0')
            script.write('##title=test script\nreturn "Hello, world!"')
            f._setObject(script.id, script, set_owner=0)
            get_transaction().commit()
            names = os.listdir(dir)
            self.assert_(('script0.py') not in names, names)
            self.assert_(('script0.dtml') in names, names)
            self.assert_(('script0') in names, names)

            f._delObject('script0.dtml')
            get_transaction().commit()
            names = os.listdir(dir)
            self.assert_(('script0.py') not in names, names)
            self.assert_(('script0.dtml') not in names, names)
            self.assert_(('script0') in names, names)
        finally:
            conn.close()


    def testImageExtension(self):
        # Verify that a new image is stored with the correct extension.
        path = os.path.join(os.path.dirname(__file__), 'correct.png')
        f = open(path, 'rb')
        try:
            data = f.read()
        finally:
            f.close()
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            manage_addImage(app, 'image', StringIO(data))
            get_transaction().commit()

            self.assertEqual(app.image.data, data)
            conn2 = self.db.open()
            try:
                app = conn2.root()['Application']
                self.assertEqual(app.image.data, data)
            finally:
                conn2.close()

            dir = self.conn._expandPath('/')
            names = os.listdir(dir)
            self.assert_(('image.png') in names, names)
            self.assert_(('image') not in names, names)
        finally:
            conn.close()


    def testCorrectedFileExtension(self):
        # Verify that certain content_types use the correct filename
        # extension.
        data = 'Hello, world!'
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            manage_addFile(app, 'hello', StringIO(data),
                           content_type='text/plain')
            manage_addFile(app, 'world.dat', StringIO(data),
                           content_type='text/plain')
            manage_addFile(app, 'binary_file', StringIO(data),
                           content_type='application/octet-stream')
            get_transaction().commit()

            dir = self.conn._expandPath('/')
            names = os.listdir(dir)
            self.assert_(('hello.txt') in names, names)
            self.assert_(('world.dat') in names, names)
            self.assert_(('hello') not in names, names)
            self.assert_(('binary_file') in names, names)
        finally:
            conn.close()


    def testGuessTypeBasedOnExtension(self):
        # Verify Zope chooses the right object type for
        # a new object.
        # White box test.
        dir = self.conn._expandPath('/')
        f = open(os.path.join(dir, 'test.py'), 'wt')
        f.write('return "Ok!"')
        f.close()
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            self.assert_(hasattr(app, 'test.py'))
            self.assert_(isinstance(app['test.py'], PythonScript))
            self.assertEqual(app['test.py'](), 'Ok!')
        finally:
            conn.close()


    def testGuessTypeWithChoppedExtension(self):
        # Verify that even though the extension gets stripped off
        # in Zope, Zope still sees the object as it should.
        # White box test.
        dir = self.conn._expandPath('/')
        f = open(os.path.join(dir, 'test.py'), 'wt')
        f.write('return "Ok!"')
        f.close()
        
        f = open(os.path.join(dir, self.conn.metadata_prefix
                              + 'properties'), 'wt')
        f.write('[object_names]\ntest\n')
        f.close()
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            self.assert_(hasattr(app, 'test'))
            self.assert_(isinstance(app.test, PythonScript))
            self.assertEqual(app.test(), 'Ok!')
        finally:
            conn.close()


    def testFallbackToFile(self):
        # Verify Zope uses a File object for unrecognized files on
        # the filesystem.  White box test.
        data = 'data goes here'
        dir = self.conn._expandPath('/')
        f = open(os.path.join(dir, 'test'), 'wt')
        f.write(data)
        f.close()
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            self.assert_(hasattr(app, 'test'))
            self.assert_(isinstance(app['test'], File))
            self.assertEqual(str(app['test']), data)
        finally:
            conn.close()


    def testDefaultPropertySchema(self):
        # Verify Zope uses the default property schema when no properties
        # are set.
        dir = self.conn._expandPath('/')
        os.mkdir(os.path.join(dir, 'test'))
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            self.assert_(hasattr(app, 'test'))
            self.assert_(isinstance(app['test'], Folder))
            self.assertEqual(app['test'].title, '')
            props = app['test']._properties
            for p in props:
                if p['id'] == 'title':
                    break
            else:
                self.fail('No title property found')
        finally:
            conn.close()


    def testRemainderStorage(self):
        # Verify that FSConnection puts the remainder in a separate file
        conn = self.db.open()
        try:
            content = 'tacked_on_data'
            app = conn.root()['Application']
            app._stowaway = content
            get_transaction().commit()

            # Verify the ability to load it
            conn2 = self.db.open()
            try:
                app2 = conn2.root()['Application']
                self.assertEqual(app2._stowaway, content)
            finally:
                conn2.close()

            # Verify the stowaway is in the remainder file.
            dir = self.conn._expandPath('/')
            f = open(os.path.join(dir, self.conn.metadata_prefix
                                  + 'remainder'), 'rb')
            data = f.read()
            f.close()
            self.assert_(data.find(content) >= 0)
        finally:
            conn.close()


    def testDottedNames(self):
        # FSConnection should allow dotted names that don't look like
        # property or remainder files.
        conn = self.db.open()
        try:
            app = conn.root()['Application']
            f = Folder()
            f.id = '.Holidays'
            app._setObject(f.id, f, set_owner=0)
            f2 = Folder()
            f2.id = '.Holidays.properties.dat'
            app._setObject(f2.id, f2, set_owner=0)
            get_transaction().commit()
        finally:
            conn.close()


    def testGuessFileContentType(self):
        # Verify that file content type guessing happens.
        data = '<html><body>Cool stuff</body></html>'
        dir = self.conn._expandPath('/')
        f = open(os.path.join(dir, 'testobj'), 'wt')
        f.write(data)
        f.close()

        conn = self.db.open()
        try:
            app = conn.root()['Application']
            self.assert_(hasattr(app, 'testobj'))
            self.assertEqual(app.testobj.content_type, 'text/html')
        finally:
            conn.close()



class Zope2FSUnderscoreTests (Zope2FSTests):
    
    def _createConnections(self, path):
        conn = FSConnection(path, metadata_prefix='_')
        return {'fs': conn}


if __name__ == '__main__':
    unittest.main()

