##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Extension of the ZODB Connection class

$Id: connection.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import sys
from time import time
from types import StringType, TupleType
from cStringIO import StringIO
from cPickle import Unpickler, Pickler

from Acquisition import aq_base
from ZODB import Persistent
from ZODB.POSException \
     import ConflictError, ReadConflictError, InvalidObjectReference, \
     StorageError
from ZODB.Connection import Connection
from ZODB.ConflictResolution import ResolvedSerial
from zLOG import LOG, ERROR

from consts import HASH0, DEBUG
from apelib.core.io import ObjectSystemIO, ClassifiedState
from apelib.core.interfaces import IKeyedObjectSystem


class ApeConnection (Connection):
    """Mapper-driven Connection

    Uses a mapper to serialize the state of objects before
    pickling, and to deserialize objects based on the pickled
    state.

    The mapper might, for example, serialize all objects as
    tabular records.
    """
    _osio = None
    _scan_ctl = None

    __implements__ = (IKeyedObjectSystem,
                      getattr(Connection, '__implements__', ()))


    def _setDB(self, odb):
        Connection._setDB(self, odb)
        if odb._scan_ctl is not None:
            ctl = self._scan_ctl
            if ctl is None:
                self._scan_ctl = ctl = odb._scan_ctl.newConnection()
            if ctl.ready():
                ctl.setOIDs(self._cache.cache_data.keys())
                # If there were any invalidations, process them now.
                if self._invalidated:
                    self._flush_invalidations()


    def getObjectSystemIO(self):
        osio = self._osio
        if osio is None:
            root_mapper = self._db._mapper_resource.access(self)
            osio = ObjectSystemIO(root_mapper, self)
            self._osio = osio
        return osio


    def close(self):
        db = self._db
        try:
            Connection.close(self)
        finally:
            if db is not None and self._osio is not None:
                self._osio = None
                db._mapper_resource.release(self)


    def __getitem__(self, oid, tt=type(())):
        obj = self._cache.get(oid, None)
        if obj is not None:
            return obj

        __traceback_info__ = (oid)
        p, serial = self._storage.load(oid, self._version)
        __traceback_info__ = (oid, p)
        file=StringIO(p)
        unpickler=Unpickler(file)
        # unpickler.persistent_load=self._persistent_load

        try:
            classified_state = unpickler.load()
        except:
            raise "Could not load oid %s, pickled data in traceback info may\
            contain clues" % (oid)
        osio = self.getObjectSystemIO()
        object = osio.newObject(classified_state)
        assert object is not None

        object._p_oid=oid
        object._p_jar=self
        object._p_changed=None

        self._cache[oid] = object
        if oid=='\0\0\0\0\0\0\0\0':
            self._root_=object # keep a ref
        return object


    def _persistent_load(self, oid, hints=None):

        __traceback_info__=oid

        obj = self._cache.get(oid, None)
        if obj is not None:
            return obj

        if hints:
            mapper_names = hints.get('mapper_names')
            if mapper_names is not None:
                classified_state = ClassifiedState(None, None, mapper_names)
                osio = self.getObjectSystemIO()
                object = osio.newObject(classified_state)
                if object is not None:
                    object._p_oid=oid
                    object._p_jar=self
                    object._p_changed=None
                    self._cache[oid] = object
                    return object

        # We don't have enough info for fast loading.  Load the whole object.
        return self[oid]


    def mayBegin(self, transaction):
        if hasattr(self, '_begun') and not self._begun:
            self._storage.tpc_begin(transaction)
            self._begun = 1


    def commit(self, object, transaction):
        if object is self:
            self.mayBegin(transaction)
            # We registered ourself.  Execute a commit action, if any.
            if self._Connection__onCommitActions is not None:
                method_name, args, kw = \
                             self._Connection__onCommitActions.pop(0)
                apply(getattr(self, method_name), (transaction,) + args, kw)
            return
        oid=object._p_oid
        assert oid != 'unmanaged', repr(object)
        #invalid=self._invalidated.get
        invalid = self._invalid

        modified = getattr(self, '_modified', None)
        if modified is None:
            modified = self._invalidating
        
        if oid is None or object._p_jar is not self:
            # new object
            oid = self.new_oid()
            object._p_jar=self
            object._p_oid=oid
            self._creating.append(oid)

        elif object._p_changed:
            if (
                (invalid(oid) and not hasattr(object, '_p_resolveConflict'))
                or
                invalid(None)
                ):
                raise ConflictError(object=object)
            modified.append(oid)

        else:
            # Nothing to do
            return

        self.mayBegin(transaction)

        stack=[object]

        file=StringIO()
        seek=file.seek
        pickler=Pickler(file,1)
        # SDH: external references are computed in a different way.
        # pickler.persistent_id=new_persistent_id(self, stack.append)
        dbstore=self._storage.store
        file=file.getvalue
        cache=self._cache
        get=cache.get
        dump=pickler.dump
        clear_memo=pickler.clear_memo


        version=self._version

        while stack:
            object=stack[-1]
            del stack[-1]
            oid=object._p_oid
            assert oid != 'unmanaged', repr(object)
            serial = self.getSerial(object)
            if serial == HASH0:
                # new object
                self._creating.append(oid)
            else:
                #XXX We should never get here
                # SDH: Actually it looks like we should, but only
                # for the first object on the stack.
                if (
                    (invalid(oid) and
                     not hasattr(object, '_p_resolveConflict'))
                    or
                    invalid(None)
                    ):
                    raise ConflictError(object=object)
                modified.append(oid)

            # SDH: hook in the serializer.
            # state=object.__getstate__()
            oid_encoder = self._db._oid_encoder
            keychain = oid_encoder.decode(oid)
            osio = self.getObjectSystemIO()
            event, classified_state = osio.serialize(keychain, object)
            ext_refs = event.getExternalRefs()
            if ext_refs:
                for (ext_keychain, ext_ref) in ext_refs:
                    if self.getSerial(ext_ref) == HASH0:
                        ext_oid = oid_encoder.encode(ext_keychain)
                        if ext_ref._p_jar is not None:
                            if ext_ref._p_jar is not self:
                                raise InvalidObjectReference, (
                                    "Can't refer from %s in %s to %s in %s"
                                    % (repr(object), repr(self), repr(ext_ref),
                                       repr(ext_ref._p_jar)))
                        else:
                            ext_ref._p_jar = self
                        if ext_ref._p_oid:
                            if ext_ref._p_oid != ext_oid:
                                raise StorageError('Conflicting OIDs')
                        else:
                            ext_ref._p_oid = ext_oid
                        stack.append(ext_ref)

            unmanaged = event.getUnmanagedPersistentObjects()
            if unmanaged:
                self.handleUnmanaged(object, unmanaged)

            seek(0)
            clear_memo()
            dump(classified_state)
            p=file(1)
            s=dbstore(oid,serial,p,version,transaction)
            self._store_count = self._store_count + 1

            # Put the object in the cache before handling the
            # response, just in case the response contains the
            # serial number for a newly created object
            try: cache[oid] = object
            except ValueError:
                # "Cannot re-register an object under a different
                # oid".  This can happen when the user is working on
                # the filesystem and creates an object with an ID that
                # was used recently.  Try to fix it by minimizing
                # the cache and trying again.
                cache.minimize()
                cache[oid] = object
            except:
                if aq_base(object) is not object:
                    # Yuck, someone tried to store a wrapper.  Try to
                    # cache it unwrapped.
                    cache[oid] = aq_base(object)
                else:
                    raise

            self._handle_serial(s, oid)


    def setstate(self, object):
        oid=object._p_oid

        if self._storage is None:
            msg = ("Shouldn't load state for %s "
                   "when the connection is closed" % `oid`)
            LOG('ZODB', ERROR, msg)
            raise RuntimeError(msg)

        try:
            p, serial = self._storage.load(oid, self._version)
            self._load_count = self._load_count + 1

            # XXX this is quite conservative!
            # We need, however, to avoid reading data from a transaction
            # that committed after the current "session" started, as
            # that might lead to mixing of cached data from earlier
            # transactions and new inconsistent data.
            #
            # Note that we (carefully) wait until after we call the
            # storage to make sure that we don't miss an invaildation
            # notifications between the time we check and the time we
            # read.
            #invalid = self._invalidated.get
            invalid = self._invalid
            if invalid(oid) or invalid(None):
                if not hasattr(object.__class__, '_p_independent'):
                    get_transaction().register(self)
                    raise ReadConflictError(object=object)
                invalid=1
            else:
                invalid=0

            file=StringIO(p)
            unpickler=Unpickler(file)
            # SDH: external references are reassembled elsewhere.
            # unpickler.persistent_load=self._persistent_load
            classified_state = unpickler.load()

            # SDH: Let the object mapper do the state setting.
            # if hasattr(object, '__setstate__'):
            #     object.__setstate__(state)
            # else:
            #     d=object.__dict__
            #     for k,v in state.items(): d[k]=v
            keychain = self._db._oid_encoder.decode(oid)
            osio = self.getObjectSystemIO()
            event = osio.deserialize(keychain, object, classified_state)

            unmanaged = event.getUnmanagedPersistentObjects()
            if unmanaged:
                self.handleUnmanaged(object, unmanaged)

            self.setSerial(object, serial)

            if invalid:
                if object._p_independent():
                    try: del self._invalidated[oid]
                    except KeyError: pass
                else:
                    get_transaction().register(self)
                    raise ConflictError(object=object)

        except ConflictError:
            raise
        except:
            LOG('ZODB',ERROR, "Couldn't load state for %s" % `oid`,
                error=sys.exc_info())
            raise


    def register(self, object):
        """Register an object with the appropriate transaction manager.
        """
        assert object._p_jar is self
        if object._p_oid is not None:
            get_transaction().register(object)
        # else someone is trying to trick ZODB into registering an
        # object with no OID.  OFS.Image.File._read_data() does this.
        # Since ApeConnection really needs meaningful OIDs, just ignore
        # the attempt.


    def __repr__(self):
        if self._version:
            ver = ' (in version %s)' % `self._version`
        else:
            ver = ''
        return '<%s at %08x%s>' % (self.__class__.__name__, id(self), ver)


    def handleUnmanaged(self, object, unmanaged):
        for o in unmanaged:
            if isinstance(o, Persistent):
                if o._p_jar is None:
                    o._p_oid = 'unmanaged'
                    o._p_jar = UnmanagedJar(self, object._p_oid)
                else:
                    # Turn off the "changed" flag
                    o._p_changed = 0


    # IKeyedObjectSystem implementation

    def getObject(self, keychain, hints=None):
        oid = self._db._oid_encoder.encode(keychain)
        return self._persistent_load(oid, hints)

    def identifyObject(self, obj):
        oid = obj._p_oid
        if oid is None:
            return None
        if obj._p_jar is not self:
            raise InvalidObjectReference, (
                "Can't refer to %s, located in %s, from %s"
                % (repr(obj), repr(obj._p_jar), repr(self)))
        return self._db._oid_encoder.decode(oid)

    def newKey(self):
        oid = self.new_oid()
        keychain = self._db._oid_encoder.decode(oid)
        return keychain[-1]

    loadStub = getObject  # Deprecated


    def getClass(self, module, name):
        return self._db._classFactory(self, module, name)


    def checkSerials(self):
        """Verifies that all cached objects are in sync with the data.

        This is useful for finding gateways that generate inconsistent
        hashes.
        """
        for oid, ob in self._cache.items():
            if ob._p_changed is not None:
                p, serial = self._storage.load(oid, self._version)
                if serial != self.getSerial(ob):
                    keychain = self._db._oid_encoder.decode(oid)
                    raise StorageError(
                        "Inconsistent serial for keychain %s" % repr(keychain))
    

    def exportFile(self, oid, file=None):
        raise NotImplementedError, 'ZEXP Export not implemented'

    def importFile(self, file, clue='', customImporters=None):
        raise NotImplementedError, 'ZEXP Import not implemented'


    # A note on serials: Serials need to be stored independently of
    # objects because the current Persistent base class uses _p_serial
    # to derive _p_mtime.  Applications like Zope use _p_mtime, but
    # the _p_serial for Ape isn't always a date, so Ape can't use
    # _p_serial to store serials.  Instead, ApeConnection puts them in
    # a _serials dictionary.

    _serials = None
    SERIAL_CLEANUP_THRESHOLD = 1000

    def getSerial(self, ob):
        oid = ob._p_oid
        if oid is None or self._cache.get(oid, None) is not ob:
            return HASH0
        serials = self._serials
        if serials is None:
            return HASH0
        return serials.get(oid, HASH0)

    def setSerial(self, ob, s):
        oid = ob._p_oid
        assert oid is not None
        if s is None:
            s = HASH0
        serials = self._serials
        if serials is None:
            serials = {}
            self._serials = serials
        if not serials.has_key(oid):
            # When the number of recorded serials exceeds the number of
            # cache entries by SERIAL_CLEANUP_THRESHOLD, prune the serials
            # dictionary.
            if (len(serials) >= len(self._cache) +
                self.SERIAL_CLEANUP_THRESHOLD):
                # clean up
                cache_get = self._cache.get
                for oid in serials.keys():
                    ob = cache_get(oid, None)
                    if ob is None or ob._p_changed is None:
                        del serials[oid]
        serials[oid] = s

    def _handle_serial(self, store_return, oid=None, change=1):
        """Handle the returns from store() and tpc_vote() calls."""

        # These calls can return different types depending on whether
        # ZEO is used.  ZEO uses asynchronous returns that may be
        # returned in batches by the ClientStorage.  ZEO1 can also
        # return an exception object and expect that the Connection
        # will raise the exception.

        # When commit_sub() exceutes a store, there is no need to
        # update the _p_changed flag, because the subtransaction
        # tpc_vote() calls already did this.  The change=1 argument
        # exists to allow commit_sub() to avoid setting the flag
        # again.
        if not store_return:
            return
        if isinstance(store_return, StringType):
            assert oid is not None
            serial = store_return
            obj = self._cache.get(oid, None)
            if obj is None:
                return
            if serial == ResolvedSerial:
                del obj._p_changed
            else:
                if change:
                    obj._p_changed = 0
                #obj._p_serial = serial
                self.setSerial(obj, serial)
        else:
            for oid, serial in store_return:
                if not isinstance(serial, StringType):
                    raise serial
                obj = self._cache.get(oid, None)
                if obj is None:
                    continue
                if serial == ResolvedSerial:
                    del obj._p_changed
                else:
                    if change:
                        obj._p_changed = 0
                    #obj._p_serial = serial
                    self.setSerial(obj, serial)



class UnmanagedJar:
    """Special jar for unmanaged persistent objects.

    There is one such jar for each unmanaged persistent object.  All
    it does is notify the managed persistent object of changes.

    Note that unmanaged persistent objects should never be ghosted!
    Instead, when the managed persistent object gets ghosted, it
    usually removes the last reference to the unmanaged object, which
    is then deallocated.
    """

    def __init__(self, real_jar, real_oid):
        self.real_jar = real_jar
        self.real_oid = real_oid

    def register(self, ob):
        o = self.real_jar[self.real_oid]
        o._p_changed = 1

    def modifiedInVersion(self, oid):
        # XXX PersistentExtra wants this
        return ''
