##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Extension of the ZODB DB class

$Id: db.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

from ZODB.DB import DB, Transaction, cPickle, cStringIO, allocate_lock

from apelib.core.interfaces import IMapper
from apelib.core.exceptions import ConfigurationError

from connection import ApeConnection
from storage import ApeStorage
from oidencoder import OIDEncoder
from resource import StaticResource
from interfaces import IResourceAccess, IOIDEncoder


def callMapperFactory(factory, kw):
    """Returns (mapper, tpc_conns) given the name of a factory and arguments.
    """
    pos = factory.rfind('.')
    if pos < 0:
        raise ConfigurationError(
            'factory must be a string containing <module>.<name>')
    module = factory[:pos]
    name = factory[pos + 1:]
    m = __import__(module, {}, {}, (name,))
    f = getattr(m, name)
    return f(**kw)


class ApeDB (DB):
    """Mapper-driven Database
    """

    klass = ApeConnection

    # SDH: some extra args.
    def __init__(self, storage,
                 mapper_resource=None,
                 factory=None,
                 oid_encoder=None,
                 scan_interval=10,
                 pool_size=7,
                 cache_size=400,
                 cache_deactivate_after=60,
                 version_pool_size=3,
                 version_cache_size=100,
                 version_cache_deactivate_after=10,
                 **kw
                 ):
        """Create an object database.
        """
        if mapper_resource is None:
            if factory is not None:
                # Use a mapper factory
                mapper, connections = callMapperFactory(factory, kw)
                assert IMapper.isImplementedBy(mapper)
                mapper_resource = StaticResource(mapper)
            else:
                if kw:
                    raise ConfigurationError('Extra keyword args: %s' % kw)
                if isinstance(storage, ApeStorage):
                    # Use the mapper from the storage
                    mapper_resource = storage.getMapperResource()
                else:
                    raise ConfigurationError('No mapper or factory specified')
        else:
            # mapper_resource was specified
            if kw:
                raise ConfigurationError('Extra keyword args: %s' % kw)
            assert IResourceAccess.isImplementedBy(mapper_resource)
            assert factory is None
        
        # Allocate locks:
        l=allocate_lock()
        self._a=l.acquire
        self._r=l.release

        # Setup connection pools and cache info
        self._pools={},[]
        self._temps=[]
        self._pool_size=pool_size
        self._cache_size=cache_size
        self._cache_deactivate_after = cache_deactivate_after
        self._version_pool_size=version_pool_size
        self._version_cache_size=version_cache_size
        self._version_cache_deactivate_after = version_cache_deactivate_after

        self._miv_cache={}

        # Setup storage
        self._storage=storage
        storage.registerDB(self, None)
        if not hasattr(storage,'tpc_vote'): storage.tpc_vote=lambda *args: None

        if oid_encoder is None:
            oid_encoder = OIDEncoder()
        else:
            assert IOIDEncoder.isImplementedBy(oid_encoder)
        self._oid_encoder = oid_encoder
        self._mapper_resource = mapper_resource
        scan_interval = int(scan_interval)
        if scan_interval > 0:
            from scanner import ScanControl
            ctl = ScanControl(db=self, scan_interval=scan_interval)
            self._scan_ctl = ctl
            ctl.scanner.setStorage(storage)
            storage.setScanner(ctl.scanner)
        else:
            self._scan_ctl = None

        # Pass through methods:
        for m in ('history',
                  'supportsUndo', 'supportsVersions', 'undoLog',
                  'versionEmpty', 'versions'):
            setattr(self, m, getattr(storage, m))

        if hasattr(storage, 'undoInfo'):
            self.undoInfo=storage.undoInfo

