##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Cache scanner.

Keeps a cache up to date by scanning for changes.

$Id: scanner.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

from thread import allocate_lock
from time import time

from BTrees.OOBTree import OOBTree, OOSet, difference
from BTrees.IOBTree import IOBTree
from zLOG import LOG, DEBUG

# FUTURE_TIMEOUT defines how long to keep source information regarding
# OIDs that might be used soon.
FUTURE_TIMEOUT = 10 * 60


class ScanControl:

    def __init__(self, db=None, scan_interval=10):
        self.db = db
        self.next_conn_id = 1
        self.conn_oids = IOBTree()   # IOBTree({ conn_id -> OOSet([oid]) } })
        self.oids = OOSet()          # OOSet([oid])
        self.scanner = Scanner()
        self.lock = allocate_lock()
        self.scan_interval = scan_interval
        self.next_scan = time() + scan_interval


    def newConnection(self):
        self.lock.acquire()
        try:
            conn_id = self.next_conn_id
            self.next_conn_id = conn_id + 1
            return ConnectionScanControl(self, conn_id)
        finally:
            self.lock.release()


    def setConnectionOIDs(self, conn_id, oids):
        changed = 0
        new_oids = OOSet()
        self.lock.acquire()
        try:
            if oids:
                self.conn_oids[conn_id] = OOSet(oids)
            else:
                if self.conn_oids.has_key(conn_id):
                    del self.conn_oids[conn_id]
            for set in self.conn_oids.values():
                new_oids.update(set)
            if self.oids != new_oids:
                self.oids = new_oids
                changed = 1
        finally:
            self.lock.release()
        if changed:
            self.scanner.setOIDs(new_oids)
        self.mayScan()


    def mayScan(self):
        now = time()
        if now >= self.next_scan:
            self.next_scan = now + self.scan_interval
            LOG('Ape', DEBUG, 'Scanning %d objects.' % len(self.oids))
            inv = self.scanner.scan()
            self.scanner.pruneFuture()
            LOG('Ape', DEBUG,
                'Finished scanning. %d objects changed.' % len(inv))
            if inv:
                d = {}
                for oid in inv:
                    d[oid] = 1
                if self.db is not None:
                    self.db.invalidate(d)
                else:
                    LOG('Ape', DEBUG, "No database set, so can't invalidate!")


class ConnectionScanControl:

    def __init__(self, ctl, conn_id):
        self.ctl = ctl
        self.conn_id = conn_id
        self.next_update = 0

    def ready(self):
        now = time()
        if now >= self.next_update:
            self.next_update = now + self.ctl.scan_interval
            return 1
        return 0

    def setOIDs(self, oids):
        self.ctl.setConnectionOIDs(self.conn_id, oids)


class Scanner:

    def __init__(self):
        self.current = OOBTree()  # OOBTree({ oid -> {source->state} })
        self.future = {}          # { oid -> ({source->state}, atime) }
        self.uncommitted = {}     # { tid -> {oid->{source->state}} }
        self.lock = allocate_lock()
        self.storage = None

    def setStorage(self, s):
        # This is needed for calling storage.getSources().
        self.storage = s

    def setOIDs(self, oids):
        new_sources = {}  # { oid -> sourcedict }
        self.lock.acquire()
        try:
            removed = difference(self.current, oids)
            for oid in removed.keys():
                del self.current[oid]
            added = difference(oids, self.current)
            for oid in added.keys():
                if self.future.has_key(oid):
                    # Source info for this OID was provided earlier.
                    sources, atime = self.future[oid]
                    del self.future[oid]
                    self.current[oid] = sources
                else:
                    new_sources[oid] = None
        finally:
            self.lock.release()
        if new_sources:
            # Load source info the slow way.
            if self.storage is not None:
                LOG('Ape', DEBUG, 'Getting sources for %d oids.'
                    % len(new_sources))
                for oid in new_sources.keys():
                    new_sources[oid] = self.storage.getSources(oid)
            else:
                LOG('Ape', DEBUG, "Can't get sources for %d oids. "
                    "Assuming no sources!" % len(new_sources))
                # This will cause the scanner to miss changes, but
                # since no storage is known, there is little we can
                # do.
                for oid in new_sources.keys():
                    new_sources[oid] = {}
            self.lock.acquire()
            try:
                for oid, sources in new_sources.items():
                    if not self.current.has_key(oid):
                        self.current[oid] = sources
                    # else something else added the source info
                    # while self.lock was released.
            finally:
                self.lock.release()


    def setSources(self, oid, sources):
        if sources is None:
            sources = {}
        self.lock.acquire()
        try:
            if self.current.has_key(oid):
                # This OID is known to be in use.
                self.current[oid] = sources
            else:
                # This OID might be useful soon.
                self.future[oid] = (sources, time())
        finally:
            self.lock.release()


    def setUncommittedSources(self, tid, oid, sources):
        self.lock.acquire()
        try:
            t = self.uncommitted.setdefault(tid, {})
            t[oid] = sources
        finally:
            self.lock.release()


    def scan(self):
        to_scan = {}        # { repo -> { source -> state } }
        to_invalidate = {}  # { oid -> 1 }
        self.lock.acquire()  # lock because oid_states might be self.current.
        try:
            for oid, sources in self.current.items():
                for source, state in sources.items():
                    repo, location = source
                    to_scan.setdefault(repo, {})[source] = state
        finally:
            self.lock.release()
        changes = {}
        for repo, d in to_scan.items():
            c = repo.freshen(d)
            if c:
                changes.update(c)
        if changes:
            # Something changed.  Map the changes back to oids and
            # update self.current.
            self.lock.acquire()
            try:
                for oid, sources in self.current.items():
                    for source, state in sources.items():
                        if changes.has_key(source):
                            to_invalidate[oid] = 1
                            sources[source] = changes[source]
            finally:
                self.lock.release()
        return to_invalidate.keys()


    def pruneFuture(self):
        if self.future:
            self.lock.acquire()
            try:
                # OIDs older than some timeout will probably never be loaded.
                cutoff = time() - FUTURE_TIMEOUT
                for oid, (sources, atime) in self.future.items():
                    if atime < cutoff:
                        del self.future[oid]
            finally:
                self.lock.release()
            LOG('Ape', DEBUG,
                'Future sources cache size: %d objects.' % len(self.future))


    def afterCommit(self, tid):
        self.lock.acquire()
        try:
            if not self.uncommitted.has_key(tid):
                return
            t = self.uncommitted[tid]
            del self.uncommitted[tid]
        finally:
            self.lock.release()
        # Update the sources with new states for the committed OIDs.
        to_scan = {}        # { repo -> { source -> state } }
        for oid, sources in t.items():
            for source, state in sources.items():
                repo, location = source
                to_scan.setdefault(repo, {})[source] = state
        changes = {}
        for repo, d in to_scan.items():
            c = repo.freshen(d)
            if c:
                changes.update(c)
        for oid, sources in t.items():
            new_sources = {}
            for source, state in sources.items():
                state = changes.get(source, state)
                new_sources[source] = state
            self.setSources(oid, new_sources)


    def afterAbort(self, tid):
        self.lock.acquire()
        try:
            if self.uncommitted.has_key(tid):
                del self.uncommitted[tid]
        finally:
            self.lock.release()

