##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Serializers specific to ZODB3.

$Id: serializers.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import os
from cStringIO import StringIO
from cPickle import Pickler, Unpickler, UnpickleableError
import time
from types import DictType

from Persistence import Persistent, PersistentMapping
from ZODB.TimeStamp import TimeStamp

from apelib.core.interfaces \
     import ISerializer, IFullSerializationEvent, IFullDeserializationEvent
from apelib.core.events import SerializationEvent, DeserializationEvent
from apelib.core.exceptions import SerializationError
from apelib.core.schemas import RowSequenceSchema, FieldSchema


class BasicPersistentMapping:
    """Basic PersistentMapping (de)serializer

    This version assumes the PM maps string keys to object references.
    """
    __implements__ = ISerializer

    schema = RowSequenceSchema()
    schema.addField('key', 'string', 1)
    schema.addField('keychain', 'keychain')

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return isinstance(object, PersistentMapping)

    def serialize(self, obj, event):
        assert self.canSerialize(obj)
        res = []
        for key, value in obj.items():
            keychain = event.identifyObject(value)
            if keychain is None:
                keychain = event.makeKeychain(key, 1)
            event.notifySerializedRef(key, value, 0, keychain)
            res.append((key, keychain))
        event.ignoreAttribute('data')
        event.ignoreAttribute('_container')
        return res

    def deserialize(self, obj, event, state):
        assert self.canSerialize(obj)
        data = {}
        for (key, keychain) in state:
            value = event.dereference(key, keychain)
            data[key] = value
        obj.__init__(data)


class FixedPersistentMapping:
    """Unchanging persistent mapping.

    Generally used for a ZODB root object.
    """

    __implements__ = ISerializer

    def __init__(self):
        # map: { name -> (keychain, mapper) }
        self.map = {}

    def add(self, name, keychain, mapper_names=None):
        self.map[name] = (keychain, mapper_names)

    def getSchema(self):
        return None  # No storage

    def canSerialize(self, object):
        return isinstance(object, PersistentMapping)

    def serialize(self, object, event):
        names = object.keys()
        names.sort()
        expected = self.map.keys()
        expected.sort()
        assert names == expected, '%s != %s' % (names, expected)

        for name in names:
            keychain, mapper_names = self.map[name]
            subob = object[name]
            event.notifySerializedRef(name, subob, 0, keychain)

        # One of the two will work. ;-)
        event.ignoreAttribute('data')
        event.ignoreAttribute('_container')


    def deserialize(self, object, event, state):
        assert state is None
        data = {}
        for name, (keychain, mapper_names) in self.map.items():
            subob = event.dereference(name, keychain,
                                      {'mapper_names': mapper_names})
            data[name] = subob
        # The PersistentMapping doesn't have its data or _container
        # attribute yet, and we don't know what its name should be
        # since PersistentMapping's internal structure is not fixed.
        # So call the PersistentMapping's constructor.
        object.__init__(data)


class RollCall:
    """Helps ensure all parts of an object get serialized.

    Designed for debugging purposes.
    """
    __implements__ = ISerializer

    def getSchema(self):
        return None  # No storage

    def canSerialize(self, object):
        return 1

    def serialize(self, object, event):
        assert IFullSerializationEvent.isImplementedBy(event)
        attrs = event.getSerializedAttributeNames()
        attrs_map = {}
        for attr in attrs:
            attrs_map[attr] = 1
        missed = []
        for k in object.__dict__.keys():
            if not k.startswith('_v_') and not attrs_map.has_key(k):
                missed.append(repr(k))
        if missed:
            raise SerializationError(
                'Attribute(s) %s of object at %s not serialized' %
                (', '.join(missed), repr(event.getKeychain())))
        return None

    def deserialize(self, object, event, state):
        assert state is None


class RemainingState:
    """(De)serializes the remaining state of a Persistent object"""

    __implements__ = ISerializer

    schema = FieldSchema('data', 'string')

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        try:
            return isinstance(object, Persistent)
        except TypeError:
            # XXX Python 2.1 thinks Persistent is not a class
            return 0


    def serialize(self, object, event):
        assert IFullSerializationEvent.isImplementedBy(event)
        assert isinstance(object, Persistent)

        # Allow pickling of cyclic references to the object.
        event.notifySerialized('self', object, 0)

        # Ignore previously serialized attributes
        state = object.__dict__.copy()
        for key in state.keys():
            if key.startswith('_v_'):
                del state[key]
        for attrname in event.getSerializedAttributeNames():
            if state.has_key(attrname):
                del state[attrname]
        if not state:
            # No data needs to be stored
            return ''

        outfile = StringIO()
        p = Pickler(outfile)
        unmanaged = []

        def persistent_id(ob, getInternalRef=event.getInternalRef,
                          unmanaged=unmanaged):
            ref = getInternalRef(ob)
            if ref is None:
                if hasattr(ob, '_p_oid'):
                    # Persistent objects that end up in the remainder
                    # are unmanaged.  Tell ZODB about them so that
                    # ZODB can deal with them specially.
                    unmanaged.append(ob)
            return ref

        p.persistent_id = persistent_id
        try:
            p.dump(state)
        except UnpickleableError, exc:
            # Try to reveal which attribute is unpickleable.
            attrname = None
            attrvalue = None
            for key, value in state.items():
                del unmanaged[:]
                outfile.seek(0)
                outfile.truncate()
                p = Pickler(outfile)
                p.persistent_id = persistent_id
                try:
                    p.dump(value)
                except UnpickleableError:
                    attrname = key
                    attrvalue = value
                    break
            if attrname is not None:
                # Provide a more informative exception.
                if os.environ.get('APE_TRACE_UNPICKLEABLE'):
                    # Provide an opportunity to examine
                    # the "attrvalue" attribute.
                    import pdb
                    pdb.set_trace()
                raise RuntimeError(
                    'Unable to pickle the %s attribute, %s, '
                    'of %s at %s.  %s.' % (
                    repr(attrname), repr(attrvalue), repr(object),
                    repr(event.getKeychain()), str(exc)))
            else:
                # Couldn't help.
                raise

        p.dump(unmanaged)
        s = outfile.getvalue()
        event.addUnmanagedPersistentObjects(unmanaged)
        return s


    def deserialize(self, object, event, state):
        assert IFullDeserializationEvent.isImplementedBy(event)
        assert isinstance(object, Persistent)

        # Set up to recover cyclic references to the object.
        event.notifyDeserialized('self', object)

        if state:
            infile = StringIO(state)
            u = Unpickler(infile)
            u.persistent_load = event.loadInternalRef
            s = u.load()
            object.__dict__.update(s)
            try:
                unmanaged = u.load()
            except EOFError:
                # old pickle with no list of unmanaged objects
                pass
            else:
                event.addUnmanagedPersistentObjects(unmanaged)


class ModTimeAttribute:
    """Sets the _p_mtime attribute."""

    __implements__ = ISerializer

    schema = FieldSchema('mtime', 'int')

    def getSchema(self):
        return self.schema

    def canSerialize(self, obj):
        try:
            return isinstance(obj, Persistent)
        except TypeError:
            # XXX Python 2.1 thinks Persistent is not a class
            return 0

    def setTime(self, obj, t):
        """Sets the last modification time of a Persistent obj to float t.
        """
        args = time.gmtime(t)[:5] + (t%60,)
        obj._p_serial = repr(TimeStamp(*args))

    def serialize(self, obj, event):
        now = long(time.time())
        if obj._p_changed:
            # Indicate that this object just changed.  Note that the time
            # is a guess.
            self.setTime(obj, now)
        return now

    def deserialize(self, obj, event, state):
        self.setTime(obj, state)

