##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Storage implementation that loads/stores using a mapper.

$Id: storage.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

import md5
from cPickle import Pickler, Unpickler
from cStringIO import StringIO

from ZODB import POSException, BaseStorage

from apelib.core.io import GatewayIO
from apelib.core.exceptions import NoStateFoundError, ConfigurationError
from consts import HASH0, HASH1, DEBUG
from oidencoder import OIDEncoder
from interfaces import IResourceAccess, IOIDEncoder


class ApeStorage(BaseStorage.BaseStorage):

    def __init__(self, mapper_resource, connections,
                 oid_encoder=None, name='', clear_all=0):
        """Initializes an ApeStorage.

        mapper_resource is a resource for loading the mapper.
        connections is a mapping that maps names to ITPCConnections.
        oid_encoder is an IOIDEncoder.
        """
        assert IResourceAccess.isImplementedBy(mapper_resource)
        self._mapper_resource = mapper_resource
        if oid_encoder is None:
            oid_encoder = OIDEncoder()
        else:
            assert IOIDEncoder.isImplementedBy(oid_encoder)
        self._oid_encoder = oid_encoder
        gwio = GatewayIO(mapper_resource.access(self), connections)
        self._gwio = gwio
        self._conn_list = gwio.getConnectionList()
        gwio.openConnections()
        gwio.initDatabases(clear_all)
        names = []
        sort_keys = []
        for c in gwio.getConnectionList():
            names.append(c.getName())
            sort_keys.append(c.sortKey())
        self._sort_key = tuple(sort_keys)
        if not name:
            name = 'ApeStorage: ' + ', '.join(names)
        self._ltid = None
        self._scanner = None
        BaseStorage.BaseStorage.__init__(self, name)

    def setScanner(self, s):
        self._scanner = s

    def __len__(self):
        return 1

    def getSize(self):
        # Stub
        return 1

    def sortKey(self):
        return self._sort_key

    def getMapperResource(self):
        return self._mapper_resource

    def initDatabases(self, clear_all=0):
        self._gwio.initDatabases(clear_all=clear_all)

    def hash64(self, value):
        """Returns an 8-byte hash value.
        """
        h = '%08x' % hash(value)
        if h == HASH0:
            # Avoid the special zero hash.
            h = HASH1
        if DEBUG:
            print '64-bit hash of %r is %r' % (value, h)
        return h

    def load(self, oid, version):
        if version:
            raise POSException.Unsupported, "Versions aren't supported"
        self._lock_acquire()
        try:
            self._mapper_resource.access(self)  # Update mapper
            keychain = self._oid_encoder.decode(oid)
            event, classified_state, hash_value = self._gwio.load(keychain)
            file = StringIO()
            p = Pickler(file)
            p.dump(classified_state)
            data = file.getvalue()
            h = self.hash64(hash_value)
            if DEBUG:
                print 'loaded', `oid`, `h`
            if self._scanner is not None:
                gw = event.getMapper().getGateway()
                sources = gw.getSources(event)
                self._scanner.setSources(oid, sources)
            return data, h
        finally:
            self._lock_release()

    def store(self, oid, h64, data, version, transaction):
        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)

        if version:
            raise POSException.Unsupported, "Versions aren't supported"

        self._lock_acquire()
        try:
            self._mapper_resource.access(self)  # Update mapper
            keychain = self._oid_encoder.decode(oid)

            # First detect conflicts.
            # The "h64" argument, if its value is not 0,
            # was previously generated by hash64().
            if DEBUG:
                print 'storing', `oid`, `h64`
            if h64 != HASH0:
                # Overwriting an old object.  Use the hash to verify
                # that the new data was derived from the old data.
                event, old_cs, old_hash = self._gwio.load(keychain)
                old_h64 = self.hash64(old_hash)
                if h64 != old_h64:
                    raise POSException.ConflictError(
                        "Storing %s based on old data. %s != %s" % (
                        repr(keychain),
                        repr(h64), repr(old_h64)))
            else:
                # A new object.  Attempts to load should lead to
                # NoStateFoundError or a hash of None, otherwise
                # there's a conflict.
                try:
                    event, cs, old_hash = self._gwio.load(keychain)
                except NoStateFoundError:
                    pass
                else:
                    if old_hash is not None:
                        raise POSException.ConflictError(
                            "%s already exists" % repr(keychain))

            # Now unpickle and store the data.
            file = StringIO(data)
            u = Unpickler(file)
            classified_state = u.load()
            event, new_hash = self._gwio.store(keychain, classified_state)
            new_h64 = self.hash64(new_hash)
            if self._scanner is not None:
                gw = event.getMapper().getGateway()
                sources = gw.getSources(event)
                self._scanner.setUncommittedSources(self._serial, oid, sources)
        finally:
            self._lock_release()

        if DEBUG:
            print 'stored', `oid`, `h64`, `new_h64`
        return new_h64

    def getSources(self, oid):
        keychain = self._oid_encoder.decode(oid)
        self._lock_acquire()
        try:
            return self._gwio.getSources(keychain)
        finally:
            self._lock_release()

    def new_oid(self):
        keychain = self._gwio.newKeychain()
        return self._oid_encoder.encode(keychain)

    def lastTransaction(self):
        return self._ltid

    def _clear_temp(self):
        pass

    def _abort(self):
        for c in self._conn_list:
            c.abort()
        if self._scanner is not None:
            self._scanner.afterAbort(self._serial)

    def _begin(self, tid, u, d, e):
        for c in self._conn_list:
            c.begin()

    def _finish(self, tid, user, desc, ext):
        for c in self._conn_list:
            c.finish()
        self._ltid = self._serial
        if self._scanner is not None:
            self._scanner.afterCommit(self._serial)

    def _vote(self):
        for c in self._conn_list:
            c.vote()

    def pack(self, t, referencesf):
        pass

    def _splat(self):
        """Spit out a string showing state.
        """
        return ''

    def close(self):
        for c in self._conn_list:
            c.close()
        self._mapper_resource.release(self)

