##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Zope 2 object classification.

$Id: classifier.py,v 1.2 2004/01/09 14:05:33 johanc Exp $
"""

from mimetypes import guess_extension

from Acquisition import aq_base
from OFS.ObjectManager import ObjectManager

from apelib.core.interfaces import IConfigurableClassifier
from apelib.core.exceptions import SerializationError, DeserializationError

# guess_extension() is useful, but it's unoptimized and sometimes
# chooses strange extensions.  fixed_extensions does nothing other than
# suggest a filename extension given a content type.  It contains
# some better defaults.
fixed_extensions = {
    'text/plain': '.txt',
    'text/html':  '.html',
    'image/png':  '.png',
    'image/jpg':  '.jpg',
    'image/jpeg': '.jpg',
    'image/gif':  '.gif',
    'application/octet-stream': '',  # No extension--too overloaded.
    }


class Classifier:
    """A classifier with some minimal Zope 2 extensions.
    """
    __implements__ = IConfigurableClassifier


    def __init__(self, gw=None):
        self.gw = gw
        self.key_to_mapper = {}
        self.class_to_mapper = {}
        self.ext_to_mapper = {}
        self.fallback_to_mapper = {}
        self.options = {}  # { (mapper_name, option) -> value }

    def getGateway(self):
        return self.gw


    def register(self, condition, value, mapper_name):
        if condition == 'key':
            self.key_to_mapper[value] = mapper_name
        elif condition == 'extension':
            self.ext_to_mapper[value] = mapper_name
        elif condition == 'class':
            self.class_to_mapper[value] = mapper_name
        elif condition == 'fallback':
            assert value in ('directory', 'file',
                             'folderish_object', 'fileish_object'), value
            self.fallback_to_mapper[value] = mapper_name
        else:
            raise ValueError('Unknown classification condition: %s'
                             % repr(condition))


    def setOption(self, mapper_name, option, value):
        assert option in ('default_extension', 'content_type_attr'), option
        self.options[(mapper_name, option)] = value
            

    def classifyObject(self, value, keychain):
        """Chooses a mapper and classification for storing an object."""
        mapper_name = self.key_to_mapper.get(keychain[-1])
        if mapper_name is not None:
            # Shortcut.
            return {'mapper_name': mapper_name}, mapper_name
        klass = value.__class__
        class_name = '%s.%s' % (klass.__module__, klass.__name__)
        classification = {'class_name': class_name}
        mapper_name = self.class_to_mapper.get(class_name)
        if mapper_name is None:
            folderish = isinstance(value, ObjectManager)
            # Store in a fallback format
            if folderish:
                fallback = 'folderish_object'
            else:
                fallback = 'fileish_object'
            mapper_name = self.fallback_to_mapper.get(fallback)
            if mapper_name is not None:
                # Force a particular mapper
                classification['mapper_name'] = mapper_name
        if mapper_name is None:
            raise SerializationError(
                'No mapper known for class %s' % repr(class_name))
        cta = self.options.get((mapper_name, 'content_type_attr'))
        if cta is not None:
            ct = str(getattr(aq_base(value), cta, None))
            ext = fixed_extensions.get(ct)
            if ext is None:
                ext = guess_extension(ct)
        else:
            ext = self.options.get((mapper_name, 'default_extension'))
        if ext:
            classification['extension'] = ext

        return classification, mapper_name


    def classifyState(self, event):
        """Chooses a mapper and classification for loading an object."""
        mapper_name = self.key_to_mapper.get(event.getKey())
        if mapper_name is not None:
            # Shortcut.
            return {'mapper_name': mapper_name}, mapper_name
        classification, serial = self.gw.load(event)
        class_name = classification.get('class_name')
        if class_name and ':' in class_name:
            # Backward compatibility
            class_name = class_name.replace(':', '.')
            classification['class_name'] = class_name
        mapper_name = classification.get('mapper_name')
        if mapper_name is None:
            # bw compat: look for certain meta_types.
            mt = classification.get('meta_type')
            if mt == '(folderish object)':
                mapper_name = self.fallback_to_mapper.get('folderish_object')
            elif mt == '(fileish object)':
                mapper_name = self.fallback_to_mapper.get('fileish_object')
        if mapper_name is None and class_name is not None:
            mapper_name = self.class_to_mapper.get(class_name)
        if mapper_name is None:
            t = classification.get('node_type')
            if t == 'd':
                # Directory
                mapper_name = self.fallback_to_mapper.get('directory')
            elif t == 'f':
                # File
                ext = classification.get('extension')
                if ext:
                    if not ext.startswith('.'):
                        ext = '.' + ext
                    mapper_name = self.ext_to_mapper.get(ext.lower())
                if not mapper_name:
                    mapper_name = self.fallback_to_mapper.get('file')
        if mapper_name is None:
            raise DeserializationError(
                'No mapper known for class %s' % repr(class_name))
        return classification, mapper_name


    def store(self, event, classification):
        """Stores the classification of an object."""
        return self.gw.store(event, classification)

