##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Serializers for OFSP (object file system product) objects

$Id: ofsserial.py,v 1.3 2004/01/10 17:54:26 johanc Exp $
"""

from cPickle import dumps, loads
from types import DictType, StringType

from Acquisition import aq_base
from OFS.SimpleItem import Item_w__name__
from OFS.ObjectManager import ObjectManager
from OFS.Image import File
from OFS.PropertyManager import PropertyManager

from apelib.core.interfaces import ISerializer
from apelib.core.schemas import FieldSchema, RowSequenceSchema
from apelib.core.exceptions import SerializationError
from apelib.core.serializers import OptionalSerializer
from apelib.zodb3.gateways import ReadOnlyItems


string_repr_types = {
    # Properties that are safe to render as strings for storage.
    # Other kinds of properties get pickled.
    'string': 1,
    'float': 1,
    'int': 1,
    'long': 1,
    'date': 1,
    'date_international': 1,
    'text': 1,
    'boolean': 1,
}


#EP hack to avoid indexing of object (due to overriden _setPropValue)
#before the context has been setup.
from ZPublisher.Converters import type_converters
from types import ListType
def _updateProperty(self, id, value):
    # Update the value of an existing property. If value
    # is a string, an attempt will be made to convert
    # the value to the type of the existing property.
    self._wrapperCheck(value)
    if not self.hasProperty(id):
        raise 'Bad Request', 'The property %s does not exist' % escape(id)
    if type(value)==type(''):
        proptype=self.getPropertyType(id) or 'string'
        if type_converters.has_key(proptype):
            value=type_converters[proptype](value)
    #self._setPropValue(id, value)
    if type(value) == ListType:
        value = tuple(value)
    setattr(self,id,value)



class FilePData:
    """Serializer of the 'data' attribute of OFS.File and OFS.Image"""

    __implements__ = ISerializer

    schema = FieldSchema('data', 'string')

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return isinstance(object, File)

    def serialize(self, object, event):
        event.notifySerialized('data', object.data, 1)
        event.ignoreAttribute('size')
        event.ignoreAttribute('width')
        event.ignoreAttribute('height')
        return str(object.data)

    def deserialize(self, object, event, state):
        data, size = object._read_data(state)
        if not object.__dict__.get('content_type'):
            # Guess the content type.
            content_type = object._get_content_type(
                state, data, object.__name__)
        else:
            # The properties serializer is authoritative.  Defer to it.
            content_type = None
        object.update_data(data, content_type, size)
        event.notifyDeserialized('data', object.data)


class FolderItems:
    """Zope 2 folder items (de)serializer
    """
    __implements__ = ISerializer

    schema = RowSequenceSchema()
    schema.addField('id', 'string', 1)
    schema.addField('keychain', 'keychain')

    # The fixed_oids flag must be turned on when serializing to
    # an object system with meaningful OIDs (like the filesystem.)
    # It doesn't need to be on for object systems with arbitrary
    # OIDs such as SQL databases.
    fixed_oids = 1

    def getSchema(self):
        return self.schema

    def canSerialize(self, obj):
        return isinstance(obj, ObjectManager)

    def serialize(self, obj, event):
        assert isinstance(obj, ObjectManager), repr(obj)
        state = []
        event.ignoreAttribute('_objects')
        mps = getattr(obj, '_mount_points', None)
        for id, subob in obj.objectItems():
            if mps and mps.has_key(id):
                # Store the mount point rather than the mounted object.
                subob = mps[id]
            base = aq_base(subob)
            keychain = event.identifyObject(base)
            if keychain is None:
                keychain = event.makeKeychain(id, 1)
            event.notifySerializedRef(id, base, 1, keychain)
            state.append((id, keychain))
        if self.fixed_oids:
            event.ignoreAttribute('_use_fixed_oids_')
            # Add a marker that tells the folder it has to move/rename
            # in a special way.  The _setOb patch sees this attribute.
            obj._use_fixed_oids_ = 1
        return state

    def deserialize(self, obj, event, state):
        assert isinstance(obj, ObjectManager)
        if self.fixed_oids:
            obj._use_fixed_oids_ = 1
        for (id, keychain) in state:
            subob = event.dereference(id, keychain)
            setattr(obj, id, subob)
            obj._objects += ({'id': id, 'meta_type':
                                 subob.__class__.meta_type},)



class IdAttribute:
    """Zope 2 id attribute."""

    __implements__ = ISerializer

    schema = FieldSchema('id', 'string')

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return 1

    def getAttrNameFor(self, object):
        if isinstance(object, Item_w__name__):
            return '__name__'
        else:
            return 'id'

    def serialize(self, object, event):
        attrname = self.getAttrNameFor(object)
        id = getattr(object, attrname)
        assert id, 'ID of %r is %r' % (object, id)
        event.notifySerialized(attrname, id, 1)
        return id

    def deserialize(self, object, event, state):
        attrname = self.getAttrNameFor(object)
        setattr(object, attrname, state)
        # Allow references under either attribute name.
        event.notifyDeserialized('id', state)
        event.notifyDeserialized('__name__', state)



class OFSProperties:
    """Serializer for OFS.PropertyManager properties."""

    __implements__ = ISerializer

    schema = RowSequenceSchema()
    schema.addField('id', 'string', 1)
    schema.addField('type', 'string')
    schema.addField('data', 'string')

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return isinstance(object, PropertyManager)

    def serialize(self, object, event):
        res = []
        assert isinstance(object, PropertyManager), repr(object)
        assert object._properties is object._propertyMap()
        event.ignoreAttribute('_properties')
        for p in object._properties:
            name = p['id']
            t = p['type']
            event.ignoreAttribute(name)
            data = object.getProperty(name)
            if t == 'lines':
                v = '\n'.join(data)
            elif string_repr_types.get(t):
                v = str(data)
            else:
                # Pickle the value and any extra info about the property.
                # Extra info is present in select and multi-select properties.
                d = p.copy()
                del d['id']
                del d['type']
                if d.has_key('mode'):
                    del d['mode']
                d['value'] = data
                v = dumps(d)
            res.append((name, t, v))
        return res

    def deserialize(self, object, event, state):
        assert isinstance(object, PropertyManager)
        assert object._properties is object._propertyMap()
        if not state:
            # No stored properties.  Revert the object to its
            # class-defined property schema.
            if object.__dict__.has_key('_properties'):
                del object._properties
            return

        old_props = object.propdict()
        new_props = {}
        for id, t, v in state:
            p = old_props.get(id)
            if p is None:
                p = {'mode': 'wd'}
            else:
                p = p.copy()
            p['id'] = id
            p['type'] = t
            if v and not string_repr_types.get(t) and t != 'lines':
                # v is a pickle.
                # Check the pickle for extra property info.
                d = loads(v)
                if isinstance(d, DictType):
                    del d['value']
                    if d:
                        # The data is stored with extra property info.
                        p.update(d)
            new_props[id] = p

        if old_props != new_props:
            object._properties = tuple(new_props.values())

        for id, t, v in state:
            if t == 'lines':
                data = v.split('\n')
            elif t == 'boolean':
                if (not v
                    or v == '0'
                    or v[:1].lower() in 'fn'): # match 0, [f]alse, [n]o
                    data = 0
                else:
                    data = 1
            elif string_repr_types.get(t):
                data = str(v)
            elif v:
                d = loads(v)
                if isinstance(d, DictType):
                    # The data is stored with extra property info.
                    data = d['value']
                else:
                    data = d
            else:
                # Fall back to a default.
                data = ''
            _updateProperty(object, id, data)


class ReadOnlyRoot(ReadOnlyItems):
    """Zope 2 application root."""

    def __init__(self, root_key):
        try:
            root_key = int(root_key)
        except ValueError:
            root_key = str(root_key)
        ReadOnlyItems.__init__(self, {'Application': (root_key,)})


class OptionalOFSProperties(OptionalSerializer):

    def __init__(self):
        OptionalSerializer.__init__(self, OFSProperties(), [])

