##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################
"""DBTab and DatabaseFactory classes.

$Id: DBTab.py,v 1.2 2004/01/09 14:05:14 johanc Exp $
"""

import os
import sys
from ConfigParser import ConfigParser
from thread import allocate_lock
builtin = sys.modules['__builtin__']

from ZODB.ActivityMonitor import ActivityMonitor
import Globals

from Exceptions import DBTabConfigurationError
from StorageTypes import storage_types, asBoolean
from ClassFactories import class_factories


database_types = {
    'DB': 'ZODB.DB',
    }

connection_types = {
    'Connection': 'ZODB.Connection',
    'LowConflictConnection': 'Products.TemporaryFolder.LowConflictConnection',
    }


class DatabaseFactory:

    # Core options
    storage_class = None
    storage_args = None
    database_class = None
    database_args = None
    connection_class = None

    # Mount options
    mount_paths = None        # ((virtual_path, real_root, real_path),)
    container_class = None    # string: "module_name.class_name"

    # Other options
    class_factory = None
    open_at_startup = None


    def __init__(self, name):
        self.name = name

    def getName(self):
        return self.name

    def setStorageParams(self, params):
        c = params.get('type')
        if not c:
            raise DBTabConfigurationError, 'A storage type is required.'
        module = None
        pos = c.rfind('.')
        if pos >= 0:
            # Specified the module
            module = c[:pos]
            c = c[pos + 1:]
        converter = None
        info = storage_types.get(c)
        if info:
            if not module or info[0] == module:
                # Use a default argument converter and maybe a default module.
                module, converter = info
        elif not module:
            raise DBTabConfigurationError(
                'Unknown storage type: %s' % c)
        m = __import__(module, globals(), locals(), [c])
        self.storage_class = getattr(m, c)

        args = params.copy()
        args['name'] = self.name
        del args['type']
        if converter is not None:
            args = converter(**args)
        self.storage_args = args


    def _convertDatabaseArgs(self, **kw):
        """Converts the arguments to be passed to the DB constructor."""
        for name in (
            'pool_size', 'cache_size', 'cache_deactivate_after',
            'version_pool_size', 'version_cache_size',
            'version_cache_deactivate_after',
            ):
            if kw.has_key(name):
                kw[name] = int(kw[name])
        if not kw.has_key('cache_size'):
            # Override the ZODB default of 400.
            kw['cache_size'] = 5000
        return kw


    def computeMountPaths(self, spec):
        mps = []
        # Allow multiple mounts by splitting on semicolons.
        parts = spec.split(';')
        for part in parts:
            part = part.strip()
            if part:
                real_root = None
                if ':' in part:
                    # 'virtual_path:real_path'
                    virtual_path, real_path = part.split(':', 1)
                    if real_path.startswith('~'):
                        # Use a special root.
                        # 'virtual_path:~real_root/real_path'
                        real_root, real_path = real_path[1:].split('/', 1)
                else:
                    # Virtual path is the same as the real path.
                    virtual_path = real_path = part
                mps.append((virtual_path, real_root, real_path))
        return mps


    def setDatabaseParams(self, params):
        c = params.get('class', 'DB')
        module = None
        pos = c.rfind('.')
        if pos >= 0:
            module = c[:pos]
            c = c[pos + 1:]
        if not module:
            # Use a default module.
            module = database_types.get(c)
            if not module:
                raise DBTabConfigurationError(
                    'Unknown database class: %s' % c)
        m = __import__(module, globals(), locals(), [c])
        self.database_class = getattr(m, c)

        c = params.get('connection_class', '')
        if c:
            pos = c.rfind('.')
            module = None
            if pos >= 0:
                module = c[:pos]
                c = c[pos + 1:]
            if not module:
                # Use a default module.
                module = connection_types.get(c)
                if not module:
                    raise DBTabConfigurationError(
                        'Unknown connection class: %s' % c)
            m = __import__(module, globals(), locals(), [c])
            self.connection_class = getattr(m, c)
        else:
            self.connection_class = None

        # Pull out the options for the factory rather than the database.
        args = params.copy()
        if args.has_key('class'):
            del args['class']
        if args.has_key('connection_class'):
            del args['connection_class']

        # Read string parameters
        for key in (
            'class_factory', 'container_class',
            ):
            if args.has_key(key):
                setattr(self, key, args[key])
                del args[key]
            else:
                setattr(self, key, None)

        # Read boolean parameters
        for key in (
            'open_at_startup',
            ):
            if args.has_key(key):
                v = asBoolean(args[key])
                setattr(self, key, v)
                del args[key]
            else:
                setattr(self, key, None)

        # Read deprecated parameters
        if args.has_key('auto_create'):
            # b/w compatibility
            if asBoolean(args['auto_create']) and not self.container_class:
                self.container_class = 'OFS.Folder.Folder'
            del args['auto_create']

        if args.has_key('mount_paths'):
            mps = args['mount_paths']
            self.mount_paths = self.computeMountPaths(mps)
            del args['mount_paths']
        else:
            self.mount_paths = None

        args = self._convertDatabaseArgs(**args)
        self.database_args = args


    def create(self):
        cf_name = self.class_factory or 'auto'
        cf = class_factories.get(cf_name)
        if cf is None:
            raise DBTabConfigurationError(
                'Unknown class factory: %s' % cf_name)
        storage = self.storage_class(**self.storage_args)
        database = self.database_class(storage, **self.database_args)
        if self.connection_class is not None:
            database.klass = self.connection_class
        database.setClassFactory(cf)
        # Each database has an independent record of activity.
        database.setActivityMonitor(ActivityMonitor())
        return database

    def getVirtualMountPaths(self):
        if not self.mount_paths:
            return ()
        return [item[0] for item in self.mount_paths]

    def getOpenAtStartup(self):
        return self.open_at_startup

    def getMountParams(self, mount_path):
        """Returns (real_root, real_path, container_class) for a virtual
        mount path.
        """
        if self.mount_paths:
            for (virtual_path, real_root, real_path) in self.mount_paths:
                if virtual_path == mount_path:
                    return (real_root, real_path, self.container_class)
        raise LookupError('Nothing known about mount path %s' % mount_path)



class DBTab:
    """A Zope database configuration, similar in purpose to /etc/fstab.
    """

    def __init__(self, filenames, fallback_filenames=None):
        self._started = 0
        self.filenames = filenames
        self.fallback_filenames = fallback_filenames
        self.opened = {}            # { name -> Database instance }
        self.lock = allocate_lock()

        (db_factories, mount_paths) = self._readConfig()
        self.db_factories = db_factories  # { name -> DatabaseFactory }
        self.mount_paths = mount_paths    # { virtual path -> name }


    def startup(self):
        """Opens the databases set to open_at_startup."""
        if self._started:
            return
        self._started = 1
        for name, factory in self.db_factories.items():
            if factory.getOpenAtStartup():
                self.getDatabase(name=name)


    def listMountPaths(self):
        """Returns a sequence of (virtual_mount_path, database_name).
        """
        return self.mount_paths.items()


    def listDatabaseNames(self):
        """Returns a sequence of names.
        """
        return self.db_factories.keys()


    def hasDatabase(self, name):
        """Returns true if name is the name of a configured database."""
        return self.db_factories.has_key(name)


    def _mountPathError(self, mount_path):
        if mount_path == '/':
            raise DBTabConfigurationError(
                "No root database configured")
        else:
            raise DBTabConfigurationError(
                "No database configured for mount point at %s"
                % mount_path)


    def getDatabase(self, mount_path=None, name=None, is_root=0):
        """Returns an opened database.  Requires either mount_path or name.
        """
        self.startup()
        if name is None:
            if mount_path is None:
                raise ValueError('Either mount_path or name is required')
            name = self.mount_paths.get(mount_path)
            if name is None:
                self._mountPathError(mount_path)
        db = self.opened.get(name)
        if db is None:
            if not self.db_factories.has_key(name):
                raise KeyError('%s is not a configured database' % repr(name))
            self.lock.acquire()
            try:
                # Check again, since the database may have been created
                # by another thread before the lock was acquired.
                db = self.opened.get(name)
                if db is None:
                    db = self._createDatabase(name, is_root)
            finally:
                self.lock.release()
        return db


    def getDatabaseFactory(self, mount_path=None, name=None):
        if name is None:
            if mount_path is None:
                raise ValueError('Either mount_path or name is required')
            name = self.mount_paths.get(mount_path)
            if name is None:
                self._mountPathError(mount_path)
        return self.db_factories[name]


    def _createDatabase(self, name, is_root):
        factory = self.db_factories[name]
        db = factory.create()
        self.opened[name] = db
        if not is_root:
            Globals.opened.append(db)
        # If it is the root database, Zope will add the database to
        # Globals.opened.  A database should not be listed twice.
        return db

    def _parseParams(self, config, section, vars=None):
        params = {}
        for k in config.options(section):
            v = config.get(section, k, vars=vars)
            k = k.lower()
            params[k] = v
        return params


    def _getEnvVars(self):
        res = os.environ.copy()
        if hasattr(res, 'data'):
            # Pull the data out if it's a UserDict instance.
            res = res.data
        for name in (
            'ZOPE_HOME', 'SOFTWARE_HOME', 'CLIENT_HOME', 'INSTANCE_HOME',
            ):
            v = getattr(builtin, name, None)
            if v is not None:
                res[name] = v
        return res


    def _readConfig(self):
        """Reads the configuration, returning
        (db_factories, virtual_mount_paths).
        """
        config = ConfigParser()
        config.read(self.filenames)
        if not config.sections() and self.fallback_filenames:
            config.read(self.fallback_filenames)
        vars = self._getEnvVars()
        db_factories = {}
        mount_paths = {}
        for section in config.sections():
            s1 = section.strip()
            s2 = s1.lower()
            if s2.startswith('storage:'):
                name = s1[8:].strip()
                factory = db_factories.get(name)
                if factory is None:
                    factory = DatabaseFactory(name)
                    db_factories[name] = factory
                params = self._parseParams(config, section, vars)
                factory.setStorageParams(params)
            elif s2.startswith('database:'):
                name = s1[9:].strip()
                factory = db_factories.get(name)
                if factory is None:
                    factory = DatabaseFactory(name)
                    db_factories[name] = factory
                params = self._parseParams(config, section, vars)
                factory.setDatabaseParams(params)
                mps = factory.getVirtualMountPaths()
                for mp in mps:
                    if mount_paths.has_key(mp):
                        raise DBTabConfigurationError(
                            "Attempted to mount more than one "
                            "database at %s" % mp)
                    mount_paths[mp] = name
        return db_factories, mount_paths

