##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""DBTab default storage types.

$Id: StorageTypes.py,v 1.2 2004/01/09 14:05:14 johanc Exp $
"""

import re

from Exceptions import DBTabConfigurationError


def asBoolean(s):
    if not s:
        return 0
    ss = str(s).lower()
    if ss in ('1', 'y', 't', 'yes', 'true', 'on'):
        return 1
    elif ss in ('0', 'n', 'f', 'no', 'false', 'off'):
        return 0
    else:
        raise ValueError, 'Not a boolean value: %s' % s


def convertFileStorageArgs(quota=None, stop=None, **kw):
    if kw.has_key('name'):
        # FileStorage doesn't accept a 'name' arg
        del kw['name']
    if quota is not None:
        kw['quota'] = long(quota) or None
    if stop is not None:
        stop = long(stop)
        if not stop:
            stop = None
        else:
            from ZODB.utils import p64
            stop = p64(stop)
        kw['stop'] = stop

    # Boolean args
    for name in (
        'create', 'read_only'
        ):
        if kw.has_key(name):
            kw[name] = asBoolean(kw[name])

    return kw


# Match URLs of the form 'zeo://zope.example.com:1234'
zeo_url_re = re.compile('zeo:/*(?P<host>[A-Za-z0-9\.-]+):(?P<port>[0-9]+)')

def convertAddresses(s):
    # Allow multiple addresses using semicolons as a split character.
    res = []
    for a in s.split(';'):
        a = a.strip()
        if a:
            mo = zeo_url_re.match(a)
            if mo is not None:
                # ZEO URL
                host, port = mo.groups()
                res.append((host, int(port)))
            else:
                # Socket file
                res.append(a)
    return res


def convertClientStorageArgs(addr=None, **kw):
    if addr is None:
        raise DBTabConfigurationError(
            'An addr parameter is required for ClientStorage.')
    kw['addr'] = convertAddresses(addr)

    # Integer args
    for name in (
        'cache_size', 'min_disconnect_poll', 'max_disconnect_poll',
        ):
        if kw.has_key(name):
            kw[name] = int(kw[name])

    # Boolean args
    for name in (
        'wait', 'read_only', 'read_only_fallback',
        ):
        if kw.has_key(name):
            kw[name] = asBoolean(kw[name])

    # The 'client' parameter must be None to be false.  Yuck.
    if kw.has_key('client') and not kw['client']:
        kw['client'] = None

    return kw


def convertBDBStorageArgs(**kw):
    from BDBStorage.BerkeleyBase import BerkeleyConfig
    config = BerkeleyConfig()
    for name in dir(BerkeleyConfig):
        if name.startswith('_'):
            continue
        val = kw.get(name)
        if val is not None:
            if name != 'logdir':
                val = int(val)
            setattr(config, name, val)
            del kw[name]
    # XXX: Nobody ever passes in env
    assert not kw.has_key('env')
    kw['config'] = config
    return kw


def convertDemoStorageArgs(base_type=None, **kw):
    if base_type:
        # XXX partly copied from DBTab.py
        pos = base_type.rfind('.')
        if pos >= 0:
            # Specified the module
            module = base_type[:pos]
            class_name = base_type[pos + 1:]
        else:
            module = None
            class_name = base_type
        converter = None
        info = storage_types.get(class_name)
        if info:
            if not module or info[0] == module:
                # Use a default argument converter and maybe a default module.
                module, converter = info
        elif not module:
            raise DBTabConfigurationError(
                'Unknown storage type: %s' % base_type)
        m = __import__(module, globals(), locals(), [class_name])
        base_class = getattr(m, class_name)

        # Pass all of the remaining arguments to the base storage constructor.
        base_args = kw
        if converter is not None:
            base_args = converter(**base_args)

        base = base_class(**base_args)
        kw = {'base': base, 'name': kw['name']}
    else:
        base = None
    return kw


storage_types = {
    'FileStorage': ('ZODB.FileStorage', convertFileStorageArgs),
    'DemoStorage': ('ZODB.DemoStorage', convertDemoStorageArgs),
    'MappingStorage': ('ZODB.MappingStorage', None),
    'TemporaryStorage': ('Products.TemporaryFolder.TemporaryStorage', None),
    'ClientStorage': ('ZEO.ClientStorage', convertClientStorageArgs),
    'BDBFullStorage': ('BDBStorage.BDBFullStorage', convertBDBStorageArgs),
    'BDBMinimalStorage':
    ('BDBStorage.BDBMinimalStorage', convertBDBStorageArgs),
    }
